/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "gmp/gmp.h"
#include "gc/gc.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>

/*-------------------------------------------------------
 *  The 'SWiT' datum (= atom or pair)
 *-------------------------------------------------------
 */
struct SWiT {
  void *type; /* or cdr */
  void *data; /* or car */
};
typedef struct SWiT *SWiT;

/*------------------------------------------------------------------------
 * When the datum is an atom, the type field is (coerced to) an odd
 * integer. When it's a pair, the type field is a pointer that is always even.
 * The data field must be coerced to the corresponding type.
 * In R5RS, the disjoint types are said to be:
 *
 *   boolean?          pair?
 *   symbol?           number?
 *   char?             string?
 *   vector?           port?
 *   procedure?
 *
 * null? and eof-object? also define two other types, disjointed from
 * the above list. Moreover, 'promises' that are values created with
 * 'delay' are distinct from the other types, but a predicate for them
 * has been forgotten (in R5RS). In SWiT, we provide promise? that is true 
 * iff its argument is a 'promise'.
 *
 * In SWiT, there are types that are not defined in standard Scheme,
 * they are caracterized by the following predicates:
 *
 *   unbound-object?   socket?
 *   environment?      queue?
 *   quad?             turtle?
 *   bezier?
 *
 * Some more types may be added in the future.
 *
 * Some subtypes are also defined, e.g.:
 *  numbers: integer? real? rational? complex? 
 *           fixnum? bignum? exact? inexact? positive? negative? etc.
 *  procedures: closure? continuation? primitive?
 *  ports: input-port? output-port? string-port? socket-port?
 *  symbols: macro? keyword? gentemp? type?
 *  characters: char-alphabetic? char-numeric? char-whitespace?
 *               char-upper-case? char-lower-case?
 *
 * Some constants (as ##eof, ##unbound, empty list)
 * are unique member of their `type' (and uniquely defined).
 * Other constants as 'true', 'false', empty string, 'zero', 'one',
 * all 8bits-characters, empty vector, keywords, belong to a type but
 * are uniquely defined (i.e. never duplicated).
 *------------------------------------------------------------------------
 */
#define SWIT_TYPE(obj) ((unsigned int)((obj)->type))
#define SWIT_DATA(obj, ctype) ((ctype)((obj)->data))

/*------------------------------------------------------------------------
 * When the datum is a pair, the type field becomes the 'cdr', and the
 * data field is the 'car'. They are both pointers to other encapsulated
 * data.
 *------------------------------------------------------------------------
 *
 * A pair is distinguished from atoms because its type field (a pointer) 
 * is even:
 */
#define SWIT_PAIRP(obj) ((SWIT_TYPE(obj) % 2) == 0)

/* Two bits (the second and the third) of the car are masked 
 * and can hold marks:
 */
#define SWIT_CAR(obj) ((SWiT)(((unsigned int)((obj)->data)) & 0xfffffff9))
#define SWIT_CDR(obj) ((SWiT)((obj)->type))

#define SET_CAR(cell, val) ((cell)->data = val)
#define SET_CDR(cell, val) ((cell)->type = val)

/*-----------------------------------------------------------------------
 *   An environment is a linked list of 'frames'. A frame is an array of
 *   values (but the names of the formals are also kept). Keeping names
 *   is necessary : 1) to disassemble the code in a human readable way,
 *   2) for functions that operate on environments such as eval,
 *   symbol-value, set-symbol! etc. (see the files 'switSymbol.c',
 *   and 'switMisc.c'). The values kept in environments
 *   are generally accessed with their `lexical address' (computed at 
 *   compile-time).
 *-----------------------------------------------------------------------
 */
#define SWIT_ENV(obj) ((struct SWiT_Env *)((obj)->data))
#define ENV_FRAME(obj) ((obj)->values)
#define ENV_NAMES(obj) ((obj)->names)
#define ENV_NEXT(obj) ((obj)->next)

struct SWiT_Env
{
  SWiT *values;           /* vector of values */
  SWiT names;             /* list of names    */
  struct SWiT_Env *next;  /* NULL when global env is reached */
};
typedef struct SWiT_Env *SWiT_Env;


/*----------------------------------------------
 *
 *    DATA : string, vector,
 *    ^^^^   closure, continuation,
 *           promise, input port,
 *           number (bignum, ratio, real, complex)
 *           quads
 *
 *   (other data types are 'immediate' :
 *     fixnum, character, boolean)
 *----------------------------------------------
 */

/*-------------------------------------------------------------------
 *   Strings (implemented as byte arrays: they are not '\0'-terminated)
 *-------------------------------------------------------------------
 */

#define STRING_LENGTH(str) (((unsigned int)((str)->type)) >> 8)
#define STRING_ELS(str) ((void *)((str)->data))


/*--------------------
 *    Vectors
 *--------------------
 */
#define VECTOR_LENGTH(obj) (((unsigned int)((obj)->type)) >> 8)
#define VECTOR_ELS(obj) SWIT_DATA(obj, SWiT *)

/*--------------
 *  Closures
 *--------------
 */

#define SWIT_CLOSURE(obj) SWIT_DATA(obj, struct SWiT_Closure *)
#define CLOSURE_LBD(clo) ((clo)->lbd)
#define CLOSURE_ENV(clo) ((clo)->env)

struct SWiT_Closure
{
  struct SWiT_Lambda *lbd;
  SWiT_Env env;
};
typedef struct SWiT_Closure *SWiT_Closure;



/*--------------------------
 *  Continuations
 *--------------------------
 */
#define SWIT_CONT(obj) SWIT_DATA(obj, struct SWiT_Ccont *)
#define CCONT_ENV(cont) ((cont)->env)
#define CCONT_UNEV(cont) ((cont)->unev)
#define CCONT_VALS(cont) ((cont)->vals)
#define CCONT_NEXT(cont) ((cont)->next)

struct SWiT_Ccont {
  struct SWiT_Expr *unev;
  struct SWiT_Env *env;
  void *vals;
  struct SWiT_Ccont *next;
};
typedef struct SWiT_Ccont *SWiT_Ccont;


/*--------------
 *  Promises
 *--------------
 */
#define SWIT_PROMISE(obj) SWIT_DATA(obj, struct SWiT_Promise *)
#define PROMISE_VAL(obj) ((obj)->val)
#define PROMISE_EXPR(obj) ((obj)->expr)
#define PROMISE_ENV(obj) ((obj)->env)

struct SWiT_Promise
{
  SWiT val;                   /* val if already computed, NULL otherwise */
  struct SWiT_Expr *expr;     /* expression to be evaluated once */
  struct SWiT_Env *env;       /* saved environment for this evaluation */
};
typedef struct SWiT_Promise *SWiT_Promise;


/*-----------------
 *  Ports
 *-----------------
 */
typedef int(* io_fun)(int descr, void *buff, int cnt);
typedef int(* seek_fun)(int descr, int offset, int mode);
#define IO_BUFFER_SIZE 128
#define SWIT_PORT(obj) SWIT_DATA(obj, struct SWiT_Port *)
#define PORT_BUFFER(port) ((port)->buffer)
#define PORT_INDEX(port) ((port)->index)
#define PORT_EOB(port) ((port)->eob)
#define PORT_DESCR(port) ((port)->descr)
#define PORT_READ_FUN(port) ((port)->rfun)
#define PORT_WRITE_FUN(port) ((port)->wfun)
#define PORT_SEEK_FUN(port) ((port)->sfun)
#define PORT_JBUF(port) ((port)->jbuf)
#define PORT_LINENO(port) ((port)->line)

struct SWiT_Port
{
  int descr;                   /* `descriptor' (file, socket, string)  */
  char *buffer;                /* io buffer                            */
  unsigned int index;          /* current position in buffer           */
  unsigned int eob;            /* end of buffer                        */
  unsigned int line;           /* the current line #                   */
  io_fun rfun;                 /* read function                        */
  io_fun wfun;                 /* write function                       */
  seek_fun sfun;               /* seek function                        */
  jmp_buf *jbuf;               /* jump ptr for eof detection           */
};
typedef struct SWiT_Port *SWiT_Port;

/*----------------------
 * Bignums (from gmp)
 *----------------------
 */
#define SWIT_BIGNUM(obj) SWIT_DATA(obj, MP_INT *)
#define SWIT_BIG_VAL(obj) (*SWIT_BIGNUM(obj))


/*---------------------------------------------
 * Ratio (from gmp) ** NOT YET IMPLEMENTED **
 *---------------------------------------------
 */
#define SWIT_RATIO(obj) SWIT_DATA(obj, MP_RAT *)
#define SWIT_RATIO_VAL(obj) (*SWIT_RATIO(obj))


/*------------------------------------------------
 * Inexact real numbers (implemented as doubles)
 *------------------------------------------------
 */
#define SWIT_DBL(obj) SWIT_DATA(obj, double *)
#define SWIT_DBL_VAL(obj) (*SWIT_DBL(obj))

/*-------------------------------------------
 * Complex numbers ** NOT YET IMPLEMENTED **
 *-------------------------------------------
 */
#define SWIT_CPLX(obj) SWIT_DATA(obj, struct SWiT_Complex *)
#define CPLX_RE(obj) ((obj)->re)
#define CPLX_IM(obj) ((obj)->im)

struct SWiT_Complex
{
  SWiT re;
  SWiT im;
};
typedef struct SWiT_Complex *SWiT_Complex;


/*------------------------------------
 * Characters, fixnums
 *------------------------------------
 */
#define SWIT_CHAR(obj) SWIT_DATA(obj, int)
#define SWIT_FIXNUM(obj) SWIT_DATA(obj, long int)


/*---------------------------------------------------------------------
 *   Primitives (they look like symbols but they are not symbols ...)
 *---------------------------------------------------------------------
 */

/*
 *    Types of the C functions :
 */
typedef SWiT (* SWiT_Prim0) (void);
typedef SWiT (* SWiT_Prim0n) (int argc, SWiT *argv);
typedef SWiT (* SWiT_Prim1) (SWiT arg);
typedef SWiT (* SWiT_Prim1n) (SWiT arg, int argc, SWiT *argv);
typedef SWiT (* SWiT_Prim2) (SWiT arg1, SWiT arg2);
typedef SWiT (* SWiT_Prim2n) (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
typedef SWiT (* SWiT_Prim3) (SWiT arg1, SWiT arg2, SWiT arg3);

/*
 * Access macros :
 */
#define PRIM_FUN(prim) ((prim)->fun)
#define PRIM_NAME(prim) ((prim)->name)
#define PRIM_STRING(prim) (STRING_ELS(PRIM_NAME(prim)))
#define PRIM_NEXT(prim) ((prim)->next)
#define PRIM_SIG(prim) ((prim)->sig)
#define PRIM_T_ARG1(prim) ((prim)->targ1)
#define PRIM_T_ARG2(prim) ((prim)->targ2)
#define PRIM_T_ARG3(prim) ((prim)->targ3)
#define PRIM_T_ARGS(prim) ((prim)->targs)
#define PRIM_T_LAST(prim) ((prim)->tlast)
#define PRIM_VAL(prim) ((prim)->val)

#define CHECK_(arg, prim, val, mask) \
({unsigned int _x = SWIT_TYPE(val); \
  unsigned int _y = PRIM_T_ ## arg(prim); \
   (_y ? \
     ((_x == _y) ? \
       1 \
       : (((_y % 2) == 0) ? \
          (((_x % 2) == 0) ? \
             1 \
             : ((_x == NIL) ? (_y == LIST) : 0)) \
          : !((_x ^ _y) & (~_y >> 16))))\
    : 1);})

#define SWIT_PRIM0(obj) SWIT_DATA(obj, struct SWiT_Sym_prim0 *)
struct SWiT_Sym_prim0
{
  SWiT_Prim0 fun;      /* A pointer to the C function.       */
  SWiT next;           /* Next primitive in same entry.      */
  SWiT name;           /* Its name                           */
  SWiT sig;           /* The signature (for the typing system) */
};
typedef struct SWiT_Sym_prim0 *SWiT_Sym_prim0;

#define SWIT_PRIM0n(obj) SWIT_DATA(obj, struct SWiT_Sym_prim0n *)
struct SWiT_Sym_prim0n
{
  SWiT_Prim0n fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targs;      /* The type of arguments              */
  int tlast;      /* The type of the last argument      */
};
typedef struct SWiT_Sym_prim0n *SWiT_Sym_prim0n;

#define SWIT_PRIM1(obj) SWIT_DATA(obj, struct SWiT_Sym_prim1 *)
struct SWiT_Sym_prim1
{
  SWiT_Prim1 fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targ1;          /* The type of its argument               */
  SWiT val;           /* The default value when prim01 */
};
typedef struct SWiT_Sym_prim1 *SWiT_Sym_prim1;

#define SWIT_PRIM1n(obj) SWIT_DATA(obj, struct SWiT_Sym_prim1n *)
struct SWiT_Sym_prim1n
{
  SWiT_Prim1n fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targ1;      /* The type of its first argument     */
  int targs;      /* The type of variable part elements */
  int tlast;      /* The type of the last argument      */
};
typedef struct SWiT_Sym_prim1n *SWiT_Sym_prim1n;

#define SWIT_PRIM2(obj) SWIT_DATA(obj, struct SWiT_Sym_prim2 *)
struct SWiT_Sym_prim2
{
  SWiT_Prim2 fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targ1;          /* The type of its first argument          */
  int targ2;          /* The type of its second argument         */
  SWiT val;           /* The default value when prim12  */
};
typedef struct SWiT_Sym_prim2 *SWiT_Sym_prim2;

#define SWIT_PRIM2n(obj) SWIT_DATA(obj, struct SWiT_Sym_prim2n *)
struct SWiT_Sym_prim2n
{
  SWiT_Prim2n fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targ1;      /* The type of its first argument     */
  int targ2;      /* The type of its second argument    */
  int targs;      /* The type of variable part elements */
  int tlast;      /* The type of the last argument      */
};
typedef struct SWiT_Sym_prim2n *SWiT_Sym_prim2n;

#define SWIT_PRIM3(obj) SWIT_DATA(obj, struct SWiT_Sym_prim3 *)
struct SWiT_Sym_prim3
{
  SWiT_Prim3 fun;
  SWiT next;
  SWiT name;
  SWiT sig;
  int targ1;      /* The type of its first argument  */
  int targ2;      /* The type of its second argument */
  int targ3;      /* The type of its third argument  */
};
typedef struct SWiT_Sym_prim3 *SWiT_Sym_prim3;

/*-------------------------------------------------------------
 *  EXPRESSIONS :  symbols, primitives, lambda,
 *  ^^^^^^^^^^^    variables, begin, if, cond,
 *                 case, quote, calls, set!
 *                 define, do, try
 *
 *-------------------------------------------------------------
 */

#define SWIT_INST(obj) ((obj)->inst)
#define SWIT_CODE(obj) ((obj)->code)

typedef struct SWiT_Expr *SWiT_Expr;

struct SWiT_Expr {
  unsigned int inst;/* the inst. # for dispatching in the evaluator */
  void *code;   /* coerced to the different structures defined below  */
};

/* Note : the ``inst'' part even and small (< 100) so that it cannot be
 *        be confused (by the evaluator) with self evaluating values --
 *        whose types are odd integers -- nor pairs -- whose cdr pointers
 *        are not marked and are certainly greater than 100.
 */

/*----------------------------------------------------------------
 *   Symbols :
 *
 * Note: a symbol can be 'encapsulated' as an expression
 * or a datum (this duality of symbols is specific to lisp).
 *----------------------------------------------------------------
 */
#define SWIT_SYM(obj) SWIT_DATA(obj, struct SWiT_Symbol *)
#define SWIT_SYMBOL(obj) ((struct SWiT_Symbol *)((obj)->code))
#define SYMBOL_VAL(sym) ((sym)->val)
#define SYMBOL_NAME(sym) ((sym)->name)
#define SYMBOL_STRING(sym) STRING_ELS(SYMBOL_NAME(sym))
#define SYMBOL_NEXT(sym) ((SWiT)(((unsigned int)((sym)->next)) & 0xfffffff8))
#define SET_SYMBOL_NEXT(sym, val) ((sym)->next = val)
#define SYMBOL_PLIST(sym) ((sym)->plist)
#define SYMBOL_FLAGS(sym) (((unsigned int)((sym)->next)) & 0x00000007)
#define SET_SYMBOL_FLAG(sym, flag) (((unsigned int)((sym)->next)) |= flag)
#define SYMBOL_READ_ONLY_MASK 1
#define SYMBOL_KEYWORD_MASK 2
#define SYMBOL_MACRO_MASK 3
#define SYMBOL_GENTEMP_MASK 4
#define SYMBOL_TYPE_MASK 5
#define SYMBOL_MACROP(sym) (SYMBOL_FLAGS(sym) == SYMBOL_MACRO_MASK)
#define SYMBOL_READ_ONLYP(sym) (SYMBOL_FLAGS(sym) == SYMBOL_READ_ONLY_MASK)
#define SYMBOL_KEYWORDP(sym) (SYMBOL_FLAGS(sym) == SYMBOL_KEYWORD_MASK)
#define SYMBOL_GENTEMPP(sym) (SYMBOL_FLAGS(sym) == SYMBOL_GENTEMP_MASK)
#define SYMBOL_TYPEP(sym) (SYMBOL_FLAGS(sym) == SYMBOL_TYPE_MASK)

struct SWiT_Symbol
{
  SWiT val;                  /* The value of the (global) symbol.     */
  SWiT name;                 /* its name : returned by symbol->string */
  SWiT plist;                /* The property list                     */
  void *next;                /* Next symbol in same hash table entry. */
};
typedef struct SWiT_Symbol *SWiT_Symbol;


/*------------------------
 *   'Lambda' form :
 *------------------------
 */
#define SWIT_LAMBDA(obj) ((struct SWiT_Lambda *)((obj)->code))
#define LAMBDA_NPARAMS(obj) ((obj)->nparams)
#define LAMBDA_NAMES(obj) ((obj)->names)
#define LAMBDA_BODY(obj) ((obj)->body)
#define LAMBDA_PCODE(obj) ((obj)->pcode)

struct SWiT_Lambda
{
  unsigned int nparams; /* Number of formals in the lower half-word
			    If there is a rest argument, the upper
			    half-word is non-zero */
  SWiT names;           /* (Proper) list of the formals names */
  SWiT_Expr body;       /* Body of the lambda-expression.
			    If there was more than one expression, it has
			    been already converted into a 'begin' form */
  SWiT pcode;           /* the pseudo-code */
};
typedef struct SWiT_Lambda *SWiT_Lambda;


/*---------------------------------------------------------------------
 *   Local variables :
 *
 *  db = depth is the 'de Bruijn' code
 *  idx = index in the list of parameters (compiled as a vector)
 *
 *---------------------------------------------------------------------
 */

#define SWIT_VAR(obj) ((unsigned long int)((obj)->code))
#define VAR_DB(var) (var & 0xffff)
#define VAR_IDX(var) (var >> 16)


/*--------------------------------------------------------
 *   'Begin' form (also used for vector of expressions):
 *-------------------------------------------------------
 */
#define SWIT_BEGIN(obj) ((struct SWiT_Begin *)((obj)->code))
#define BEG_NFORMS(obj) ((obj)->nforms)
#define BEG_FORMS(obj) ((obj)->forms)
struct SWiT_Begin
{
  unsigned int nforms;     /* number of forms : always > 0  */
  SWiT_Expr *forms;        /* vector of forms               */
};
typedef struct SWiT_Begin *SWiT_Begin;


/*-----------------------------------
 *  'set!' (global variable) form :
 *-----------------------------------
 */
#define SWIT_GSET(obj) ((struct SWiT_Gset *)((obj)->code))
#define GSET_SYM(obj) ((obj)->sym)
#define GSET_EXPR(obj) ((obj)->expr)

struct SWiT_Gset
{
  SWiT_Symbol sym;
  SWiT_Expr expr;
};
typedef struct SWiT_Gset *SWiT_Gset;


/*--------------------------
 *   'Define' form :
 *-------------------------
 */
#define SWIT_DEFINE(obj) ((struct SWiT_Define *)((obj)->code))
#define DEFINE_SYM(obj) ((obj)->sym)
#define DEFINE_EXPR(obj) ((obj)->expr)

struct SWiT_Define
{
  SWiT sym;
  SWiT_Expr expr;
};
typedef struct SWiT_Define *SWiT_Define;


/*--------------------------
 *   'Undefine' form :
 *-------------------------
 */
#define SWIT_UNDEFINE(obj) ((struct SWiT *)((obj)->code))

/*-------------------------------
 *  'set!' (local variable) form :
 *-------------------------------
 */
#define SWIT_LSET(obj) ((struct SWiT_Lset *)((obj)->code))
#define LSET_VAR(obj) ((obj)->var)
#define LSET_EXPR(obj) ((obj)->expr)

struct SWiT_Lset
{
  unsigned int var;
  SWiT_Expr expr;
};
typedef struct SWiT_Lset *SWiT_Lset;


/*-----------------------------------
 *   Procedure calls of primitives:
 *-----------------------------------
 */
#define CALL_PRIM(obj) ((obj)->prim)
#define CALL_FUN(obj) PRIM_FUN(CALL_PRIM(obj))
#define CALL_NAME(obj) PRIM_NAME(CALL_PRIM(obj))
#define CALL_STRING(obj) STRING_ELS(CALL_NAME(obj))
#define CALL_NARGS(obj) ((obj)->nargs)
#define CALL_ARG(obj) ((obj)->arg)
#define CALL_ARG1(obj) ((obj)->arg1)
#define CALL_ARG2(obj) ((obj)->arg2)
#define CALL_ARG3(obj) ((obj)->arg3)
#define CALL_ARGS(obj) ((obj)->args)
#define CALL_PROC(obj) ((obj)->proc)

#define SWIT_CALL0(obj) ((struct SWiT_Sym_prim0 *)((obj)->code))
#define SWIT_CALL0n(obj) ((struct SWiT_Call0n *)((obj)->code))
struct SWiT_Call0n
{
  SWiT_Sym_prim0n prim;
  unsigned int nargs;
  SWiT_Expr *args;
};
typedef struct SWiT_Call0n *SWiT_Call0n;

#define SWIT_CALL1n(obj) ((struct SWiT_Call1n *)((obj)->code))
struct SWiT_Call1n
{
  SWiT_Sym_prim1n prim;
  SWiT_Expr arg;
  unsigned int nargs;
  SWiT_Expr *args;
};
typedef struct SWiT_Call1n *SWiT_Call1n;

#define SWIT_CALL2n(obj) ((struct SWiT_Call2n *)((obj)->code))
struct SWiT_Call2n
{
  SWiT_Sym_prim2n prim;
  SWiT_Expr arg1;
  SWiT_Expr arg2;
  unsigned int nargs;
  SWiT_Expr *args;
};
typedef struct SWiT_Call2n *SWiT_Call2n;

#define SWIT_CALL1(obj) ((struct SWiT_Call1 *)((obj)->code))
struct SWiT_Call1
{
  SWiT_Sym_prim1 prim;
  SWiT_Expr arg;
};
typedef struct SWiT_Call1 *SWiT_Call1;

#define SWIT_CALL2(obj) ((struct SWiT_Call2 *)((obj)->code))
struct SWiT_Call2
{
  SWiT_Sym_prim2 prim;
  SWiT_Expr arg1;
  SWiT_Expr arg2;
};
typedef struct SWiT_Call2 *SWiT_Call2;

#define SWIT_CALL3(obj) ((struct SWiT_Call3 *)((obj)->code))
struct SWiT_Call3
{
  SWiT_Sym_prim3 prim;
  SWiT_Expr arg1;
  SWiT_Expr arg2;
  SWiT_Expr arg3;
};
typedef struct SWiT_Call3 *SWiT_Call3;


/*---------------------------
 *  General procedure call :
 *---------------------------
 */
#define SWIT_CALL(obj) ((struct SWiT_Call *)((obj)->code))
struct SWiT_Call
{
  SWiT_Expr proc;     /* An expression that must evaluate to a procedure */
  unsigned int nargs; /* Number of arguments, `arity check' will be  done (at run-time) after proc is evaluated  */
  SWiT_Expr *args;    /* The vector of arguments                     */
};
typedef struct SWiT_Call *SWiT_Call;


/*------------------------------------------
 *  'Let', 'Let*', 'Letrec' special forms :
 *------------------------------------------
 */
#define SWIT_LET(obj) ((struct SWiT_Let *)((obj)->code))
#define LET_LBD(obj) ((obj)->lbd)
#define LET_ARGS(obj) ((obj)->vargs)

struct SWiT_Let
{
  SWiT_Lambda lbd;
  SWiT_Expr *vargs;
};
typedef struct SWiT_Let *SWiT_Let;


/*-----------------
 *   'If' form :
 *-----------------
 */
#define SWIT_IF(obj) ((struct SWiT_If *)((obj)->code))
#define IF_TEST(obj) ((obj)->test)
#define IF_THEN(obj) ((obj)->pthen)
#define IF_ELSE(obj) ((obj)->pelse)

struct SWiT_If
{
  SWiT_Expr test;
  SWiT_Expr pthen;
  SWiT_Expr pelse;

};
typedef struct SWiT_If *SWiT_If;


/*-----------------
 *   'Try' form :
 *-----------------
 */
#define SWIT_TRY(obj) ((struct SWiT_Try *)((obj)->code))
#define TRY_EXPR(obj) ((obj)->expr)
#define TRY_PROTECT(obj) ((obj)->protect)
#define TRY_HDL(obj) ((obj)->hdl)
#define TRY_EXN(obj) ((obj)->exn)

struct SWiT_Try
{
  SWiT_Expr expr;    /* expression whose value is returned if exn not raised */
  SWiT exn;          /* list of exception `names' (usually symbols)  */
  SWiT_Expr protect; /* a protected form                             */
  SWiT_Expr *hdl;    /* handlers for these exceptions                */
};
typedef struct SWiT_Try *SWiT_Try;


/*------------------
 *  'Cond' form :
 *------------------
 */
#define SWIT_COND(obj) ((struct SWiT_Cond *)((obj)->code))
#define COND_TESTS(obj) ((obj)->tests)
#define COND_SEQS(obj) ((obj)->seqs)
#define COND_NCLAUSES(obj) ((obj)->nclauses)

struct SWiT_Cond
{
  SWiT_Expr *tests;         /* the vector of tests
			         (cond (test1 seq1) (test2 seq2) ...) */
  SWiT_Expr *seqs;          /* the vector of sequences following the
			        tests */
  unsigned int nclauses;    /* the common length of tests and seqs
			        vectors */
};
typedef struct SWiT_Cond *SWiT_Cond;


/*--------------------
 *   'Case' form :
 *--------------------
 */
#define SWIT_CASE(obj) ((struct SWiT_Case *)((obj)->code))
#define CASE_KEY(obj)   ((obj)->key)
#define CASE_TESTS(obj) ((obj)->tests)
#define CASE_SEQS(obj) ((obj)->seqs)
#define CASE_NCLAUSES(obj) ((obj)->nclauses)

struct SWiT_Case
{
  SWiT_Expr key;               /* the key in (case key ...) */
  SWiT *tests;                 /* the (constant) vector of tests:
			           (case key (test1 seq1) (test2 seq2) ...) */
  SWiT_Expr *seqs;             /* the vector of sequences following the
			           tests */
  unsigned int nclauses;       /* the common length of tests and seqs
				   vectors */
};
typedef struct SWiT_Case *SWiT_Case;


/*-------------------
 *  'Quote' form :
 *-------------------
 */
/* used in primitives for optionnal arguments : */
extern SWiT_Expr swit_make_quote (SWiT obj);

#define SWIT_QUOTE(obj) ((struct SWiT *)((obj)->code))


/*-----------------------------------------------
 * The 'eval' procedure defined in 'switEval.c'
 *-----------------------------------------------
 */
extern SWiT swit_eval(SWiT_Expr expr, SWiT_Env env, SWiT_Ccont);

struct SWiT_Exn {
  struct SWiT_Try *try;
  struct SWiT_Env *env;
  struct SWiT_Ccont *cont;
  struct SWiT_Exn *next;
};
typedef struct SWiT_Exn *SWiT_Exn;


/* Compiler constants */
#define SWIT_MAX_ARGS 0xffff	  /* max number of args to function */
#define MAX_SYMBOL_SIZE 0xffff    /* max length of symbol names */
#define TWO_INT_MASK 0xffff       /* for decoding local variables code */

/*
 * Global scheme symbols, keywords or constants :
 */
extern SWiT swit_null_list;
extern SWiT swit_null_string;
extern SWiT swit_null_vect;
extern SWiT swit_void_env;
extern SWiT swit_true;
extern SWiT swit_false;
extern SWiT swit_eof;
extern SWiT swit_zero;
extern SWiT swit_one;
extern SWiT swit_unbound;
extern SWiT swit_quit;
extern SWiT swit_space;
extern SWiT swit_tab;
extern SWiT swit_escape;
extern SWiT swit_backspace;
extern SWiT swit_newline;
extern SWiT swit_formfeed;
extern SWiT swit_return;
extern SWiT swit_stdin_port;
extern SWiT swit_stdout_port;
extern SWiT swit_stderr_port;
extern SWiT swit_vector_start;
extern SWiT swit_object_start;
extern SWiT swit_prom_or_clos_start;
extern SWiT swit_prom;
extern SWiT swit_clos;
extern SWiT swit_splice_abbrev;
extern SWiT swit_let;
extern SWiT swit_letstar;
extern SWiT swit_letrec;
extern SWiT swit_lambda;
extern SWiT swit_define;
extern SWiT swit_defmacro;
extern SWiT swit_if;
extern SWiT swit_try;
extern SWiT swit_set;
extern SWiT swit_begin;
extern SWiT swit_cond;
extern SWiT swit_case;
extern SWiT swit_quote;
extern SWiT swit_quasiquote;
extern SWiT swit_unquote;
extern SWiT swit_unquote_splicing;
extern SWiT swit_and;
extern SWiT swit_or;
extern SWiT swit_delay;
extern SWiT swit_else;
extern SWiT swit_arrow;
extern SWiT swit_time;
extern SWiT swit_verbose;
extern SWiT swit_exit;

/* error handling */
struct swit_error
{
  SWiT_Expr obj;
  SWiT_Env env;
  char *frmt;
  int n;
};

extern void swit_signal_error (SWiT err, char *msg, ...);
extern char *prim_err_msg;


/* Misc. */
extern SWiT swit_apply (SWiT fun, int nargs, SWiT *vals);
extern SWiT swit_apply_proc (char *name, int nargs, SWiT_Expr *args);
extern SWiT_Expr read_program (int token);
extern SWiT_Expr read_exp (int token, SWiT env);
extern SWiT rd_datum (int token);
extern int loading;
extern int swit_eq (SWiT obj1, SWiT obj2);
extern int swit_eqv (SWiT obj1, SWiT obj2);
extern int swit_equal (SWiT obj1, SWiT obj2);
extern int swit_list_length (SWiT arg);
extern void print_to_buffer (SWiT obj, int esc, int tcl, SWiT fv, int depth);
extern char *swit_string_to_c_string (SWiT s);
extern void str2buff (char *str);
extern void char2buff (char c);
extern void space2buff (void);
extern SWiT make_type(char *type_name);

extern char *sys_errlist[];

/* tables */
extern void *print_table[];
extern void *read_table[];
extern void *equal_table[];

/* debug */
extern void debug(SWiT obj);
extern void debug_code(SWiT_Expr obj);

/* Memory allocation */
#define swit_malloc GC_malloc /*  GC_malloc / *  malloc */
#define swit_malloc_atomic  GC_malloc /* GC_malloc_atomic / *  malloc */
#define swit_realloc GC_realloc /* realloc */
#define swit_free GC_free   /* free */
#define swit_malloc_stubborn GC_malloc /*  GC_malloc_stubborn / * malloc */
#define swit_change_stubborn(x) /*  GC_change_stubborn / * */
#define swit_end_stubborn_change(x) /* GC_end_stubborn_change / * */
#define size_swit sizeof(struct SWiT)
#define size_swit_expr sizeof(struct SWiT_Expr)
#define swit_alloc_object() (SWiT)swit_malloc(size_swit)
/*#define swit_alloc_ptrs(size) (SWiT *)swit_malloc(size_swit*size)*/
#define swit_alloc_ptrs(size) (SWiT *)swit_malloc(4*(size+1))

#define swit_alloc_expr() (SWiT_Expr)swit_malloc(size_swit_expr)
#define swit_alloc_exprs(num_objects) (SWiT_Expr *)swit_malloc(size_swit_expr*num_objects)
/*#define swit_alloc_exprs(num_objects) (SWiT_Expr *)swit_malloc(sizeof(int)*num_objects+1)*/
#define SWIT_ALLOC(structure) \
 ((SWiT_ ## structure)swit_malloc(sizeof(struct SWiT_ ## structure)))

/* Creation of literals */
extern SWiT swit_make_char (int c);
extern SWiT swit_make_pair (SWiT car, SWiT cdr);
extern SWiT swit_make_string (void *str, unsigned int size);
extern SWiT swit_make_constant_string (void *str, unsigned int size);
extern SWiT swit_make_vector (int argc, SWiT *argv);
extern SWiT swit_make_fixnum (int fix);
extern SWiT swit_make_bignum (MP_INT *i);
extern SWiT swit_make_inexact (double d);
extern SWiT swit_make_symbol (char *name, int len, SWiT init);
extern SWiT swit_make_env (SWiT_Env env);

/* Creation of primitives */
extern void swit_make_prim0(char *name, SWiT_Prim0 proc);
extern void swit_make_prim1(char *name, SWiT_Prim1 proc, int arg1);
extern void swit_make_prim01(char *name, SWiT_Prim1 proc,
			     int targ1, SWiT val);
extern void swit_make_prim12(char *name, SWiT_Prim2 proc,
			     int targ1, int targ2,
			     SWiT val);
extern void swit_make_prim2(char *name, SWiT_Prim2 proc,
			    int targ1, int targ2);
extern void swit_make_prim3(char *name, SWiT_Prim3 proc,
			    int targ1, int targ2, int targ3);
extern void swit_make_prim0n(char *name, SWiT_Prim0n proc,
			     int targs, int tlast);
extern void swit_make_prim1n(char *name, SWiT_Prim1n proc,
			     int targ1, int targs, int tlast);
extern void swit_make_prim2n(char *name, SWiT_Prim2n proc,
			     int targ1, int targ2,
			     int targs, int tlast);


/* main loop and initialisation */
extern void swit_repl(void);
extern void swit_init(void);

/*
 *-------------------------------------------------------------
 *
 * Tokens classes used by the lexical analyser :
 *
 *-------------------------------------------------------------
 */

/* literals (boolean, char, string, numbers, identifier, keywords) */
#define	LITERAL	256
/* Delimiter tokens */
#define	SPLICE	257 /* ,@ */
#define	VECTOR  258 /* #( */
#define PRIMITIVE 259 /* #' */
#define SYSTEM 260  /* #: */
#define READ_EVAL 261 /* #. */
#define RD_QUEUE 262 /* #[ */

/* in switRead.c */
extern SWiT tok_val;
extern char *tok_text;
extern int get_tok(void);

/* Ports */
extern SWiT_Port current_ip;
extern SWiT_Port current_op;
extern SWiT_Port current_ep;


/*----------------------------------------------------------------
 *  Types :
 *  An atom is of type T iff : obj.type & type_mask == T
 *  Pairs verify : obj.type % 2 == 0
 *----------------------------------------------------------------
 */
#define TYPE_MASK0    0x000f
#define TYPE_MASK1    0x00ff
#define TYPE_MASK2    0x0fff
#define STR_MASK      0x001f /* strings are at most 2^24 bytes long */
#define CPLX_MASK     0x10ff
#define INT_MASK      0x20ff
#define OBJ           0xffff
#define SUB_TYPE_MASK 0xff00

#define CHAR          0x0003
#define BOOLEAN       0x0005
#define STRING        0x0007
#define CONST_STRING  0x0017
#define VECT          0x0009 /* vectors are at most 2^24 bytes long */
#define QUEUE         0x000b
#define PROCEDURE     0x000d  /* warning: type & mask1 = 0x0d */
#   define PRIM       0x010d    /* Primitive */
#      define PRIM0   0x110d
#      define PRIM01  0x210d
#      define PRIM0N  0x310d
#      define PRIM1   0x410d
#      define PRIM12  0x510d
#      define PRIM1N  0x610d
#      define PRIM2   0x710d
#      define PRIM2N  0x810d
#      define PRIM3   0x910d
#   define CONT       0x020d     /* Continuation */
#   define CLOSURE    0x030d
#define CURRENT_ENV_PRIM  0x1110d
#define CAR_PRIM      0x1410d
#define CDR_PRIM      0x2410d
#define CALL_CC_PRIM  0x3410d
#define RAISE_PRIM    0x1510d
#define VALUES_PRIM   0x1610d
#define CALL_WITH_VALUES_PRIM 0x1710d
#define CERROR_PRIM   0x2710d
#define APPLY_PRIM    0x1810d
#define PROMISE       0x000f
#define PORT          0x0011     /* Input, output, read/write, closed */
#define NUMBER        0x0013     /* FIX, BIG, DBL, RATIO, Z[i], Q[i], R[i] */
#   define EXACT      0x0113
#      define FIXNUM  0x2113
#      define BIGNUM  0x6113
#      define RATIO   0x4113
#      define ZCPLX   0x3113
#      define QCPLX   0x5113
#   define INEXACT    0x0213     /* real, R[i] */
#      define RCPLX   0x1213
#      define REAL    0x2213
#   define COMPLEX    0x1013      /* used with CPLX_MASK */
#   define INT        0x2013      /* used with INT_MASK */
#define SYMBOL        0x0015
#define NIL           0x001b
#define UNBOUND       0x001d
#define MARK          0x001f
#define MACRO         0x0023
#define QUIT          0x0025
#define OBJECT        0x002d
#define SOCKET        0x002f
#define QUAD          0x0031
#define EOF_TYPE      0x0033
#define TURTLE        0x0035
#define ENV           0x004d     /* Environment */

#define SWIT_SOCKETP(obj) (SWIT_TYPE(obj) == SOCKET)
#define SWIT_NULLP(obj) (SWIT_TYPE(obj) == NIL)
#define SWIT_BOOLEANP(obj) (SWIT_TYPE(obj) == BOOLEAN)
#define SWIT_STRINGP(obj) ((SWIT_TYPE(obj) & TYPE_MASK0) == STRING)
#  define SWIT_CONST_STRINGP(obj) ((SWIT_TYPE(obj) & STR_MASK) == CONST_STRING)
#  define SWIT_MUTABLE_STRINGP(obj) ((SWIT_TYPE(obj) & STR_MASK) == STRING)
#define SWIT_QUEUEP(obj) (SWIT_TYPE(obj) == QUEUE)
#define SWIT_QUADP(obj) ((SWIT_TYPE(obj) & OBJ) == QUAD) /* marks */
#define SWIT_TURTLEP(obj) (SWIT_TYPE(obj) == TURTLE)
#define SWIT_OBJECTP(obj) (SWIT_TYPE(obj) == OBJECT)
/* MASK0 for vectors, because of marks (see switVector.c) */
#define SWIT_VECTORP(obj) ((SWIT_TYPE(obj) & TYPE_MASK1) == VECT)
#define SWIT_PROMP(obj) (SWIT_TYPE(obj) == PROMISE)
#define SWIT_SYMBOLP(obj) (SWIT_TYPE(obj) == SYMBOL)
#define SWIT_EOFP(obj) (SWIT_TYPE(obj) == EOF_TYPE)
#define SWIT_UNBOUNDP(obj) (SWIT_TYPE(obj) == UNBOUND)
#define SWIT_CHARP(obj) (SWIT_TYPE(obj) == CHAR)

#define SWIT_PROCP(obj) ((SWIT_TYPE(obj) & TYPE_MASK1) == PROCEDURE)
#  define SWIT_PRIMP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == PRIM)
#    define SWIT_PRIM0P(obj) (SWIT_TYPE(obj) == PRIM0)
#    define SWIT_PRIM01P(obj) (SWIT_TYPE(obj) == PRIM01)
#    define SWIT_PRIM0NP(obj) (SWIT_TYPE(obj) == PRIM0N)
#    define SWIT_PRIM1P(obj) (SWIT_TYPE(obj) == PRIM1)
#    define SWIT_PRIM12P(obj) (SWIT_TYPE(obj) == PRIM12)
#    define SWIT_PRIM1NP(obj) (SWIT_TYPE(obj) == PRIM1N)
#    define SWIT_PRIM2P(obj) (SWIT_TYPE(obj) == PRIM2)
#    define SWIT_PRIM2NP(obj) (SWIT_TYPE(obj) == PRIM2N)
#    define SWIT_PRIM3P(obj) (SWIT_TYPE(obj) == PRIM3)
#  define SWIT_CONTP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == CONT)
#  define SWIT_CLOSUREP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == CLOSURE)

/* #define SWIT_ENVP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == ENV) */
#define SWIT_ENVP(obj) (SWIT_TYPE(obj) == ENV)

#define SWIT_PORTP(obj) (SWIT_TYPE(obj) == PORT)

#define SWIT_NUMBERP(obj) ((SWIT_TYPE(obj) & TYPE_MASK1) == NUMBER)
#  define SWIT_EXACTP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == EXACT)
#  define SWIT_INTP(obj) ((SWIT_TYPE(obj) & INT_MASK) == INT)
#    define SWIT_FIXNUMP(obj) (SWIT_TYPE(obj) == FIXNUM)
#    define SWIT_BIGNUMP(obj) (SWIT_TYPE(obj) == BIGNUM)
#    define SWIT_RATIOP(obj) (SWIT_TYPE(obj) == RATIO)
#    define SWIT_ZCPLXP(obj) (SWIT_TYPE(obj) == ZCPLX)
#    define SWIT_QCPLXP(obj) (SWIT_TYPE(obj) == QCPLX)
#  define SWIT_INEXACTP(obj) ((SWIT_TYPE(obj) & TYPE_MASK2) == INEXACT)
#    define SWIT_REALP(obj) (SWIT_TYPE(obj) == REAL)
#    define SWIT_RCPLXP(obj) (SWIT_TYPE(obj) == RCPLX)
#  define SWIT_COMPLEXP(obj) ((SWIT_TYPE(obj) & CPLX_MASK) == COMPLEX)

#define SWIT_MARKP(obj) (SWIT_TYPE(obj) == MARK)
#define SWIT_MACROP(obj) (SWIT_SYMBOLP(obj) && SYMBOL_MACROP(SWIT_SYM(obj)))


#define ANY            0x0
#define PAIR           0x02
#define LIST           0x04
#define QUAD_M (QUAD | (~TYPE_MASK0 << 16))
#define VECT_M (VECT | (~TYPE_MASK0 << 16))
#define STRING_M (STRING | (~TYPE_MASK0 << 16))
#define PROC_M (PROCEDURE | (~TYPE_MASK1 << 16))
#define NUM_M (NUMBER | (~TYPE_MASK1 << 16))

/* Exceptions */
extern SWiT exn;
extern SWiT exn_val;
extern SWiT exn_type;
extern SWiT exn_range;
extern SWiT exn_arity;
extern SWiT exn_unbound;
extern SWiT exn_math;
extern SWiT exn_io;
extern SWiT exn_read;
extern SWiT exn_unknown;
extern SWiT exn_cont;
extern SWiT exn_env;
extern SWiT exn_size;

/* Declaration of error messages (defined in messages_<language>.h) */
/* generic */
extern char *unknown_error;
/* init */
extern char *banner_msg;
extern char *browse_msg;
extern char *implementation_date;
extern char *load_init;
extern char *lib_path_error;
extern char *loading_file;
extern char *skip_file;
extern char *file_not_exists;
extern char *init_done;
extern char *no_tk;
extern char *load_err;
extern char *file_not_found;
/* top level */
extern char *temps;
extern char *err_prompt;
extern char *cerr_prompt;
extern char *error_at_line;
extern char *exception_newline;
extern char *exception;
extern char *uncaught_exception;
extern char *in;
extern char *arith_error;
extern char *stack_overflow;
extern char *bus_error;
extern char *seg_fault;
extern char *interrupt_received;
extern char *ret_to_top;
extern char *comp_impl_err;
extern char *type_impl_err;
/* eval */
extern char *wrong_type;
extern char *wrong_arity;
extern char *must_be_thunk;
extern char *not_a_thunk;
extern char *prim_redef;
extern char *call_cc_wrong_arg;
extern char *improper_list;
extern char *not_a_proc;
/* character */
extern char *out_of_range;
/* pair */
extern char *not_a_pair;
extern char *proper_list;
extern char *proper_list1;
extern char *list_of_pairs;
extern char *index_too_large;
extern char *not_an_alist;
/* symbol */
extern char *wrong_number;
extern char *iskeyword;
extern char *cannot_set;
extern char *isunbound;
extern char *cannot_unbound;
extern char *sig_not_defined;
/* vector */
extern char *finite_list;
extern char *bsearch_error;
extern char *bsearch_error2;
/* string */
extern char *out_of_bound1;
extern char *out_of_bound2;
extern char *index_out_of_range;
extern char *must_be_strings;
extern char *impl_err;
extern char *length_must_be_even;
/* number */
extern char *too_big;
extern char *wrong_radix;
/* control */
extern char *type_error_msg;
extern char *apply_wrong_number;
extern char *same_length;
extern char *last_proper;
/* read */
extern char *wrong_shell_command;
extern char *unexpected_token;
extern char *bad_bool;
extern char *unexpected_numbersign;
extern char *unexpected_space;
extern char *num_too_long;
extern char *number_too_long;
extern char *bad_exact_num;
extern char *bad_num;
extern char *bad_octal;
extern char *bad_char;
extern char *bad_char_name;
extern char *bad_literal;
extern char *string_too_long;
extern char *int_expected;
extern char *closed_port;
extern char *cannot_peek;
extern char *wrong_prim_name;
extern char *already_reserved;
/* io */
extern char *cannot_close;
extern char *wrong_port;
extern char *cannot_seek;
extern char *closed;
extern char *no_port;
extern char *must_be_proc;
extern char *wrong_address;
extern char *cannot_output;
extern char *filename_too_long;
extern char *bye;
extern char *permission_denied;
extern char *already_closed;
extern char *err_inside_string;
extern char *err_inside_expr;
extern char *err_at_line;
/* socket */
extern char *already_reserved;
/* print */
extern char *out_err;
/* queue */
extern char *empty_queue;
extern char *negative_index;
/* quad */
extern char *too_many_args;
extern char *wrong_index;
extern char *pt_infty;
extern char *must_be_vect3;
extern char *must_be_bezier;
extern char *stack_empty;
extern char *null_vect3;
extern char *null_quad;
/* misc */
extern char *no_parent;
extern char *too_big_or_neg;
extern char *err_mode;
extern char *must_be_proc3;
/* code */
extern char *unknown_code;
extern char *prim_call_err;
extern char *wrong_code;
/* quads and model */
extern char *wrong_quad_coord;
extern char *non_zero;
/* tk */
extern char *swit_call_wrong_number;
extern char *swit_val_wrong_number;
extern char *swit_expr_wrong_number;
extern char *swit_call_isnot_fun;
extern char *swit_expr_no_env;
extern char *swit_val_not_bound;
extern char *tk_in;
/* disassemble */
extern char *dis_impl_err;
