/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *                 6.6 CHARACTERS
 *==================================================================
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 * char?                 OBJ         -> BOOLEAN
 * char=?                CHAR X CHAR -> BOOLEAN
 * char<=?               CHAR X CHAR -> BOOLEAN
 * char<?                CHAR X CHAR -> BOOLEAN
 * char>?                CHAR X CHAR -> BOOLEAN
 * char>=?               CHAR X CHAR -> BOOLEAN
 * char-ci=?             CHAR X CHAR -> BOOLEAN
 * char-ci<?             CHAR X CHAR -> BOOLEAN
 * char-ci>?             CHAR X CHAR -> BOOLEAN
 * char-ci<=?            CHAR X CHAR -> BOOLEAN
 * char-ci>=?            CHAR X CHAR -> BOOLEAN
 * char-alphabetic?      CHAR        -> BOOLEAN
 * char-numeric?         CHAR        -> BOOLEAN
 * char-whitespace?      CHAR        -> BOOLEAN
 * char-upper-case?      CHAR        -> BOOLEAN
 * char-lower-case?      CHAR        -> BOOLEAN
 * char->integer         CHAR        -> INTEGER
 * integer->char         INTEGER     -> CHAR
 * char-upcase           CHAR        -> CHAR
 * char-downcase         CHAR        -> CHAR
 *==================================================================
 */

#include "swit.h"

/* locals (prototypes) */

static SWiT char_p_prim (SWiT arg);
static SWiT char_eq_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_lt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_gt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_le_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ge_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ci_eq_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ci_lt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ci_gt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ci_le_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT char_ci_ge_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
/* eported to switEquiv.c */
int char_eq (SWiT arg1, SWiT arg2);
static int char_lt (SWiT arg1, SWiT arg2);
static int char_gt (SWiT arg1, SWiT arg2);
static int char_le (SWiT arg1, SWiT arg2);
static int char_ge (SWiT arg1, SWiT arg2);
static int char_ci_eq (SWiT arg1, SWiT arg2);
static int char_ci_lt (SWiT arg1, SWiT arg2);
static int char_ci_gt (SWiT arg1, SWiT arg2);
static int char_ci_le (SWiT arg1, SWiT arg2);
static int char_ci_ge (SWiT arg1, SWiT arg2);
static SWiT char_alphabetic_prim (SWiT arg);
static SWiT char_numeric_prim (SWiT arg);
static SWiT char_whitespace_prim (SWiT arg);
static SWiT char_upper_case_prim (SWiT arg);
static SWiT char_lower_case_prim (SWiT arg);
static SWiT char_to_integer_prim (SWiT arg);
static SWiT integer_to_char_prim (SWiT arg);
static SWiT char_upcase_prim (SWiT arg);
static SWiT char_downcase_prim (SWiT arg);

static SWiT char_table[256];

/* globals */

/* named characters */
SWiT swit_space;
SWiT swit_tab;
SWiT swit_escape;
SWiT swit_backspace;
SWiT swit_newline;
SWiT swit_formfeed;
SWiT swit_return;


/*====================================
 *           PRINTING
 *======================================
 */

static void
print_char (int ch, int esc)
{
  if (esc) {
    switch (ch) {
    case ' ':
      str2buff("#\\space");
      break;
    case '\n':
      str2buff("#\\newline");
      break;
    case '\t':
      str2buff("#\\tab");
      break;
    case '\e':
      str2buff("#\\escape");
      break;
    case '\f':
      str2buff("#\\formfeed");
      break;
    case '\r':
      str2buff("#\\return");
      break;
    case '\b':
      str2buff("#\\backspace");
      break;
    default:
      if (ch < 32) {
	switch (ch)
	  {
	  case '\n': str2buff("#\\n");break;
	  case '\t': str2buff("#\\t");break;
	  case '\f': str2buff("#\\f");break;
	  case '\r': str2buff("#\\r");break;
	  case '\b': str2buff("#\\b");break;
	  case '\e': str2buff("#\\e");break;
	  default:
	    {
	      char cstr[4];
	      str2buff("#\\");
	      cstr[0] = '0';
	      cstr[1] = 48 + ch / 8;
	      cstr[2] = 48 + ch % 8;
	      cstr[3] = '\0';
	      str2buff(cstr);
	    }
	  }
      } else if (ch < 256) {
	str2buff("#\\");
	char2buff((char)ch);
      } else {
	str2buff("#\\");
	char2buff((char)(ch / 256));
	char2buff((char)(ch % 256));
      }
    }
  } else {
    if (ch < 256) {
      char2buff((char)ch);
    } else {
      char2buff((char)(ch / 256));
      char2buff((char)(ch % 256));
    }
  }
}

static void
print_character (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  print_char(SWIT_CHAR(obj), esc);
}


static SWiT
create_char (int c)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = CHAR;
  SWIT_CHAR(obj) = c;
  return (obj);
}

/*
 * Small characters are made once and for all :
 * ('eq?' can be used to compare such characters
 *  and is faster than char=?)
 */
SWiT
swit_make_char (int c)
{
  if ((c < 256) && (c >= 0)){
    return char_table[c];
  } else {
    return create_char(c);
  }
}


static unsigned int
my_toupper (unsigned int ch)
{
  if ((ch < 256) && (ch >= 0)) {
    return toupper(ch);
  } else {
    return ch;
  }
}

/*==================================================================
 *                       PRIMITIVES
 *==================================================================
 */

static SWiT
char_p_prim (SWiT arg)
{
  return (SWIT_CHARP(arg) ? swit_true : swit_false);
}

#define GEN_NARY_COMP(name) \
static SWiT  \
name ## _prim(SWiT arg1, SWiT arg2, int argc, SWiT *argv) \
{ \
   int n = name(arg1, arg2);\
   if (n == 256) goto err;\
   if (! n) { \
       return (swit_false); \
   } else if (argc) { \
      SWiT x = arg2;\
      int i,n; \
      for (i=0;i<argc;i++) { \
         SWiT y = x;\
         x = argv[i];\
         n = name(y, x);\
         if (n == 256) goto err;\
         if (! n) { \
             return (swit_false); \
         } \
      } \
   } \
   return (swit_true); \
 err: \
   prim_err_msg = "Wrong character";\
   exn = exn_range;\
   return NULL;\
}

#define GEN_CHAR_COMP(name, op) \
static int  \
name (SWiT arg1, SWiT arg2) \
{ \
  return (SWIT_CHAR(arg1) op SWIT_CHAR(arg2));\
}

GEN_CHAR_COMP(char_lt, <)
GEN_CHAR_COMP(char_gt, >)
GEN_CHAR_COMP(char_le, <=)
GEN_CHAR_COMP(char_ge, >=)
GEN_NARY_COMP(char_eq)
GEN_NARY_COMP(char_lt)
GEN_NARY_COMP(char_gt)
GEN_NARY_COMP(char_le)
GEN_NARY_COMP(char_ge)

/* exported to switEquiv.c */
int
char_eq (SWiT arg1, SWiT arg2)
{
  return (SWIT_CHAR(arg1) == SWIT_CHAR(arg2));
}


#define GEN_CHAR_CI_COMP(name, op) \
static int  \
name (SWiT arg1, SWiT arg2) \
{ \
    return (my_toupper(SWIT_CHAR(arg1)) op my_toupper(SWIT_CHAR(arg2))); \
}

GEN_CHAR_CI_COMP(char_ci_eq, ==)
GEN_CHAR_CI_COMP(char_ci_lt, <)
GEN_CHAR_CI_COMP(char_ci_gt, >)
GEN_CHAR_CI_COMP(char_ci_le, <=)
GEN_CHAR_CI_COMP(char_ci_ge, >=)


GEN_NARY_COMP(char_ci_eq)
GEN_NARY_COMP(char_ci_lt)
GEN_NARY_COMP(char_ci_gt)
GEN_NARY_COMP(char_ci_le)
GEN_NARY_COMP(char_ci_ge)

static SWiT
char_alphabetic_prim (SWiT arg)
{
  return (isalpha(SWIT_CHAR(arg)) ? swit_true : swit_false);
}

static SWiT
char_numeric_prim (SWiT arg)
{
  return (isdigit(SWIT_CHAR(arg)) ? swit_true : swit_false);
}

static SWiT
char_whitespace_prim (SWiT arg)
{
  return (isspace(SWIT_CHAR(arg)) ? swit_true : swit_false);
}

static SWiT
char_upper_case_prim (SWiT arg)
{
  return (isupper(SWIT_CHAR(arg)) ? swit_true : swit_false);
}

static SWiT
char_lower_case_prim (SWiT arg)
{
  return (islower(SWIT_CHAR(arg)) ? swit_true : swit_false);
}

extern SWiT int_to_integer(int n);

static SWiT
char_to_integer_prim (SWiT arg)
{
  return int_to_integer(SWIT_CHAR (arg));
}

static SWiT
integer_to_char_prim (SWiT arg)
{
  if SWIT_FIXNUMP(arg) {
    int c = SWIT_FIXNUM(arg);
    if (c < 0) goto err;
    return swit_make_char(SWIT_FIXNUM(arg));
  } else if SWIT_BIGNUMP(arg) {
    MP_INT *big = SWIT_BIGNUM(arg);
    if (mpz_size(big) > 1) {
      goto err;
    } else {
      int c = mpz_get_si(big);
      if ((c < 0) || (c > 65535)) goto err;
      return swit_make_char(c);
    }
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
err:
  prim_err_msg = out_of_range;
  exn = exn_range;
  return NULL;
}

static SWiT
char_upcase_prim (SWiT arg)
{
  int n = SWIT_CHAR(arg);
  if (n > 255) {
    return arg;
  } else {
    return swit_make_char(toupper(n));
  }
}

static SWiT
char_downcase_prim (SWiT arg)
{
  unsigned int n = SWIT_CHAR(arg);
  if (n > 255) {
    return arg;
  } else {
    return swit_make_char(tolower(n));
  }
}


/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_char (void)
{
  int i;

  print_table[CHAR >> 1] = print_character;

  for (i = 0; i < 256; i++) {
    char_table[i] = create_char(i);
  }

  swit_newline = swit_make_char('\n');
  swit_formfeed = swit_make_char('\f');
  swit_return = swit_make_char('\r');
  swit_tab = swit_make_char('\t');
  swit_escape = swit_make_char('\e');
  swit_backspace = swit_make_char('\b');
  swit_space = swit_make_char(' ');

  swit_make_prim1 ("char?", char_p_prim, ANY);
  swit_make_prim2n ("char=?", char_eq_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char<?", char_lt_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char>?", char_gt_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char<=?", char_le_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char>=?", char_ge_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char-ci=?", char_ci_eq_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char-ci<?", char_ci_lt_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char-ci>?", char_ci_gt_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char-ci<=?", char_ci_le_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim2n ("char-ci>=?", char_ci_ge_prim, 
		    CHAR, CHAR, CHAR, CHAR);
  swit_make_prim1 ("char-alphabetic?", char_alphabetic_prim,
		   CHAR);
  swit_make_prim1 ("char-numeric?", char_numeric_prim,
		   CHAR);
  swit_make_prim1 ("char-whitespace?", char_whitespace_prim,
		   CHAR);
  swit_make_prim1 ("char-upper-case?", char_upper_case_prim,
		   CHAR);
  swit_make_prim1 ("char-lower-case?", char_lower_case_prim,
		   CHAR);
  swit_make_prim1 ("char->integer", char_to_integer_prim,
		   CHAR);
  swit_make_prim1 ("integer->char", integer_to_char_prim,
		   NUM_M);
  swit_make_prim1 ("char-upcase", char_upcase_prim,
		   CHAR);
  swit_make_prim1 ("char-downcase", char_downcase_prim,
		   CHAR);
}
