/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*----------------------------------------------------------------
 *                       6.9 CONTROL FEATURES
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *    procedure?        'a -> BOOLEAN
 *    apply             PROCEDURE X 'a ... X LIST -> 'b
 *    map               PROCEDURE X LIST X LIST ... -> LIST
 *    for-each          PROCEDURE X LIST X LIST ... -> BOOLEAN
 *    force             'a -> 'b
 *    call-with-current-continuation  PROCEDURE -> CONTINUATION
 *
 * Extensions :
 * ^^^^^^^^^^
 *    promise?            'a ->  BOOLEAN
 *    closure?            'a ->  BOOLEAN
 *    primitive?          'a ->  BOOLEAN
 *    map*               PROCEDURE X LIST X LIST ... -> LIST
 *    append-map         PROCEDURE X LIST X LIST ... -> LIST
 *    append-map*        PROCEDURE X LIST X LIST ... -> LIST
 *    append-map!        PROCEDURE X LIST X LIST ... -> LIST
 *    append-map*!       PROCEDURE X LIST X LIST ... -> LIST
 *----------------------------------------------------------------
 */

#include "swit.h"

extern void print_lambda();
extern void mem2buff();
extern SWiT swit_class;
extern void swit_print_code (SWiT_Expr obj, SWiT fv);
extern void print_code_to_buffer (SWiT_Expr obj, SWiT fv, int depth);
extern int search (SWiT sym, SWiT names);
extern SWiT swit_symbol_value (SWiT sym, SWiT env);
extern void print_sym();
extern void print_str();
extern SWiT make_fv(SWiT_Env e);
extern SWiT swit_append2 (SWiT lst1, SWiT lst2);

static SWiT procedure_p_prim (SWiT val);
static SWiT force_prim (SWiT arg);
static SWiT promise_p_prim (SWiT val);
static SWiT primitive_p_prim (SWiT val);
static SWiT closure_p_prim (SWiT val);
static SWiT continuation_p_prim (SWiT val);


/*=====================================================================
 *    Print procedures (closure, contituation, primitive)
 *=====================================================================
 */
static void
print_closure (SWiT_Closure clos, int tcl, SWiT fv, int depth)
{
  if (tcl) {
    str2buff("{closure}"); /* todo */
  } else {
    depth++;
    str2buff("#<closure ");
    /* should print closure env here (todo) */
    fv = swit_append2(make_fv(CLOSURE_ENV(clos)), fv); /* correct ??*/
    print_lambda(CLOSURE_LBD(clos), fv, depth);
    str2buff(">");
  }
}


/* exported to disassemble.c */
void
prim2buff (SWiT name)
{
  mem2buff(STRING_ELS(name), STRING_LENGTH(name));
}

static void
print_procedure (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  switch (SWIT_TYPE(obj) & OBJ) 
    {
    case PRIM0:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM0(obj)));
      break;
    case PRIM0N:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM0n(obj)));
      break;
    case PRIM01:
    case PRIM1:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM1(obj)));
      break;
    case PRIM12:
    case PRIM2:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM2(obj)));
      break;
    case PRIM1N:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM1n(obj)));
      break;
    case PRIM2N:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM2n(obj)));
      break;
    case PRIM3:
      str2buff("#'");
      prim2buff(PRIM_NAME(SWIT_PRIM3(obj)));
      break;
    case CONT:
      str2buff("#<continuation>"); /* external representation yet to be found */
      break;
    case CLOSURE:
      print_closure(SWIT_CLOSURE(obj), tcl, fv, depth);
      break;
    }
}

/*
 * WARNING:
 *  The bluid-in primitive #'apply is not defined here 
 *  (but hard-coded in eval, see switEval.c)
*/
/* exported (Io, Misc, Tk, Vector etc.) */
extern SWiT_Expr make_call_code (SWiT_Expr fun, int len, SWiT_Expr *args);
extern SWiT_Ccont done_cont;
SWiT
swit_apply (SWiT fun, int nargs, SWiT *vals)
{
  SWiT_Expr expr; 
  SWiT_Expr qvals[nargs];
  int i;
  for (i = 0; i < nargs; i++) {
    qvals[i] = swit_make_quote(vals[i]);
  }
  expr = make_call_code(swit_make_quote(fun), nargs, qvals);
  return swit_eval(expr, NULL, done_cont);
}


/*
 * The following procedure is generally used to call a scheme procedure
 * (e.g. the compiler, the disassembler that are written in swit)
 * returns NULL on error.
 */
SWiT
swit_apply_proc (char *name, int nargs, SWiT_Expr *args)
{
  SWiT_Symbol s = SWIT_SYM(swit_make_symbol(name, strlen(name), swit_unbound));
  SWiT proc = SYMBOL_VAL(s);
  if (!SWIT_PROCP(proc)) {
    /* internal message (for debugging) : should not occur */
    fprintf(stderr, "swit_apply_proc `%s': not a procedure", name);
    return NULL;
  } else {
    SWiT *vals = swit_alloc_ptrs(nargs);
    SWiT result;
    register int i;
    for (i = nargs-1; i> -1; i--) {
      vals[i] = swit_eval(args[i], NULL, done_cont);
    }
    if ((result = swit_apply(proc, nargs, vals))) {
      return result;
    }
    /* internal message (for debugging) : should not occur */
    fprintf(stderr, "swit_apply_proc `%s': unknown error", name);
    return NULL;
  }
}


static SWiT
procedure_p_prim (SWiT val)
{
  return (SWIT_PROCP(val) ? swit_true : swit_false);
}


/* 
 * test (from Jaffer's test.scm) :
 *  (letrec ((p (delay (if c 3 (begin (set! c #t) (+ (force p) 1)))))
 *           (c #f))
 *     (force p)) => 3 (returning 4 is not correct)
 */
SWiT
force_prim (SWiT arg)
{
  if SWIT_PROMP(arg) {
    SWiT_Promise promise = SWIT_PROMISE (arg);
    SWiT res, val = PROMISE_VAL(promise);
    int typ;
    if (!val) {
      val = swit_eval(PROMISE_EXPR(promise), PROMISE_ENV(promise), done_cont);
      if (!PROMISE_VAL(promise)) {
	PROMISE_VAL(promise) = val;
      }
    }
    res = PROMISE_VAL(promise);
    typ = SWIT_TYPE(res);
    if ((typ == BOOLEAN)
          || (typ == CHAR) 
	  || (res == swit_null_string) 
	  || (res == swit_null_list)) {
      return res;
    } else {
      SWIT_TYPE(arg) = typ;
      arg->data = res->data;
    }
  }
  return(arg);
}


/*===========================================
 *     PRINTING (promise)
 *===========================================
 */

static void
print_promise (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  SWiT_Promise prom = SWIT_PROMISE(obj);

  str2buff("#<promise ");
  if (PROMISE_VAL(prom)) {
    /* may occur (?) only while error handling */
    print_to_buffer(PROMISE_VAL(prom), 1, 0, fv, depth);
  } else {
    int i, env_flag = 0;
    depth++;
    /* should print promise env here (todo) */
    fv = swit_append2(make_fv(PROMISE_ENV(prom)), fv);
    print_code_to_buffer(PROMISE_EXPR(prom), fv, depth);
    if (env_flag) {
      char str[env_flag+1];
      for (i = 0; i < env_flag; i++) {
	str[i] = ')';
      }
      str[env_flag]='\0';
      str2buff(str);
    }
  }
  str2buff(">");
}


static SWiT
promise_p_prim (SWiT val)
{
  return (SWIT_PROMP(val) ? swit_true : swit_false);
}

static SWiT
primitive_p_prim (SWiT val)
{
  return (SWIT_PRIMP(val) ? swit_true : swit_false);
}

static SWiT
closure_p_prim (SWiT val)
{
  return (SWIT_CLOSUREP(val) ? swit_true : swit_false);
}

static SWiT
continuation_p_prim (SWiT val)
{
  return (SWIT_CONTP(val) ? swit_true : swit_false);
}

void
swit_init_control (void)
{
  print_table[PROCEDURE >> 1] = print_procedure;
  print_table[PROMISE >> 1] = print_promise;

  swit_make_prim1 ("procedure?", procedure_p_prim, ANY);
  swit_make_prim2n ("apply", NULL, PROC_M, ANY, ANY, LIST);  /* in eval */
  swit_make_prim1 ("force", force_prim, ANY);
  swit_make_prim2 ("call-with-values", NULL, 0, 0); /* in eval */
  swit_make_prim1n ("values", NULL, 0, 0, 0); /* in eval */
  swit_make_prim1 ("call/cc", NULL, 0); /* in eval */
  swit_make_prim1 ("call-with-current-continuation", NULL, 0);  /* alias */
  swit_make_prim1 ("promise?", promise_p_prim, ANY);
  swit_make_prim1 ("primitive?", primitive_p_prim, ANY);
  swit_make_prim1 ("closure?", closure_p_prim, ANY);
  swit_make_prim1 ("continuation?", continuation_p_prim, ANY);
}
