/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*----------------------------------------------------
 *         6.2 EQUIVALENCE PREDICATES
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *   eqv?        'a X 'b -> BOOLEAN
 *   eq?         'a X 'b -> BOOLEAN
 *   equal?      'a X 'b -> BOOLEAN
 *
 * Extensions :
 * ^^^^^^^^^^
 *   graph-equal?   'a X 'b -> BOOLEAN
 *
 *  BUG : if a 'Ctrl-C' or an error is issued during
 *        graph-equal? the argument will not be
 *        restored ('peeled').
 *-----------------------------------------------------
 */

#include "swit.h"

/* defined in Pair */
extern void mark (SWiT expr);
extern SWiT peel (SWiT oldcel);
extern SWiT EX;
extern SWiT M00;
extern SWiT M01;
extern SWiT M10;
extern SWiT M11;

extern int char_eq (SWiT ch1, SWiT ch2);

/* static int vector_equal (SWiT vec1, SWiT vec2); */
static int tree_equal (SWiT lst1, SWiT lst2);
static int list_equal (SWiT lst1, SWiT lst2);
static SWiT equal_prim (SWiT arg1, SWiT arg2);
static SWiT graph_equal_prim (SWiT arg1, SWiT arg2);
static SWiT eq_prim (SWiT arg1, SWiT arg2);
static SWiT eqv_prim (SWiT arg1, SWiT arg2);

/* exported in Pair (for memq, etc.) */
int swit_eq (SWiT obj1, SWiT obj2);
int swit_eqv (SWiT obj1, SWiT obj2);  /* exported also in Eval (for case)*/
int swit_equal (SWiT obj1, SWiT obj2);
int swit_graph_equal (SWiT obj1, SWiT obj2);

typedef int(* Equal_Fun)(SWiT, SWiT);
/* to be called as : ((Equal_Fun)(equal_table[TYPE >> 1]))(obj1, obj2) */


static SWiT
eq_prim (SWiT arg1, SWiT arg2)
{
  return (arg1 == arg2 ? swit_true : swit_false);
}

static SWiT
eqv_prim (SWiT arg1, SWiT arg2)
{
  return (swit_eqv(arg1, arg2) ? swit_true : swit_false);
}


static SWiT
equal_prim (SWiT arg1, SWiT arg2)
{
  return (swit_equal(arg1, arg2) ? swit_true : swit_false);
}

static SWiT
graph_equal_prim (SWiT arg1, SWiT arg2)
{
  return (swit_graph_equal(arg1, arg2) ? swit_true : swit_false);
}

int
swit_eq (SWiT obj1, SWiT obj2)
{
  return ((obj1 == obj2) ? 1 : 0);
}

/* imported from switNumber.c */
extern int bin_eq (SWiT n1, SWiT n2);

int
swit_eqv (SWiT obj1, SWiT obj2)
{
  if (obj1 == obj2) {
    return 1;
  } else if (SWIT_NUMBERP(obj1) && SWIT_NUMBERP(obj2)) {
    return (bin_eq(obj1, obj2));
  } else if (SWIT_CHARP(obj1) && SWIT_CHARP(obj2)) { /* 16 bit chars */ 
    return (char_eq(obj1, obj2));
  } else if (SWIT_CLOSUREP(obj1) && SWIT_CLOSUREP(obj2)) {
    SWiT_Closure clos1 = SWIT_CLOSURE(obj1);
    SWiT_Closure clos2 = SWIT_CLOSURE(obj2);
    if (CLOSURE_ENV(clos1) != CLOSURE_ENV(clos2)) {
      return 0;
    } else {
      SWiT code1 = SWIT_CDR(LAMBDA_PCODE(CLOSURE_LBD(clos1)));
      SWiT code2 = SWIT_CDR(LAMBDA_PCODE(CLOSURE_LBD(clos2)));
      /* same arity */
      if (SWIT_FIXNUM(SWIT_CAR(SWIT_CAR(code1))) != SWIT_FIXNUM(SWIT_CAR(SWIT_CAR(code2)))) return 0;
      /* same body */
      return (swit_equal(SWIT_CDR(code1), SWIT_CDR(code2)));
    }
  } else {
    return 0;
  }
}

int
swit_equal (SWiT obj1, SWiT obj2)
{
  if (swit_eqv(obj1, obj2)) {
    return 1;
  } else if SWIT_PAIRP(obj1) {
    if SWIT_PAIRP(obj2) {
      return tree_equal(obj1, obj2);
    } else {
      return 0;
    }
  } else {
    int t1 = SWIT_TYPE(obj1) & TYPE_MASK0;
    int t2 = SWIT_TYPE(obj2) & TYPE_MASK0;
    int t0;
    if (t1 == t2) {
      switch (t1)
	{
	  /*	case VECT: return vector_equal(obj1, obj2);*/
	case STRING:
	  {
	    int len1 = STRING_LENGTH(obj1);
	    int len2 = STRING_LENGTH(obj2);
	    return ((len1 == len2)
		    && (memcmp(STRING_ELS(obj1),
			       STRING_ELS(obj2), len1) == 0));
	  }
	default:
	  t0 = SWIT_TYPE(obj1) & TYPE_MASK1;
	  if (t0 != (SWIT_TYPE(obj2) & TYPE_MASK1)) {
	    return 0;
	  } else {
	    return ((Equal_Fun)(equal_table[t1 >> 1]))(obj1, obj2);
	  }
	}
    } else {
      return 0;
    }
  }
}

static int
tree_equal (SWiT lst1, SWiT lst2)
{
  while SWIT_PAIRP(lst1) {
    if SWIT_PAIRP(lst2) {
      if (swit_equal (SWIT_CAR (lst1), SWIT_CAR (lst2))) {
	lst1 = SWIT_CDR (lst1);
	lst2 = SWIT_CDR (lst2);
      } else {
	return 0;
      }
    } else {
      return 0;
    }
  }
  return (SWIT_PAIRP(lst2) ? 0 : swit_equal(lst1, lst2));
}

int
swit_graph_equal (SWiT obj1, SWiT obj2)
{
  if (swit_eqv(obj1, obj2)) {
    return 1;
  } else if SWIT_PAIRP(obj1) {
    if SWIT_PAIRP(obj2) {
      return list_equal(obj1, obj2);
    } else {
      return 0;
    }
    /*  } else if SWIT_VECTORP(obj1) {
	return vector_equal(obj1, obj2); */
  } else if SWIT_STRINGP(obj1) {
    int len1 = STRING_LENGTH(obj1);
    int len2 = STRING_LENGTH(obj2);
    return ((len1 == len2)
	    && (memcmp(STRING_ELS(obj1), STRING_ELS(obj2), len1) == 0));
  } else {
    return ((Equal_Fun)(equal_table[SWIT_TYPE(obj1) >> 1]))(obj1, obj2);
  }
}

#define M(cell) (SWIT_CAR(cell))
#define A(cell) (SWIT_CAR(SWIT_CDR(cell)))
#define D(cell) (SWIT_CDR(SWIT_CDR(cell)))

static int
equal_graph (SWiT expr1, SWiT expr2)
{
  if (SWIT_PAIRP(expr1)) {
    if (SWIT_PAIRP(expr2)) {
      SWiT m1 = M(expr1);
      SWiT m2 = M(expr2);
      if (m1 == m2) {
	if (m1 == M00) {
	  return(equal_graph(A(expr1), A(expr2))
		 && equal_graph(D(expr1), D(expr2)));
	} else if (m1 == M10) {
	  SWiT a1 = A(expr1);
	  SWiT a2 = A(expr2);
	  if (SWIT_PAIRP(a1) && SWIT_PAIRP(a2)) {
	    return(1);
	  } else {
	    return (swit_equal(a1, a2));
	  }
	  equal_graph(D(expr1), D(expr2));
	} else if (m1 == M01) {
	  SWiT d1 = D(expr1);
	  SWiT d2 = D(expr2);
	  equal_graph(A(expr1), A(expr2));
	  if (SWIT_PAIRP(d1) && SWIT_PAIRP(d2)) {
	    return(1);
	  } else {
	    return (swit_equal(d1, d2));
	  }
	} else { /* M11 */
	  SWiT a1 = A(expr1);
	  SWiT a2 = A(expr2);
	  SWiT d1 = D(expr1);
	  SWiT d2 = D(expr2);
	  if (SWIT_PAIRP(a1) && SWIT_PAIRP(a2)) {
	    return(1);
	  } else {
	    return (swit_equal(a1, a2));
	  }
	  if (SWIT_PAIRP(d1) && SWIT_PAIRP(d2)) {
	    return(1);
	  } else {
	    return (swit_equal(d1, d2));
	  }
	}
      } else {
	return(0);
      }
    } else {
      return(0);
    }
  } else if (SWIT_PAIRP(expr2)) {
    return(0);
  } else {
    return (swit_eqv(expr1, expr2));
  }
}

static int
list_equal (SWiT expr1, SWiT expr2)
{
  int result;
  mark(expr1);
  mark(expr2);
  result = equal_graph(expr1, expr2);
  peel(expr1);
  peel(expr2);
  return (result);
}


/*=====================================================
 *                 INITIALISATION
 *=====================================================
 */
void
swit_init_equiv (void)
{
  swit_make_prim2 ("eqv?", eqv_prim,
		   ANY, ANY);
  swit_make_prim2 ("eq?", eq_prim,
		   ANY, ANY);
  swit_make_prim2 ("equal?", equal_prim,
		   ANY, ANY);
  swit_make_prim2 ("graph-equal?", graph_equal_prim,
		   ANY, ANY);
}
