/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/


/*-----------------------------------------------------------
 *                          6.10.1  PORTS
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *    call-with-input-file     STRING X PROCEDURE -> <OBJ>
 *    call-with-output-file    STRING X PROCEDURE -> <OBJ>
 *    input-port?              <OBJ> -> BOOLEAN
 *    output-port?             <OBJ> -> BOOLEAN
 *    current-input-port       INPUT-PORT
 *    current-output-port      OUTPUT-PORT
 *    with-input-from-file     STRING X PROCEDURE -> <OBJ>
 *    with-output-to-file      STRING X PROCEDURE -> <OBJ>
 *    open-input-file          STRING -> INPUT-PORT
 *    open-output-file         STRING -> OUTPUT-PORT
 *    close-input-port         INPUT-PORT -> BOOLEAN
 *    close-output-port        OUTPUT-PORT -> BOOLEAN
 *
 * Extensions :
 * ^^^^^^^^^^
 *    call-with-input-string   STRING X PROCEDURE -> <OBJ>
 *    with-input-from-string   STRING X PROCEDURE -> <OBJ>
 *    string->port             STRING -> PORT
 *    port-string              PORT -> STRING
 *    port?                    <OBJ> -> BOOLEAN
 *    string-port?             <OBJ> -> BOOLEAN
 *
 *
 *                     6.10.4 FILE SYSTEM INTERFACE
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *    load                     STRING -> BOOLEAN
 *    transcript-on            STRING -> BOOLEAN
 *    transcript-off           void -> BOOLEAN
 *
 * Extensions :
 * ^^^^^^^^^^
 *    file-exists?             STRING -> BOOLEAN
 *    system                   STRING -> BOOLEAN
 *    quit                     -> BOOLEAN
 *    getenv                   STRING -> STRING
 *    date                     -> STRING
 *-----------------------------------------------------------
 */

#include "swit.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

extern int issocket(int des);

#define current_ip SWIT_PORT(cur_in_port)
#define current_op SWIT_PORT(cur_out_port)
#define current_ep SWIT_PORT(cur_err_port)
#define read_jbuf PORT_JBUF(current_ip)
#define lineno PORT_LINENO(current_ip)

/* imported (from file 'switRead.c') */
extern int start_expr;
extern SWiT_Expr compile_program (SWiT datum);
extern SWiT_Expr compile_lcode (SWiT code);
extern SWiT swit_load_path;
extern fd_set fds;
extern int start_string;
extern int inside_string;

/* from switEval.c */
extern int tk;

/* exported */
char *filename = "";
SWiT swit_stdin_port;
SWiT swit_stdout_port;
SWiT swit_stderr_port;
SWiT cur_in_port;
SWiT cur_out_port;
SWiT cur_err_port;

SWiT null_in_port;
SWiT null_out_port;


/*
 *  Forward definitions :
 */
int file_exists (char *);

/* File ports */
static SWiT call_with_input_file_prim (SWiT arg1, SWiT arg2);
static SWiT call_with_output_file_prim (SWiT arg1, SWiT arg2);
static SWiT port_p_prim (SWiT arg);
static SWiT current_input_port_prim (void);
static SWiT current_output_port_prim (void);
static SWiT open_input_file_prim (SWiT arg);
static SWiT open_output_file_prim (SWiT arg);
static SWiT with_input_from_file_prim (SWiT arg1, SWiT arg2);
static SWiT with_output_to_file_prim (SWiT arg1, SWiT arg2);
static SWiT close_port_prim (SWiT arg);
/* String ports */
static SWiT call_with_io_string_prim (SWiT arg1, SWiT arg2);
/* System interface */
static SWiT load_prim (SWiT arg);
static SWiT file_exists_prim (SWiT arg);
static SWiT system_prim (SWiT cmd);
static SWiT quit_prim (void);
static SWiT getenv_prim (SWiT arg);
static SWiT date_prim (void);


/*
 *================================================================
 *                   6.10.1.   PORTS
 *
 * ports are `generic' : they all use with the same functions
 * for input, output and seek.
 *
 * Implemented ports are : files, strings and sockets.
 *================================================================
 */

/*----------------------------------
 *   CONSTRUCTION
 *----------------------------------
 */
SWiT
swit_make_port (int descr, io_fun rd_fun, io_fun wr_fun, seek_fun sk_fun)
{
  SWiT obj = swit_alloc_object();
  SWiT_Port port = SWIT_ALLOC(Port);
  PORT_BUFFER(port) = (char *)swit_malloc_atomic(sizeof(char)*IO_BUFFER_SIZE);
  PORT_INDEX(port) = 0;
  PORT_EOB(port) = 0;
  PORT_LINENO(port) = 1;
  PORT_DESCR(port) = descr;
  PORT_READ_FUN(port) = rd_fun;
  PORT_WRITE_FUN(port) = wr_fun;
  PORT_SEEK_FUN(port) = sk_fun;
  PORT_JBUF(port) = (jmp_buf *)swit_malloc(sizeof(jmp_buf));
  SWIT_TYPE(obj) = PORT;
  SWIT_PORT(obj) = port;

  /* add the descr to fd set (safe string ports and std ports) */
  if (descr > 2) FD_SET(descr, &fds);

  return (obj);
}


/*----------------------------------
 *          PRINTING
 *----------------------------------
 */
static void
print_port (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  SWiT_Port p = SWIT_PORT(obj);
  int filedes = PORT_DESCR(p);
  if (filedes != -1) {
    char str[3];
    str2buff("#<port ");
    sprintf(str, "%d", filedes);
    str2buff(str);
    char2buff('>');
  } else {
    str2buff("#<closed port>");
  }
}
    
static void
print_eof (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  str2buff("##eof");
}


extern SWiT descr_to_port (int des);
extern SWiT search_port(int descr);

static SWiT
descr_to_port_prim (SWiT arg)
{
  int des = SWIT_FIXNUM(arg);
  switch (des)
    {
    case 0: return swit_stdin_port;
    case 1: return swit_stdout_port;
    case 2: return swit_stderr_port;
    case -1: {
      prim_err_msg = closed;
      exn = exn_io;
      return NULL;
    }
    default:
      {
	if (des < -1) {
	  SWiT port = descr_to_port (des);
	  if (port) return port;
	  prim_err_msg = no_port;
	  exn = exn_io;
	  return NULL;
	} else {
	  SWiT port = search_port(des);
	  if (port != swit_false) return port;
	  prim_err_msg = no_port;
	  exn = exn_io;
	  return NULL;
	}
      }
    }
}

/*========================================================
 *                    PIPE PORTS
 *========================================================
 */

/*
SWiT
make_pipe_prim (void)
{
  int p[2];
  if (pipe(p)) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }
  return swit_make_pair(swit_make_port(p[0], (io_fun)read, NULL, NULL),
			swit_make_port(p[1], NULL, (io_fun)write, NULL));
}

*/

/*========================================================
 *                    STRING PORTS
 *========================================================
 */

/*---------------------------------------------------------
 * Implementation of string ports needs 'indexed strings',
 * i.e. strings augmented with an index that gives the
 * position for the next I/O.
 *---------------------------------------------------------
 */
static int strings_idx;
struct indexed_string {
  int pos;   /* current position in string */
  SWiT str;  /* the (swit) string */
};
typedef struct indexed_string *indexed_string_Ptr;

/*-----------------------------------------------------------------
 * string ports are given descriptors <= -2 (to be distinguished
 *  from file or socket descriptors).
 * 510 is the max # of strings ports that can be opened 
 * simultaneously (should be enough).
 *-----------------------------------------------------------------
 */
static indexed_string_Ptr strings[512];
static SWiT string_ports[512];

SWiT
descr_to_port (int des)
{
  SWiT port = string_ports[-des];
  if (port && (PORT_DESCR(SWIT_PORT(port)) == des)) {
    return port;
  } else {
    return NULL;
  }
}

/*-----------------------------------------------------
 *  string->port
 *
 *  Creates a port from a string.
 *
 *  The string must be mutable.
 *-----------------------------------------------------
 */
static int make_string_descr (SWiT arg);
static int readstr_fun (int descr, char *buf, int size);
static int writestr_fun (int descr, char *buf, int size);
/* exported to switTk.c */
int string_seek (int descr, int offset, int mode);

/* exported to swit-tk.c */
SWiT
string_to_port_prim (SWiT arg)
{
  /* get a descriptor */
  int descr = make_string_descr(arg);
  SWiT port;
  if SWIT_CONST_STRINGP(arg) {
    port = swit_make_port(-descr, (io_fun)readstr_fun, NULL, string_seek);
  } else {
    port = swit_make_port(-descr, (io_fun)readstr_fun, (io_fun)writestr_fun, string_seek);
  }
  string_ports[descr] = port;
  return port;
}


static int
make_string_descr (SWiT arg)
{
  indexed_string_Ptr s = (indexed_string_Ptr)swit_malloc_atomic(sizeof(struct indexed_string));
  s->pos = 0;
  s->str = arg;
  strings[strings_idx++] = s;
  return strings_idx-1;
}

static int
readstr_fun (int descr, char *buf, int size)
{
  indexed_string_Ptr istr = strings[-descr];
  SWiT str = istr->str;
  int pos = istr->pos;
  int top = STRING_LENGTH(str) + 1;
  int cnt = ((pos + size) < top) ? size : top - pos;
  char *s = STRING_ELS(str)+pos;
  register int i;
  if (size > IO_BUFFER_SIZE) return -1;
  for (i = 0; i < cnt; i++) {
    *buf++ = *s++;
  }
  istr->pos = pos + cnt;
  return cnt;
}

static int
writestr_fun (int descr, char *buf, int size)
{
  indexed_string_Ptr istr = strings[-descr];
  SWiT str = istr->str;
  int pos = istr->pos;
  int top = STRING_LENGTH(str);
  if ((pos == top) || SWIT_CONST_STRINGP(str)) {
    return -1; /* error */
  } else {
    /* if size is too big, only that part of *buf
       that fits into the string will be writen
       (the number of bytes effectivly writen is returned) */
    int cnt = ((pos + size) < top) ? size : top - pos;
    char *s = STRING_ELS(str)+pos;
    register int i;
    for (i = 0; i < cnt; i++) {
      *s++ = *buf++;
    }
    istr->pos = pos + cnt;
    return cnt;
  }
}

/* exported to switTk.c */
int
string_seek (int descr, int offset, int mode)
{
  indexed_string_Ptr istr = strings[-descr];
  switch (mode)
    {
    case SEEK_CUR: {
      if (offset) {
	int cur_pos = istr->pos;
	SWiT str = istr->str;
	int new = cur_pos + offset;
	if ((new >= 0) && (new < STRING_LENGTH(str))) {
	  istr->pos = new;
	  return new;
	} else {
	  return -1; /* error */
	}
      } else {
	return istr->pos;
      }
    }
    case SEEK_SET: {
      SWiT str = istr->str;
      if ((offset >= 0) && (offset < STRING_LENGTH(str))) {
	istr->pos = offset;
	return offset;
      } else {
	return -1; /* error */
      }
    }
    case SEEK_END: {
      /* offset must be negative (positive offset that would specify
         a position past the end and extend the string are not allowed) */
      SWiT str = istr->str;
      int top = STRING_LENGTH(str);
      int new = top + offset;
      if ((new >= 0) && (new < top)) {
	istr->pos = new;
	return new;
      } else {
	return -1; /* error */
      }
    }
    default:
      return -1;
    }
}


/*-----------------------------------------------------
 * The functions `call-with-input-string' and
 * `call-with-output-string' are defined in swit
 * (in the file "strings.swt")
 *-----------------------------------------------------
 */
static SWiT
call_with_io_string_prim (SWiT arg1, SWiT arg2)
{
  SWiT port = string_to_port_prim(arg1);
  SWiT ret;
  tk = 1;
  ret = swit_apply(arg2, 1, &port);
  tk = 0;
  close_port_prim(port);
  return ret;
}


static SWiT
with_input_from_string_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_in_port;
  SWiT port = string_to_port_prim(arg1);
  SWiT ret;
  tk = 1;
  cur_in_port = port;
  ret = swit_apply(arg2, 0, NULL);
  tk = 0;
  close_port_prim(port);
  cur_in_port = old_port;
  return ret;
}

static SWiT
with_output_to_string_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_out_port;
  SWiT port = string_to_port_prim (arg1);
  SWiT ret;
  cur_out_port = port;
  tk = 1;
  ret = swit_apply(arg2, 0, NULL);
  tk = 0;
  close_port_prim(port);
  cur_out_port = old_port;
  return ret;
}


/*--------------------------------------------------
 *  port-string
 *
 * Returns the string that was `open' as a port
 *
 * To get the substring upto the current seek
 * position, use `seek-cur' (defined below):
 *
 *  (substring (string-port <sp>) (seek-cur <sp>))
 *
 *--------------------------------------------------
 */
static SWiT
port_string_prim (SWiT arg)
{
  SWiT_Port p = SWIT_PORT(arg);
  int des = PORT_DESCR(p);
  if (des > -2) {
    exn = exn_io;
    prim_err_msg = wrong_port;
    return NULL;
  } else {
    indexed_string_Ptr istr = strings[-des];
    if (istr) { /* should always be the case ! */
      return istr->str;
    } else {
      prim_err_msg = impl_err;
      return NULL;
    }
  }
}


/*========================================================
 *                    FILE PORTS
 *========================================================
 */


/*--------------------------------------------------------
 *  call-with-input-file, call-with-output-file
 *--------------------------------------------------------
 */
static SWiT
call_with_file (char *f,  SWiT arg, int flags)
{
  char *load_path = STRING_ELS(SYMBOL_VAL(SWIT_SYM(swit_load_path)));
  if ((strlen(filename) + strlen(load_path ? load_path : "")) > 250) {
    prim_err_msg = filename_too_long;
    exn = exn_io;
    return NULL;
  } else { 
    char buf[256];
    int filedes;
    buf[0] = '\0';
    if (load_path
	&& (*filename != '/')
	&& ((*filename != '.') || (*(filename+1) != '/'))) {
      strcpy(buf, load_path);
      if (buf[strlen(load_path)-1] != '/') {
	buf[strlen(load_path)] = '/';
	buf[strlen(load_path)+1] = '\0';
      } 
    }
    strcat(buf, f);
    if (flags == (O_WRONLY | O_CREAT)) {
      flags = O_WRONLY | O_CREAT | O_TRUNC;
    }
    filedes = open(buf, flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
    if (filedes == -1) {
      prim_err_msg = sys_errlist[errno];
      filename = f;
      exn = exn_io;
      return NULL;
    } else {
      SWiT port = swit_make_port(filedes, (io_fun)read, 
				 (io_fun)write, (seek_fun)lseek);
      SWiT ret;
      tk = 1;
      ret = swit_apply(arg, 1, &port);
      tk = 0;
      close_port_prim(port);
      return ret;
    }
  }
}

static SWiT
call_with_file_prim (SWiT arg1, SWiT arg2)
{
  return call_with_file(STRING_ELS(arg1), arg2, O_RDWR);
}

static SWiT
call_with_input_file_prim (SWiT arg1, SWiT arg2)
{
  return call_with_file(STRING_ELS(arg1), arg2, O_RDONLY);
}

static SWiT
call_with_output_file_prim (SWiT arg1, SWiT arg2)
{
  return call_with_file(STRING_ELS(arg1), arg2, O_WRONLY | O_CREAT);
}

/*--------------------------------------------------------
 *  port?, input-port?, output-port?, string-port?
 *  (for socket-port? see : switSocket.c)
 *--------------------------------------------------------
 */

static SWiT
port_p_prim (SWiT arg)
{
  return (SWIT_PORTP(arg) ? swit_true : swit_false);
}

static SWiT
inputport_p_prim (SWiT arg)
{
  if SWIT_PORTP(arg) {
    /* plus simple : tester write_fun A FAIRE */
    /* return PORT_READ_FUN(arg) ? swit_true : swit_false; */
    
    int descr = PORT_DESCR(SWIT_PORT(arg));
    if ((descr < -1) || (descr == 0) || (issocket(descr))) {
      return swit_true;
    } else {
      int flags = fcntl(descr, F_GETFL);
      return (flags == O_RDONLY ? swit_true : swit_false);
    }
  } else {
    return swit_false;
  }
}

static SWiT
outputport_p_prim (SWiT arg)
{
  if SWIT_PORTP(arg) {
    /* plus simple : tester read_fun */
    int descr = PORT_DESCR(SWIT_PORT(arg));
    switch (descr) {
    case 1: case 2: return swit_true;
    case 0: case -1: return swit_false;
    default:
      if (descr > 0) {
	int flags = fcntl(descr, F_GETFL);
	return ((flags & O_WRONLY) ? swit_true : swit_false);
      } else { /* string port : descr < -1 */
	return SWIT_CONST_STRINGP(port_string_prim(arg)) ? swit_false : swit_true;
      }
      
    }
  }
  return swit_false;
}

static SWiT
rwport_p_prim (SWiT arg)
{
  if SWIT_PORTP(arg) {
    /* plus simple : tester write_fun et read_fun */
    int descr = PORT_DESCR(SWIT_PORT(arg));
    if ((descr < -1) || (issocket(descr))) {
      return swit_true;
    } else {
      int flags = fcntl(descr, F_GETFL);
      return (flags == O_RDWR ? swit_true : swit_false);
    }
  } else {
    return swit_false;
  }
}

static SWiT
stringport_p_prim (SWiT arg)
{
  if SWIT_PORTP(arg) {
    return (PORT_DESCR(SWIT_PORT(arg)) < -1 ? swit_true : swit_false);
  } else {
    return swit_false;
  }
}


/*----------------------------------------------------------------
 *  current-input-port, current-output-port, current-error-port
 *----------------------------------------------------------------
 */

static SWiT
current_input_port_prim (void)
{
  return (cur_in_port);
}

static SWiT
current_output_port_prim (void)
{
  return (cur_out_port);
}

static SWiT
current_error_port_prim (void)
{
  return (cur_err_port);
}


/*--------------------------------------------------------
 *  with-input-from-file, with-output-from-file
 *--------------------------------------------------------
 */

static SWiT
with_input_from_file_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_in_port;
  SWiT port = open_input_file_prim(arg1);
  if (!port) {
    exn = exn_io;
    return NULL;
  } else {
    SWiT ret;
    cur_in_port = port;
    tk = 1;
    ret = swit_apply(arg2, 0, NULL);
    tk = 0;
    close_port_prim(port);
    cur_in_port = old_port;
    return ret;
  }
}

static SWiT
with_output_to_file_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_out_port;
  SWiT port = open_output_file_prim (arg1);
  SWiT ret;
  if (!port) {
    exn = exn_io;
    return NULL;
  } else {
    tk = 1;
    cur_out_port = port;
    ret = swit_apply(arg2, 0, NULL);
    close_port_prim(port);
    cur_out_port = old_port;
    tk = 0;
    return ret;
  }
}


/*--------------------------------------------------------
 *  open-input-file, open-output-file
 *--------------------------------------------------------
 */

static SWiT
open_file (char *filename, int flags)
{
  char *load_path;
  if (filename == NULL) {
    prim_err_msg = file_not_exists;
    exn = exn_io;
    return NULL;
  }
  load_path = STRING_ELS(SYMBOL_VAL(SWIT_SYM(swit_load_path)));
  if (load_path == NULL) {
    load_path = "";
  }
  if ((strlen(filename) + strlen(load_path ? load_path : "")) > 250) {
    prim_err_msg = filename_too_long;
    exn = exn_io;
    return NULL;
  } else { 
    char buf[256];
    int filedes;
    buf[0] = '\0';
    if ((*load_path != 0) 
	&& (*filename != '/')
	&& ((*filename != '.') || (*(filename+1) != '/'))) {
      strcpy(buf, load_path);
      if (buf[strlen(load_path)-1] != '/') {
	buf[strlen(load_path)] = '/';
	buf[strlen(load_path)+1] = '\0';
      } 
    }
    strcat(buf, filename);
    if (flags == (O_WRONLY | O_CREAT)) {
      flags = O_WRONLY | O_CREAT | O_TRUNC;
    }
    filedes = open(buf, flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
    if (filedes == -1) {
      prim_err_msg = sys_errlist[errno];
      exn = exn_io;
      return NULL;
    }
    return swit_make_port(filedes, (io_fun)read, (io_fun)write, (seek_fun)lseek);
  }
}

static SWiT
open_file_prim (SWiT arg)
{
  return open_file(STRING_ELS(arg), O_RDWR | O_CREAT);
}

static SWiT
open_input_file_prim (SWiT arg)
{
  return open_file(STRING_ELS(arg), O_RDONLY);
}

static SWiT
open_output_file_prim (SWiT arg)
{
  return open_file(STRING_ELS(arg), O_WRONLY | O_CREAT);
}


/*------------------------------------------------------
 *  close-input-port, close-output-port
 *------------------------------------------------------
 */

/* exported to switSocket.c */
SWiT
close_port_prim (SWiT arg)
{
  SWiT_Port p = SWIT_PORT(arg);
  int des = PORT_DESCR(p);
  if (des < -1) { /* string port */
    int i = strings_idx-1;
    PORT_DESCR(p) = -1;
    strings[-des] = NULL;
    while ((strings[i] == NULL) && (i > 1)) i--;
    strings_idx = i+1;
    return swit_true;
  } else if ((des == 0) || (des == 1) || (des == 2)) {
    return swit_false;
  } else {
    if (issocket(des)) {
      if (shutdown(des, 2) == -1) {
	fprintf(stderr, cannot_close);
	return swit_false;
      }
    }
    if (close(des) == -1) {
      fprintf(stderr, already_closed, sys_errlist[errno]);
      return swit_false;
    }
    FD_CLR(des, &fds);
    PORT_DESCR(p) = -1;
    lineno = 1;
    return swit_true;
  }
}

/*------------------------------------------------------
 *  transcript-on, transcript-off
 *------------------------------------------------------
 */

int transcript_descr;

static SWiT
transcript_on_prim(SWiT arg)
{
  char *filename = STRING_ELS(arg);
  int filedes = open(filename, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
  if ((filedes == -1) || (lseek(filedes, 0, SEEK_END) == -1)) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  } else {
    transcript_descr = filedes;
    return swit_true;
  }
}

static SWiT
transcript_off_prim()
{
  if (close(transcript_descr) == -1) {
    fprintf(stderr, "%s\n", sys_errlist[errno]);
    return swit_false;
  }
  transcript_descr = -1;
  return swit_true;
}

static int
writestd (int descr, char *buf, int size)
{
  int cnt = write(descr, buf, size);
  if (transcript_descr != -1) {
    write(transcript_descr, buf, size);
  }
  return cnt;
}

static int
readstd (int descr, char *buf, int size)
{
  int cnt = read(descr, buf, size);
  if (transcript_descr != -1) {
    write(transcript_descr, buf, size);
  }
  return cnt;
}



/*
 *================================================================
 *                 POSITIONING (for file or string ports)
 *================================================================
 */

extern SWiT int_to_integer (int n);

static SWiT
seek_set_port_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Port port = SWIT_PORT(arg1);
  int n, descr = PORT_DESCR(port);
  seek_fun seekfun = PORT_SEEK_FUN(port);
  if (seekfun == NULL) {
    prim_err_msg = cannot_seek;
    exn = exn_io;
    return NULL;
  }
  if (descr == -1) {
    prim_err_msg = closed;
    exn = exn_io;
    return NULL;
  }
  if SWIT_FIXNUMP(arg2) {
    n = SWIT_FIXNUM(arg2);
  } else if SWIT_BIGNUMP(arg2) {
    n = mpz_get_ui(SWIT_BIGNUM(arg2));
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
  if ((*seekfun)(descr, n, SEEK_SET) == -1) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  } else {
    /* The buffer is emptied : it is always flushed
     * when output operations are performed (when input, it will
     * be simply re-read).
     */
    PORT_INDEX(port) = 0;
    PORT_EOB(port) = 0;
    return arg2;
  }
}


/* exported to switPrint.c */
SWiT
seek_cur_port_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Port port = SWIT_PORT(arg1);
  int n, cur_pos, pos, m, eob = PORT_EOB(port);
  int descr = PORT_DESCR(port);
  seek_fun seekfun = PORT_SEEK_FUN(port);
  if (seekfun == NULL) {
    prim_err_msg = cannot_seek;
    exn = exn_io;
    return NULL;
  }
  if (descr == -1) {
    prim_err_msg = closed;
    exn = exn_io;
    return NULL;
  }
  if SWIT_FIXNUMP(arg2) {
    n = SWIT_FIXNUM(arg2);
  } else if SWIT_BIGNUMP(arg2) {
    n = mpz_get_ui(SWIT_BIGNUM(arg2));
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }

  m = PORT_INDEX(port) + n;
  cur_pos = seekfun(descr, 0, SEEK_CUR);
  if ((n == 0) || ((m >= 0) && (m < eob))) {
    PORT_INDEX(port) = m;
    return int_to_integer(m + cur_pos - eob);
  } else {
    if ((pos = (*seekfun)(descr, m - eob, SEEK_CUR)) == -1) {
      prim_err_msg = sys_errlist[errno];
      exn = exn_io;
      return NULL;
    } else {
      PORT_INDEX(port) = 0;
      PORT_EOB(port) = 0;
    }
    return (int_to_integer(pos));
  }
}

static SWiT
seek_end_port_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Port port = SWIT_PORT(arg1);
  int n, pos, descr = PORT_DESCR(port);
  seek_fun seekfun = PORT_SEEK_FUN(port);
  if (seekfun == NULL) {
    prim_err_msg = cannot_seek;
    exn = exn_io;
    return NULL;
  }
  if (descr == -1) {
    prim_err_msg = closed;
    exn = exn_io;
    return NULL;
  }
  if SWIT_FIXNUMP(arg2) {
    n = SWIT_FIXNUM(arg2);
  } else if SWIT_BIGNUMP(arg2) {
    n = mpz_get_ui(SWIT_BIGNUM(arg2));
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }

  if ((pos = (*seekfun)(descr, n, SEEK_END)) == -1) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  } else {
    PORT_INDEX(port) = 0;
    PORT_EOB(port) = 0;
    return (int_to_integer(pos));
  }
}


/*
 *================================================================
 *                   6.10.4.  SYSTEM INTERFACE
 *================================================================
 */

int loading = 0;
extern SWiT swit_print_depth;
extern SWiT swit_print_span;
extern SWiT swit_load_path;
extern SWiT_Ccont done_cont;

void
swit_load_init (char *file)
{
  char filename[256];
  int filedes;
  char *lib_path = getenv("SWIT_LIBRARY_PATH");
  if (!lib_path || !*lib_path) {
    if (file_exists("/usr/local/lib/swit/compile.obj") != -1) {
      lib_path = "/usr/local/lib/swit";
    } else if (file_exists("lib/compile.obj") != -1) {
      lib_path = "lib";
    } else {
      lib_path = "../lib";
    }
  }
  strncpy(filename, lib_path, 240);
  strcat(filename, file);
  filedes = open(filename, O_RDONLY);
  if (filedes == -1) {
    fprintf(stderr, lib_path_error, filename);
    exit(1);
  } else {
    SWiT old_port = cur_in_port;
    int token;
    cur_in_port = swit_make_port(filedes, (io_fun)read, NULL, NULL);
    printf(loading_file, filename);
    loading = 1;
    if (setjmp(*read_jbuf)) {
      goto end;
    } else {
      while ((token = get_tok())) {
	/* loading until eof or error */
	if (!swit_eval(compile_program(rd_datum(token)), NULL, done_cont)) goto end;
      }
    }
  end:
    loading = 0;
    close_port_prim(cur_in_port);
    cur_in_port = old_port;
  }
}


/* TODO (?) : record the file names already loaded along with the file data
          and don't reload an unchanged file 
*/
extern SWiT big; /* big = 1000, initialize in switPrint.c */
extern int read_paren;
static int start_expr;
extern SWiT swit_type_flag;
extern SWiT swit_magic;
static int
iscode (SWiT datum)
{
  if SWIT_VECTORP(datum) {
    SWiT *code = VECTOR_ELS(datum);
    if (*code == swit_magic) return 1;
  }
  return 0;
}

static SWiT
load_prim (SWiT arg)
{
  int err;
  char *old_filename = filename;
  SWiT saved_lp = SYMBOL_VAL(SWIT_SYM(swit_load_path));
  char *load_path = STRING_ELS(saved_lp);
  if ((STRING_LENGTH(arg) + strlen(load_path ? load_path : "")) > 250) {
    prim_err_msg = filename_too_long;
    exn = exn_io;
    return NULL;
  } else { 
    int filedes;
    SWiT ret = swit_false;
    SWiT old_port = cur_in_port;
    SWiT old_print_span = SYMBOL_VAL(SWIT_SYM(swit_print_span));
    SWiT old_print_depth = SYMBOL_VAL(SWIT_SYM(swit_print_depth));
    char *buf = swit_malloc(256);
    void *j;
    int token;
    static int found;
    buf[0] = '\0';
    filename = STRING_ELS(arg);
    if (load_path
	&& (*filename != '/')
	&& ((*filename != '.') || (*(filename+1) != '/'))) {
      strcpy(buf, load_path);
      if (buf[strlen(load_path)-1] != '/') {
	buf[strlen(load_path)] = '/';
	buf[strlen(load_path)+1] = '\0';
      } 
    }
    strcat(buf, filename);
    filedes = open(buf, O_RDONLY);
    if (filedes == -1) {
      if (loading) {
	fprintf(stderr, skip_file, lineno, buf);
	found = 0;
      } else {
	fprintf(stderr, file_not_exists, buf);
      }
      return swit_false;
    }
    printf(loading_file, buf);
    if ((j = strrchr(buf, '/'))) {
      buf[strlen(buf)-strlen((char *)j)+1]='\0';
      SYMBOL_VAL(SWIT_SYM(swit_load_path)) = swit_make_string(buf, strlen(buf));
    }
    loading = 1;
    found = 1;
    cur_in_port = swit_make_port(filedes, (io_fun)read, NULL, NULL);
    SYMBOL_VAL(SWIT_SYM(swit_print_depth)) = big;
    SYMBOL_VAL(SWIT_SYM(swit_print_span)) = big;
    if ((err = setjmp(*read_jbuf))) {
      goto end;
    }
    tk = 1;
    while ((token = get_tok())) {
      SWiT datum;
      int iscode_flag;
      SWiT_Expr expr;
      start_expr = lineno;
      datum = rd_datum(token);
      iscode_flag = iscode(datum);
      expr = compile_program(datum);
      ret = swit_eval(expr, NULL, done_cont);
      if (!ret) {
	fprintf(stderr, "--- ");
	if (inside_string) {
	  fprintf(stderr, err_inside_string, start_string, filename);
	} else if (read_paren) {
	  fprintf(stderr, err_inside_expr, start_expr, filename);
	} else {
	  fprintf(stderr, err_at_line, lineno, filename);
	  fprintf(stderr, "     (");
	  fprintf(stderr, prim_err_msg, -1);
	  fprintf(stderr, ")\n");
	}
      } else {
	if (SWIT_SYMBOLP(ret)
	    && (SYMBOL_VAL((SWiT_Symbol)swit_verbose->data) == swit_true)) {
	  printf("(%s) line %d -- %s", filename, lineno,
		 (char *)STRING_ELS(SYMBOL_NAME(SWIT_SYM(ret))));
	  printf("\n");
	}
      }
    }
  end:
    SYMBOL_VAL(SWIT_SYM(swit_print_span)) = old_print_span;
    SYMBOL_VAL(SWIT_SYM(swit_print_depth)) = old_print_depth;
    SYMBOL_VAL(SWIT_SYM(swit_load_path)) = saved_lp;
    loading = 0;  
    tk = 0;
    /* eof has been reached */
    filename = old_filename;
    close_port_prim(cur_in_port);
    cur_in_port = old_port;
    return (ret);
  }
}

SWiT
swit_load (char *file)
{
  return (load_prim (swit_make_string(file, strlen(file))));
}

int
file_exists (char *filename)
{
  struct stat *buf = (struct stat *)swit_malloc_atomic(sizeof(struct stat));
  int err = stat(filename, buf);
  return (err);
}


static SWiT
file_exists_prim (SWiT arg)
{
  char *filename = STRING_ELS (arg);
  return (file_exists(filename) == -1 ? swit_false : swit_true);
}

SWiT
system_prim (SWiT cmd)
{
  int result = system(STRING_ELS(cmd));
  return (result == -1 ? swit_false : swit_true);
}

SWiT
chdir_prim (SWiT path)
{
  int result = chdir(STRING_ELS(path));
  return (result == -1 ? swit_false : swit_true);
}

static SWiT
quit_prim (void)
{
  fprintf(stdout, bye);
  exit(0);
}

static SWiT
pwd_prim (void)
{
  char *name;
  if ((name = getcwd(NULL,0)) == NULL) {
    prim_err_msg = permission_denied;
    exn = exn_io;
    return NULL;
  } else {
    return swit_make_constant_string(name, strlen(name));
  }
}

extern char **environ;

static SWiT
getenv_prim (SWiT arg)
{
  char **env = environ;
  if (arg == swit_null_string) {
    SWiT result = swit_null_list;
    while (*env) {
      result = swit_make_pair(swit_make_constant_string(*env, strlen(*env)), result);
      env++;
    }
    return result;
  } else {
    char *s = getenv(STRING_ELS(arg));
    return (*s ? swit_make_string(s, strlen(s)) : swit_null_string);
  }
}

static SWiT
date_prim (void)
{
  time_t t = (time_t)swit_malloc_atomic(sizeof(time_t));
  time(&t);
  return (swit_make_string(ctime(&t), 10));
}



/*===========================
 *    INITIALISATION
 *===========================
 */
void
swit_init_io (void)
{
  print_table[PORT >> 1] = print_port;
  print_table[EOF_TYPE >> 1] = print_eof;

  /* Initialize the set of descriptors */
  FD_ZERO(&fds);

  transcript_descr = -1;
  null_in_port = swit_make_port(-1, NULL, NULL, NULL);
  null_out_port = swit_make_port(-1, NULL, NULL, NULL);
  cur_in_port = swit_stdin_port = swit_make_port(0, (io_fun)readstd, NULL, NULL);
  cur_out_port = swit_stdout_port = swit_make_port(1, NULL, (io_fun)writestd, NULL);
  cur_err_port = swit_stderr_port = swit_make_port(2, NULL, (io_fun)writestd, NULL);

  /* Ports */
  swit_make_prim2 ("call-with-input-file", call_with_input_file_prim,
		   STRING_M, PROC_M);
  swit_make_prim2 ("call-with-output-file", call_with_output_file_prim,
		   STRING_M, PROC_M);
  swit_make_prim1 ("input-port?", inputport_p_prim,
		   ANY);
  swit_make_prim1 ("output-port?", outputport_p_prim,
		   ANY);
  swit_make_prim0 ("current-input-port", current_input_port_prim);
  swit_make_prim0 ("current-output-port", current_output_port_prim);
  swit_make_prim2 ("with-input-from-file", with_input_from_file_prim,
		   STRING_M, PROC_M);
  swit_make_prim2 ("with-output-to-file", with_output_to_file_prim,
		   STRING_M, PROC_M);
  swit_make_prim1 ("open-input-file", open_input_file_prim,
		   STRING_M);
  swit_make_prim1 ("open-output-file", open_output_file_prim,
		    STRING_M);
  swit_make_prim1 ("close-input-port", close_port_prim,
		   PORT);
  swit_make_prim1 ("close-output-port", close_port_prim,
		   PORT);
  /* extensions */
  swit_make_prim2 ("call-with-file", call_with_file_prim,
		   STRING_M, PROC_M);
  swit_make_prim1 ("port?", port_p_prim,
		   ANY);
  swit_make_prim1 ("descr->port", descr_to_port_prim,
		   FIXNUM);
  swit_make_prim1 ("read-write-port?", rwport_p_prim,
		   ANY);
  swit_make_prim1 ("open-file", open_file_prim,
		   STRING_M);
  swit_make_prim1 ("close-port", close_port_prim,
		   PORT);
  swit_make_prim0 ("current-error-port", current_error_port_prim);

  /* Pipes */
  /*  swit_make_prim0 ("make-pipe", make_pipe_prim); */

  /* String ports */
  strings_idx = 2; /* first used index in `strings' array */
  swit_make_prim2 ("call-with-io-string", call_with_io_string_prim, 
		   STRING_M, PROC_M);
  swit_make_prim2 ("with-input-from-string", with_input_from_string_prim, 
		   STRING_M, PROC_M);
  swit_make_prim2 ("with-output-to-string", with_output_to_string_prim, 
		   STRING_M, PROC_M);
  swit_make_prim1 ("string->port", string_to_port_prim,
		   STRING_M);
  swit_make_prim1 ("string-port?", stringport_p_prim,
		   ANY);
  swit_make_prim1 ("port-string", port_string_prim,
		   PORT);

/* File positioning */
  swit_make_prim12 ("seek-set", seek_set_port_prim,
		    PORT, NUM_M,swit_zero);
  swit_make_prim12 ("seek-cur", seek_cur_port_prim,
		    PORT, NUM_M, swit_zero);
  swit_make_prim12 ("seek-end", seek_end_port_prim,
		    PORT, NUM_M, swit_zero);

/* System interface */
  swit_make_prim1 ("load", load_prim,
		   STRING_M);
  swit_make_prim1 ("transcript-on", transcript_on_prim,
		   STRING_M);
  swit_make_prim0 ("transcript-off", transcript_off_prim);
  swit_make_prim1 ("file-exists?", file_exists_prim,
		   STRING_M);
  swit_make_prim1 ("system", system_prim,
		   STRING_M);
  swit_make_prim1 ("chdir", chdir_prim,
		   STRING_M);
  swit_make_prim0 ("quit", quit_prim);
  swit_make_prim0 ("pwd", pwd_prim);
  swit_make_prim01 ("getenv", getenv_prim,
		    STRING_M, swit_null_string);
  swit_make_prim0 ("date", date_prim);
}

