/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*-------------------------------------------------------
 *              MISCELLENAOUS EXTENSIONS
 *
 *  Environments :
 *  ^^^^^^^^^^^^
 *    closure-env         CLOSURE -> ENVIRONMENT
 *    extend-env          ENVIRONMENT X ENVIRONMENT -> ENVIRONMENT
 *    promise-env         PROMISE -> ENVIRONMENT
 *    continuation-env    CONTINUATION -> ENVIRONMENT
 *    environment?       'a -> BOOLEAN
 *    parent-env          ENVIRONMENT -> ENVIRONMENT
 *    frame->alist        ENVIRONMENT -> ALIST
 *    env-depth           ENVIRONMENT -> INTEGER
 *
 *  Macros :
 *  ^^^^^^
 *    macro?             'a -> BOOLEAN
 *    macro-expand1       LIST -> LIST
 *    macro-expand        LIST -> LIST
 *
 *  Garbage collector :
 *  ^^^^^^^^^^^^^^^^^
 *    gc-collect          VOID -> BOOLEAN
 *    gc-heap-size        VOID -> INTEGER
 *    gc-expand-heap      INTEGER -> BOOLEAN
 *    gc-size             OBJ -> INTEGER
 *
 *  Debugging tools :
 *  ^^^^^^^^^^^^^^^
 *    error               STRING X 'a -> BOOLEAN
 *    cerror              STRING X (VOID -> 'a) -> 'b
 *
 *  Time :
 *  ^^^^
 *    clock              VOID -> INTEGER
 *    set-timer          INTEGER X MODE X [PROCEDURE] -> INTEGER
 *    get-timer          VOID -> INTEGER
 *    pause              INTEGER -> BOOLEAN
 *
 *  Other :
 *  ^^^^^
 *    version             VOID -> STRING   
 *    quit                VOID -> ...
 *
 *-------------------------------------------------------
 */

#include "swit.h"
#include <sys/times.h>
#include <sys/time.h>

/* exported */
SWiT make_fv(SWiT_Env e);
SWiT make_env (SWiT_Env env); /* to switSymbol.c */


/* imported (from switEval.c) */
extern SWiT_Env swit_extend_env (SWiT names, SWiT *vals, SWiT_Env env);
extern void print_lambda();
extern void mem2buff();
extern SWiT swit_class;
extern void swit_print_code (SWiT_Expr obj, SWiT fv);
extern void print_code_to_buffer (SWiT_Expr obj, SWiT fv, int depth);
extern int search (SWiT sym, SWiT names);
extern SWiT swit_symbol_value (SWiT sym, SWiT env);
extern void print_sym();
extern void print_str();
extern SWiT make_fv(SWiT_Env e);
extern SWiT swit_append2 (SWiT lst1, SWiT lst2);


/*----------------------------------------------------------
 *   Locally defined 
 *----------------------------------------------------------
 */
/* environments */
static SWiT closure_env_prim (SWiT arg);
static SWiT append_env_prim (SWiT arg1, SWiT arg2);
static SWiT promise_env_prim (SWiT arg);
static SWiT env_p_prim (SWiT val);
static SWiT env_depth_prim (SWiT arg);
static SWiT_Env parent_env (SWiT_Env e, int k);
/* macros */
static SWiT macro_p_prim (SWiT val);
static SWiT macro_expand_prim(SWiT exp);
static SWiT macro_expand_all_prim (SWiT obj);
/* GC */
static SWiT gcollect_prim (void);
static SWiT gheap_size_prim (void);
static SWiT gc_expand_prim (SWiT arg);
/* debugging tools */
static SWiT error_prim (SWiT msg, SWiT exp);
/* time */
static SWiT clock_prim (void);
static SWiT get_timer_prim(void);
/* mutable */
static SWiT to_immutable_prim (SWiT arg);
static SWiT to_mutable_prim (SWiT arg);
static SWiT mutable_p_prim (SWiT arg);


/*---------------------------------------------------------
 *                     Environments
 *---------------------------------------------------------
 */

/* exported to switInit.c to initialize swit_void_env */
SWiT
swit_make_env (SWiT_Env env)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = ENV;
  SWIT_ENV(obj) = env;
  return (obj);
}


static void
print_environment (SWiT obj, int tcl, SWiT fv, int depth)
{
  /*  if (tcl) {
    str2buff("{environment}");
  } else { */
    if (obj != swit_void_env) {
      SWiT class = swit_symbol_value(swit_class, obj);
      if (class != swit_unbound) {
	SWiT str;
	SWiT print = swit_symbol_value(swit_make_symbol("print", 5, swit_unbound), obj);
	if ((print != swit_unbound) 
	    && (str = swit_apply(print, 1, &obj))) { /* BUG! pas rentrant */
	  print_str(STRING_ELS(str), STRING_LENGTH(str), 0, 0);
	} else {
	  str2buff("#<class: ");
	  print_sym(SWIT_SYM(class));
	  char2buff('>');
	}
      } else {
	str2buff("#<environment>"); /* todo (at least : env names) */
      }
    } else {
      str2buff("#<environment>");
    }
    /* } */
}


static SWiT_Env
parent_env (SWiT_Env e, int k)
{
  register int i;
  for (i = 0; i < k; i++) { 
    if (e) {
      e = ENV_NEXT(e);
    } else {
      prim_err_msg = no_parent;
      return NULL;
    }
  }
  return(e);
}
    
static SWiT
parent_env_prim (SWiT env, SWiT arg)
{
  SWiT_Env e;
  int level = SWIT_FIXNUM(arg);
  if (level < 0) {
    exn = exn_range;
    prim_err_msg = out_of_range;
    return NULL;
  }
  prim_err_msg = NULL;
  e = parent_env(SWIT_ENV(env), level);
  if (prim_err_msg) {
    exn = exn_env;
    return NULL;
  } else {
    return (e ? swit_make_env(e) : swit_void_env);
  }
}

static SWiT
env_depth_prim (SWiT arg)
{
  SWiT_Env e = SWIT_ENV(arg);
  int depth = 0;
  while (e) {
    depth++;
    e = ENV_NEXT(e);
  }
  return swit_make_fixnum(depth);
}

static SWiT
closure_env_prim (SWiT arg)
{
  if SWIT_CLOSUREP(arg) {
    SWiT_Env env = CLOSURE_ENV(SWIT_CLOSURE(arg));
    return (env ? swit_make_env(env) : swit_void_env);
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
}

static SWiT
append_env_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Env e1 = SWIT_ENV(arg1);
  if (!e1) {
    return (arg2);
  } else {
    SWiT_Env last, first, e = e1;
    first = last = swit_extend_env(ENV_NAMES(e), ENV_FRAME(e), NULL);
    e = ENV_NEXT(e);
    while (e) {
      SWiT_Env new = swit_extend_env(ENV_NAMES(e), ENV_FRAME(e), NULL);
      ENV_NEXT(last) = new;
      last = new;
      e = ENV_NEXT(e);
    }
    ENV_NEXT(last) = SWIT_ENV(arg2);
    return(swit_make_env(first));
  }
}

/*
static SWiT
nconc_env_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Env e = SWIT_ENV(arg1);
  if (!e) {
    return (arg2);
  } else {
    while (ENV_NEXT(e)) {
      e = ENV_NEXT(e);
    }
    ENV_NEXT(e) = SWIT_ENV(arg2);
    return(arg1);
  }
}
*/

static SWiT
promise_env_prim (SWiT arg)
{
  SWiT_Env env = PROMISE_ENV(SWIT_PROMISE(arg));
  return (env ? swit_make_env(env) : swit_void_env);
}

static SWiT
env_p_prim (SWiT arg)
{
  return (SWIT_ENVP(arg) ? swit_true : swit_false);
}

/* exported to switPrint.c, switTop.c and switControl.c */
SWiT
make_fv(SWiT_Env e)
{
  SWiT fv = swit_make_pair(swit_null_list, swit_null_list);
  SWiT last = fv;
  while (e) {
    SWiT new = swit_make_pair(ENV_NAMES(e), swit_null_list);
    SET_CDR(last, new);
    last = new;
    e = ENV_NEXT(e);
  }
  return(SWIT_CDR(fv));
}

/*
static SWiT
strip (SWiT lst)
{
  SWiT res = swit_make_pair(swit_null_list, swit_null_list);
  SWiT last = res;
  while (lst != swit_null_list) {
    SWiT new = swit_make_pair(SWIT_CAR(SWIT_CAR(lst)), swit_null_list);
    SET_CDR(last, new);
    last = new;
    lst = SWIT_CDR(lst);
  }
  return(SWIT_CDR(res));
}
*/

static SWiT
get_env_names_prim (SWiT arg)
{
  SWiT_Env e = SWIT_ENV(arg);
  SWiT fv = swit_make_pair(swit_null_list, swit_null_list);
  SWiT last = fv;
  while (e) {
    /* unstripped */
    SWiT new = swit_make_pair(ENV_NAMES(e), swit_null_list);
    SET_CDR(last, new);
    last = new;
    e = ENV_NEXT(e);
  }
  return(SWIT_CDR(fv));
}

static SWiT
frame_to_alist_prim (SWiT arg)
{
  SWiT_Env env = SWIT_ENV(arg);
  if (env) {
    SWiT first = swit_make_pair(swit_null_list, swit_null_list);
    SWiT last = first;
    SWiT names = ENV_NAMES(env);
    SWiT *values = ENV_FRAME(env);
    while (names != swit_null_list) {
      SWiT new = swit_make_pair
	            (swit_make_pair (SWIT_CAR(SWIT_CAR(names)), *values),
		     swit_null_list);
      SET_CDR(last, new);
      last = new;
      names = SWIT_CDR(names);
      values++;
    }
    return SWIT_CDR(first);
  } else {
    return swit_null_list;
  }
}

/*---------------------------------------------------------
 *         Macros
 *---------------------------------------------------------
 */

static SWiT *
list_to_vector (int n, SWiT lst)
{
  SWiT *vec = swit_alloc_ptrs(n+1); /* +1 = hack for vargs */
  register i;
  for (i = 0; i < n; i++) {
    vec[i] = SWIT_CAR(lst);
    lst = SWIT_CDR(lst);
  }
  return (vec);
}

static SWiT
macro_p_prim (SWiT arg)
{
  return (SWIT_MACROP(arg) ? swit_true : swit_false);
}


static SWiT
macro_expand_prim (SWiT obj)
{
  SWiT car = SWIT_CAR(obj);
  if SWIT_MACROP(car) {
    SWiT args = SWIT_CDR(obj), expansion;
    int len = swit_list_length(args);
    SWiT macro = SYMBOL_VAL(SWIT_SYM(car));
    expansion = swit_apply(macro, len, list_to_vector(len, args));
    return expansion;
  } else {
    return obj;
  }
}

static SWiT
macro_expand_all_prim (SWiT obj)
{
  if SWIT_PAIRP(obj) {
    SWiT car = SWIT_CAR(obj);
    SWiT args = SWIT_CDR(obj);
    if SWIT_MACROP(car) {
      int len = swit_list_length(args);
      SWiT macro = SYMBOL_VAL(SWIT_SYM(car));
      SWiT expand = swit_apply(macro, len, list_to_vector(len, args));
      if (!expand) {
	return NULL;
      } 
      return macro_expand_all_prim(expand);
    } else {
      SWiT lst = args;
      while SWIT_PAIRP(lst) {
	SWiT a = SWIT_CAR(lst);
	SET_CAR(lst, macro_expand_all_prim (a));
	lst = SWIT_CDR(lst);
      }
      return obj;
    }
  } else {
    return obj;
  }
}


/*---------------------------------------------------------
 *         GC
 *---------------------------------------------------------
 */

static SWiT
gcollect_prim (void)
{
  GC_gcollect();
  return(swit_true);
}

static SWiT
gheap_size_prim (void)
{
  int hs = GC_get_heap_size();
  return(swit_make_fixnum(hs));
}


static SWiT
gc_expand_prim (SWiT arg)
{
  if SWIT_FIXNUMP(arg) {
    GC_expand_hp(SWIT_FIXNUM(arg));
    return(swit_true);
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
}

/*---------------------------------------------------------
 *           Exceptions and errors
 *---------------------------------------------------------
 */
static SWiT
error_prim (SWiT msg, SWiT exp)
{
  swit_signal_error(exp, STRING_ELS(msg));
  return(NULL);
}


/*--------------------------------
 *    time and timer functions
 *--------------------------------
 */
extern SWiT int_to_integer (int n);
static SWiT
clock_prim (void)
{
  int cps = sysconf(_SC_CLK_TCK);
  return(int_to_integer((clock() * 1000) / cps));
}

static SWiT
select_prim(SWiT arg)
{
  struct timeval new;
  int t;
  if SWIT_FIXNUMP(arg) {
    t = SWIT_FIXNUM(arg);
  } else if SWIT_BIGNUMP(arg){
    MP_INT *x = SWIT_BIGNUM(arg);
    if (mpz_size(x) == 1) {
      t = mpz_get_si(x);
    } else {
      prim_err_msg = too_big;
      exn = exn_range;
      return NULL;
    }
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;    
  }
  new.tv_usec = 1000*t;
  new.tv_sec = 0;
  if (select(0, NULL, NULL, NULL, &new) == -1)
    return swit_false;
  return swit_true;
}



/*---------------------------------------------------------
 *  Mutable/immutable and others
 *---------------------------------------------------------
 */
static SWiT 
to_immutable_prim (SWiT arg)
{
  return NULL;
}

static SWiT 
to_mutable_prim (SWiT arg)
{
  return NULL;
}

static SWiT 
mutable_p_prim (SWiT arg)
{
  return swit_true;
}

/*
static SWiT
const_string_p_prim (SWiT arg)
{
  return (SWIT_CONST_STRINGP(arg) ? swit_true : swit_false);
}

static SWiT
string_to_const_prim (SWiT arg)
{
  if (arg == swit_null_string) {
    return swit_null_string;
  } else {
    SWIT_TYPE(arg) |= 0x0010;
    return (arg);
  }
}

static SWiT
const_to_string_prim (SWiT arg)
{
  if (arg == swit_null_string) {
    return swit_null_string;
  } else {
    SWIT_TYPE(arg) &= ~0x0010;
    return (arg);
  }
}
*/
extern SWiT swit_version;
static SWiT
version_prim (void)
{
  return(swit_version);
}

static SWiT
identity_prim (SWiT arg)
{
  return (arg);
}

/*---------------------------------------------------------
 *  INIT
 *---------------------------------------------------------
 */
void
swit_init_misc (void)
{
  print_table[ENV >> 1] = print_environment;

/* Environments */
  swit_make_prim0 ("current-env", NULL); /* in eval */
  swit_make_prim1 ("closure-env", closure_env_prim,
		    PROC_M);
  swit_make_prim2 ("extend-env", append_env_prim,
		    ENV, ENV);
  swit_make_prim1 ("promise-env", promise_env_prim,
		   PROMISE);
  swit_make_prim12 ("parent-env", parent_env_prim,
		    ENV, FIXNUM, 
		    swit_one);
  swit_make_prim1 ("get-env-names", get_env_names_prim,
		   ENV);
  swit_make_prim1 ("environment?", env_p_prim,
		   ANY);
  swit_make_prim1 ("frame->alist", frame_to_alist_prim,
		   ENV);
  swit_make_prim1 ("env-depth", env_depth_prim,
		   ENV);
/* Macros */
  swit_make_prim1 ("macro?", macro_p_prim,
		   SYMBOL);
  swit_make_prim1 ("macro-expand1", macro_expand_prim,
		   LIST);
  swit_make_prim1 ("macro-expand", macro_expand_all_prim, 
		   ANY);

/* GC */
  swit_make_prim1 ("gc-expand-heap", gc_expand_prim,
		   NUM_M);
  swit_make_prim0 ("gc-collect", gcollect_prim);
  swit_make_prim0 ("gc-heap-size", gheap_size_prim);

/* Exceptions, errors */
  swit_make_prim12 ("raise", NULL, 0,0,0); /* in eval */
  swit_make_prim12 ("error", error_prim,
		    STRING_M, ANY,
		    swit_else);
  swit_make_prim2 ("cerror", NULL, 0, 0); /* in eval */

/* Time */
  swit_make_prim0 ("clock", clock_prim);
  swit_make_prim1 ("pause", select_prim,
		   NUM_M);
/* Mutable and other */
  swit_make_prim1("mutable->immutable", to_immutable_prim,
		  ANY);
  swit_make_prim1("immutable->mutable", to_mutable_prim,
		  ANY);
  swit_make_prim1("mutable?", mutable_p_prim,
		  ANY);
  swit_make_prim1("identity", identity_prim,
		  ANY);
  swit_make_prim0 ("version", version_prim);
}

