/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==============================================================
 *                      QUADRATIC BEZIER CURVES
 *==============================================================
 *
 *   bezier?              OBJ  -> BOOLEAN
 *   bezier-distance      VECTOR X POINT -> REAL
 *
 *==============================================================
 *                           TURTLE
 *==============================================================
 *   turtle?                   OBJ  -> BOOLEAN
 *   make-turtle
 *   turtle-save!
 *   turtle-restore!
 *   turtle-move!
 *   turtle-get-path
 *   turtle-get-position
 *   turtle-set-position!
 *   turtle-get-orientation
 *   turtle-set-orientation!
 *   turtle-heading         QUAD -> VECT3
 *   turtle-left!           REAL -> REAL
 *   turtle-right!          REAL -> REAL
 *   turtle-up!             REAL -> REAL
 *   turtle-down!           REAL -> REAL
 *   turtle-roll-left!      REAL -> REAL
 *   turtle-roll-right!     REAL -> REAL
 *
 *==============================================================
 *                  3D VISU TRANSFORMS
 *==============================================================
 *   world->norm            POINT -> POINT
 *   norm->world            POINT -> POINT
 *   set-visu!              POINT X ORIENT -> BOOLEAN
 *==============================================================
 */

#include "swit.h"
#include "quad.h"

#include <math.h>


/*==============================================================
 *                      QUADRATIC BEZIER CURVES
 *==============================================================
 */

static void *
wt ()
{
  exn = exn_type;
  prim_err_msg = wrong_type;
  return NULL;
}

/* quadratic bezier curves are defined with a vector of three quads. */
double
pts_distance (SWiT_Quad p1, SWiT_Quad p2)
{
  double w1 = QUAD_W(p1);
  double w2 = QUAD_W(p2);
  double w = w1*w2;
  double x = QUAD_X(p1)*w2-QUAD_X(p2)*w1;
  double y = QUAD_Y(p1)*w2-QUAD_Y(p2)*w1;
  double z = QUAD_Z(p1)*w2-QUAD_Z(p2)*w1;
  return sqrt(x*x+y*y+z*z)/w;
}

static double
bz_aux (SWiT_Quad q0, SWiT_Quad q1, SWiT_Quad q2,
	SWiT_Quad q, double d1, double d2, double eps)
{
  if (abs(d1 - d2) < eps) {
    return (d1+d2)/2;
  } else {
    SWiT_Quad q01 = pt_line (q0, q1, 0.5);
    SWiT_Quad q12 = pt_line (q1, q2, 0.5);
    SWiT_Quad q012 = pt_line (q01, q12, 0.5);
    double d0 = pts_distance (q012, q);
    if (d1 < d2) {
      return bz_aux (q0, q01, q012, q, d1, d0, eps);
    } else {
      return bz_aux (q012, q12, q2, q, d0, d2, eps);
    }
  }
}

static int
bezierp (SWiT b)
{
  if (SWIT_VECTORP(b) && (VECTOR_LENGTH(b) == 3)) {
    SWiT *elts = VECTOR_ELS(b);
    return (SWIT_QUADP(elts[0]) && SWIT_QUADP(elts[1])
	    && SWIT_QUADP(elts[2]));
  } else {
    return 0;
  }
}

static SWiT
bezier_p_prim (SWiT arg)
{
  return (bezierp(arg) ? swit_true : swit_false);
}

static SWiT
bezier_distance_prim (SWiT arg1, SWiT arg2)
{
  if (bezierp(arg1)) {
    SWiT_Quad q = SWIT_QUAD(arg2);
    SWiT *elts = VECTOR_ELS(arg1);
    SWiT_Quad q0 = SWIT_QUAD(elts[0]);
    SWiT_Quad q2 = SWIT_QUAD(elts[2]);
    double d1 = pts_distance (q0, q);
    double d2 = pts_distance (q2, q);
    return swit_make_inexact(bz_aux (q0, SWIT_QUAD(elts[1]),
				     SWIT_QUAD(elts[2]), q, d1, d2,
				     (d1+d2)/100));
  } else {
    prim_err_msg = must_be_bezier;
    exn = exn_type;
    return NULL;
  }
}



/*=====================================================
 *                  (3D) TURTLE GRAPHICS
 *=====================================================
 */

#define SWIT_TURTLE(obj) ((struct SWiT_Turtle *)((obj)->data))
#define TURTLE_ORIENT(obj) ((obj)->orient)
#define TURTLE_PATH(obj) ((obj)->path)
#define TURTLE_STACK(obj) ((obj)->stack)
#define TURTLE_POSITION(obj) SWIT_CAR(TURTLE_PATH(obj))

struct SWiT_Turtle
{
  SWiT orient;         /* a quaternion defining a rotation to apply
                          to world frame                             */
  SWiT path;           /* list of ``continuous'' positions yet
                          to be used (for drawing or moving frames).
			  This ``path'' is lost when the turtle
                          jumps (turtle-set-position or restore from
                          stack. */
  SWiT stack;          /* stack of (pos . orient)                    */
};
typedef struct SWiT_Turtle *SWiT_Turtle;

/*-----------------------------------------------------------------
 *                            PRINT
 *-----------------------------------------------------------------
 */
static void
print_turtle (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  SWiT_Turtle t = SWIT_TURTLE(obj);
  depth++;
  str2buff("#<turtle ");
  print_to_buffer(TURTLE_POSITION(t),esc, tcl, fv, depth);
  space2buff();
  print_to_buffer(TURTLE_ORIENT(t),esc, tcl, fv, depth);
  char2buff('>');
}

/*============================================================
 *                            TYPE
 *============================================================
 */
SWiT swit_turtle;

static SWiT
turtle_p_prim (SWiT arg)
{
  return (SWIT_TURTLEP(arg) ? swit_true : swit_false);
}


static SWiT
make_turtle (int argc, SWiT *argv)
{
  SWiT obj = swit_alloc_object();
  SWiT_Turtle turtle = SWIT_ALLOC(Turtle);
  SWiT default_position = argc > 0 ? argv[0]
    : swit_make_quad(make_quad(0, 0, 0, 1));
  SWiT default_orient = argc > 1 ? argv[1]
    : swit_make_quad(make_quad(0, 0, 0, 1.4142));
  TURTLE_ORIENT(turtle) = default_orient;
  if (argc > 2) {
    prim_err_msg = too_many_args;
    exn = exn_arity;
    return NULL;
  }
  TURTLE_STACK(turtle) = swit_null_list;
  TURTLE_PATH(turtle) = swit_make_pair(default_position, swit_null_list);
  SWIT_TYPE(obj) = TURTLE;
  SWIT_TURTLE (obj) = turtle;
  return (obj);
}

/* position & orientation operations */
static SWiT
turtle_set_position_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg1);
  /* check position (arg2) is at finite distance ? */ 
  TURTLE_PATH(turtle) = swit_make_pair(arg2, swit_null_list);
  return arg2;
}

static SWiT
turtle_get_position_prim (SWiT arg)
{
  return TURTLE_POSITION(SWIT_TURTLE(arg));
}

static SWiT
turtle_set_orient_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg1);
  TURTLE_ORIENT(turtle) = arg2;
  return arg2;
}

static SWiT
turtle_get_orient_prim (SWiT arg)
{
  return TURTLE_ORIENT(SWIT_TURTLE(arg));
}

/* path operations */
static SWiT
turtle_move_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg1);
  SWiT_Quad q = SWIT_QUAD(TURTLE_ORIENT(turtle));
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  SWiT_Quad p = SWIT_QUAD(TURTLE_POSITION(turtle));
  double d = to_double(arg2);
  SWiT new_position = swit_make_quad(make_quad(QUAD_X(p)+ d*(x*y-z*w),
					       QUAD_Y(p)+ d*(1-x*x-z*z),
					       QUAD_Z(p)+ d*(x*w+z*y),
					       QUAD_W(p)));
  TURTLE_PATH(turtle) = swit_make_pair(new_position, TURTLE_PATH(turtle));
  return new_position;
}

static SWiT
turtle_clear_path_prim (SWiT arg)
{
  SET_CDR(TURTLE_PATH(SWIT_TURTLE(arg)), swit_null_list);
  return arg;
}

static SWiT
turtle_get_path_prim (SWiT arg)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg);
  return TURTLE_PATH(turtle);
}

/* Stack operations */
static SWiT
turtle_save_prim (SWiT arg)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg);
  TURTLE_STACK(turtle) =
    swit_make_pair(swit_make_pair(quad_copy_prim(TURTLE_POSITION(turtle)),
				  quad_copy_prim(TURTLE_ORIENT(turtle))),
		   TURTLE_STACK(turtle));
  return arg;
}

static SWiT
turtle_restore_prim (SWiT arg)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg);
  SWiT stack = TURTLE_STACK(turtle);
  if (SWIT_PAIRP(stack)) {
    SWiT saved = SWIT_CAR(stack);
    SWiT position = SWIT_CAR(saved);
    TURTLE_ORIENT(turtle) = SWIT_CDR(saved);
    SET_CDR(saved, swit_null_list);
    TURTLE_STACK(turtle) = SWIT_CDR(stack);
    TURTLE_PATH(turtle) = saved;
    return position;
  } else {
    prim_err_msg = stack_empty;
    exn = exn_type;
    return NULL;
  }
}

static SWiT
turtle_clear_stack_prim (SWiT arg)
{
  TURTLE_STACK(SWIT_TURTLE(arg)) = swit_null_list;
  return arg;
}

static SWiT
turtle_get_stack_prim (SWiT arg)
{
  return TURTLE_STACK(SWIT_TURTLE(arg));
}

/*-----------------------------------------------------------------
 *  Compute the "heading" of the turtle from the current orientation
 *  of the turtle (a quaternion), the heading is the vector J = q*j*q'
 *-----------------------------------------------------------------
 */
static SWiT
heading_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Turtle turtle = SWIT_TURTLE(arg1);
  SWiT_Quad q = SWIT_QUAD(TURTLE_ORIENT(turtle));
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  if (arg2 == swit_zero) {
    return swit_make_quad(make_quad(x*y-z*w,1-x*x-z*z,x*w+z*y,0));
  } else {
    SWiT_Quad p = SWIT_QUAD(TURTLE_POSITION(turtle));
    double d = to_double(arg2);
    return swit_make_quad(make_quad(QUAD_X(p)+ d*(x*y-z*w),
			  QUAD_Y(p)+ d*(1-x*x-z*z),
			  QUAD_Z(p)+ d*(x*w+z*y), QUAD_W(p)));
  }
}

static double deg2rad2 = 0.008726646; /* pi/360 */

static void
l_or_r (SWiT orient, double delta2)
{
  SWiT_Quad q = SWIT_QUAD(orient);
  double c = cos(delta2);
  double s = sin(delta2);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  QUAD_X(q) = x*c+y*s;
  QUAD_Y(q) = y*c-x*s;
  QUAD_Z(q) = z*c+w*s;
  QUAD_W(q) = w*c-z*s;
}

static SWiT
left_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  l_or_r(orient, delta2);
  return orient;
}

static SWiT
right_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  l_or_r(orient, -delta2);
  return orient;
}

static void
u_or_d (SWiT orient, double delta2)
{
  SWiT_Quad q = SWIT_QUAD(orient);
  double c = cos(delta2);
  double s = sin(delta2);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  QUAD_X(q) = x*c+w*s;
  QUAD_Y(q) = y*c+z*s;
  QUAD_Z(q) = z*c-y*s;
  QUAD_W(q) = w*c-x*s;
}

static SWiT
up_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  u_or_d(orient, delta2);
  return orient;
}

static SWiT
down_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  u_or_d(orient, -delta2);
  return orient;
}

static void
roll (SWiT orient, double delta2)
{
  SWiT_Quad q = SWIT_QUAD(orient);
  double c = cos(delta2);
  double s = sin(delta2);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  QUAD_X(q) = x*c-z*s;
  QUAD_Y(q) = y*c+w*s;
  QUAD_Z(q) = z*c+x*s;
  QUAD_W(q) = w*c-y*s;
}

static SWiT
rleft_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  roll(orient, delta2);
  return orient;
}

static SWiT
rright_prim (SWiT arg1, SWiT arg2)
{
  SWiT orient = TURTLE_ORIENT(SWIT_TURTLE(arg1));
  double delta2 = deg2rad2 * to_double(arg2);
  roll(orient, -delta2);
  return orient;
}



/*=====================================================
 *                  3D VISUALISATION
 *=====================================================
 */

static double visu[16];/* 11 for visu matrix, eye, x0, y0, z0, focal */
/*
;;;       [ ct            0           st             x0*ct + z0*st           ]
;;;   M'= [ sp*st        -cp         -sp*ct         -sp*A - y0*cp            ]
;;;       [-cp*st/e      -sp/e        cp*ct/e       (cp*A - y0*sp)/e         ]
;;;       [cp*st/(f*e)    sp/(f*e)   -cp*ct/(f*e)  (-cp*A + y0*sp)/(f*e) + 1/f]
;;;
;;;  with A = -x0*st+z0*ct
;;; (the last line is not stored)
*/
static void
set_visu (double x0, double y0, double z0,
	  double theta, double phi, double eye, double focal)
{
  double ct = cos(theta);
  double st = sin(theta);
  double cp = cos(phi);
  double sp = sin(phi);
  double a = z0*ct-x0*st;
  visu[0] = ct;
  visu[1] = st;
  visu[2] = x0*ct+z0*st;
  visu[3] = sp*st;
  visu[4] = -cp;
  visu[5] = -sp*ct;
  visu[6] = -sp*a - cp*y0;
  visu[7] = -cp*st/eye;
  visu[8] = -sp/eye;
  visu[9] = ct*cp/eye;
  visu[10] = (a*cp - y0*sp)/eye;
  visu[11] = eye;
  visu[12] = x0;
  visu[13] = y0;
  visu[14] = z0;
  visu[15] = focal;
}

static SWiT
set_visu_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad orig = SWIT_QUAD(arg1);
  if (QUAD_W(orig) == 0) {
    return wt();
  } else {
    SWiT_Quad persp = SWIT_QUAD(arg2);
    double w = QUAD_W(orig);
    double eye = QUAD_Z(persp);
    double focal = QUAD_W(persp);
    if ((eye == 0) || (focal == 0)) {
      prim_err_msg = non_zero;
      exn = exn_math;
      return NULL;
    }
    set_visu(QUAD_X(orig)/w, QUAD_Y(orig)/w, QUAD_Z(orig)/w,
	     QUAD_X(persp), QUAD_Y(persp), eye, focal);
    return swit_true; /* or a visu `hvect', when available */
  }
}

static SWiT
world_to_norm (double x, double y, double z, double w)
{
  double zn = visu[7]*x + visu[8]*y + visu[9]*z + visu[10]*w;
  return
    swit_make_quad(make_quad(visu[0]*x + visu[1]*z + visu[2]*w,
			     visu[3]*x + visu[4]*y + visu[5]*z + visu[6]*w,
			     zn, (w-zn)/visu[15]));
}

static SWiT
world_to_norm_prim (SWiT arg)
{
  SWiT_Quad q = SWIT_QUAD(arg);
  return world_to_norm(QUAD_X(q), QUAD_Y(q), QUAD_Z(q), QUAD_W(q));
}

/*
;;;       [ct   sp*st   -cp*e*st - x0   -x0*f ]
;;;   M = [0   -cp      -e*sp - y0      -y0*f ]
;;;       [st  -sp*ct    cp*e*ct - z0   -z0*f ]
;;;       [0    0        1               f    ]
;;;
;;;  (M = T*RT*RP*P)
  visu[0] = ct;
  visu[1] = st;
  visu[2] = x0*ct+z0*st;
  visu[3] = sp*st;
  visu[4] = -cp;
  visu[5] = -sp*ct;
  visu[6] = -sp*a - cp*y0;
  visu[7] = -cp*st/eye;
  visu[8] = -sp/eye;
  visu[9] = ct*cp/eye;
  visu[10] = (a*cp - y0*sp)/eye;
  visu[11] = eye;
  visu[12] = x0;
  visu[13] = y0;
  visu[14] = z0;
  visu[15] = focal;
*/

static SWiT
norm_to_world (double x, double y, double z, double w)
{
  double ct = visu[0];
  double st = visu[1];
  double mcp = visu[4];
  double e = visu[11];
  double zw = z+w*visu[15];
  return swit_make_quad(make_quad(
	     ct*x + visu[3]*y + mcp*st*e*z - visu[12]*zw,
	     mcp*y + e*e*visu[8]*z - visu[13]*zw,
	     st*x + visu[5]*y - e*ct*mcp*z - visu[14]*zw,
	     zw));
}


static SWiT
norm_to_world_prim (SWiT arg)
{
  SWiT_Quad q = SWIT_QUAD(arg);
  return norm_to_world(QUAD_X(q), QUAD_Y(q), QUAD_Z(q), QUAD_W(q));
}



/*=====================================================
 *                 INITIALISATION
 *=====================================================
 */
void
swit_init_model (void)
{
  swit_turtle = make_type("<turtle>");
  print_table[TURTLE>>1] = print_turtle;
  set_visu(0.0,0.0,0.0,
	   0.0,0.0,300.0,4.0);

  /* Bezier */
  swit_make_prim1("bezier?", bezier_p_prim,
		  ANY);
  swit_make_prim2("bezier-distance", bezier_distance_prim,
		  VECT_M, QUAD_M);

  /* turtle */
  swit_make_prim1("turtle?", turtle_p_prim,
		  ANY);
  swit_make_prim0n("make-turtle", make_turtle,
		   ANY, ANY);
  swit_make_prim1("turtle-get-position", turtle_get_position_prim,
		  TURTLE);
  swit_make_prim2("turtle-set-position!", turtle_set_position_prim,
		  TURTLE, QUAD_M);
  swit_make_prim1("turtle-get-orientation", turtle_get_orient_prim,
		  TURTLE);
  swit_make_prim2("turtle-set-orientation!", turtle_set_orient_prim,
		  TURTLE, QUAD_M);
  swit_make_prim2("turtle-move!", turtle_move_prim,
		  TURTLE, NUM_M);
  swit_make_prim1("turtle-get-path", turtle_get_path_prim,
		  TURTLE);
  swit_make_prim1("turtle-clear-path!", turtle_clear_path_prim,
		  TURTLE);
  swit_make_prim2("turtle-left!", left_prim,
		  TURTLE, NUM_M);
  swit_make_prim2("turtle-right!", right_prim,
		  TURTLE, NUM_M);
  swit_make_prim2("turtle-up!", up_prim,
		  TURTLE, NUM_M);
  swit_make_prim2("turtle-down!", down_prim,
		  TURTLE, NUM_M);
  swit_make_prim2("turtle-roll-left!", rleft_prim,
		  TURTLE, NUM_M);
  swit_make_prim2("turtle-roll-right!", rright_prim,
		  TURTLE, NUM_M);
  swit_make_prim12("turtle-heading", heading_prim,
		   TURTLE, NUM_M, swit_zero);
  swit_make_prim1("turtle-save!", turtle_save_prim,
		  TURTLE);
  swit_make_prim1("turtle-restore!", turtle_restore_prim,
		  TURTLE);
  swit_make_prim1("turtle-get-stack", turtle_get_stack_prim,
		  TURTLE);
  swit_make_prim1("turtle-clear-stack!", turtle_clear_stack_prim,
		  TURTLE);

  /* 3D visu transforms */
  swit_make_prim1("world->norm", world_to_norm_prim,
		  QUAD_M);
  swit_make_prim1("norm->world", norm_to_world_prim,
		  QUAD_M);
  swit_make_prim2("set-visu!", set_visu_prim,
		  QUAD_M, QUAD_M);
}
