/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*---------------------------------------------------------------------
 *                    6.5 NUMBERS
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *   number?                    NUMBER -> BOOLEAN
 *   complex?                   NUMBER -> BOOLEAN
 *   real?                      NUMBER -> BOOLEAN
 *   rational?                  NUMBER -> BOOLEAN
 *   integer?                   NUMBER -> BOOLEAN
 *   exact?                     NUMBER -> BOOLEAN
 *   inexact?                   NUMBER -> BOOLEAN
 *   =, < , > , <= , >=         NUMBER * NUMBER * NUMBER ... -> BOOLEAN
 *   zero?, positive?, negative?, odd?, even?  NUMBER -> BOOLEAN
 *   max 1+
 *   min 1+
 *   + 0+
 *   - 1+
 *   * 1+
 *   / 1+
 *   abs 1
 *   quotient 2
 *   remainder 2
 *   modulo 2
 *   gcd 0+
 *   lcm 0+
 *   (numerator : not implemented)
 *   (denominator : not implemented)
 *   floor, ceiling, truncate, round 1
 *   (rationalize : not implemented)
 *   (complex : not implemented)
 *   sqrt,sin,cos,log,tan,asin,acos 1
 *   expt 2
 *   atan (1)2
 *   exact->inexact 1
 *   inexact->exact 1
 *   number->string (1)2
 *   string->number (1)2
 *
 * Extensions :
 * ^^^^^^^^^^
 * Cube root :
 *  cbrt
 * More transcendental functions :
 *  log10
 *  expm1
 *  log1p
 *  sinh
 *  cosh
 *  tanh
 *  asinh
 *  acosh
 *  atanh
 * Random numbers :
 *  random
 *  srandom
 * Logical functions on integers (TODO !!):
 *  log-or
 *  log-xor
 *  log-and
 *  log-not
 *  log-lsh
 *  log-rsh
 *  log-comp
 *---------------------------------------------------------------------
 */
#include "swit.h"
#include <math.h>
#include <errno.h>

/* Extported to Print (for displaying optional argument) */
SWiT swit_ten;

/* "General" predicates <obj> -> <bool> */
static SWiT number_p_prim (SWiT arg);
static SWiT complex_p_prim (SWiT arg);
static SWiT real_p_prim (SWiT arg);
static SWiT rational_p_prim (SWiT arg);
static SWiT integer_p_prim (SWiT arg);
static SWiT fixnum_p_prim (SWiT arg);
static SWiT bignum_p_prim (SWiT arg);
static SWiT exact_p_prim (SWiT arg);
static SWiT inexact_p_prim (SWiT arg);

/* Predicates <int> -> <bool> */
static SWiT odd_p_prim (SWiT arg);
static SWiT even_p_prim (SWiT arg);
static SWiT prime_p_prim (SWiT arg);

/* Overloaded comparison predicates <num> x ... -> <bool> */
static SWiT zero_p_prim (SWiT arg);
static SWiT positive_p_prim (SWiT arg);
static SWiT negative_p_prim (SWiT arg);
int bin_eq (SWiT n1, SWiT n2); /* exported for eqv? (file "switEquiv.c") */
static int bin_lt (SWiT n1, SWiT n2);
static int bin_gt (SWiT n1, SWiT n2);
static int bin_le (SWiT n1, SWiT n2);
static int bin_ge (SWiT n1, SWiT n2);
static SWiT eq_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT lt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT gt_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT le_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT ge_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);

/*
 * Overloaded operators T x ... -> T
 *  (the type of the result is the same as the type of the arguments)
 */
static SWiT bin_add (SWiT arg1, SWiT arg2);
static SWiT bin_mul (SWiT arg1, SWiT arg2);
static SWiT bin_sub (SWiT arg1, SWiT arg2);
static SWiT bin_div (SWiT arg1, SWiT arg2);
static SWiT bin_max (SWiT arg1, SWiT arg2);
static SWiT bin_min (SWiT arg1, SWiT arg2);
static SWiT add_prim (int argc, SWiT *argv);
static SWiT sub_prim (SWiT arg, int argc, SWiT *argv);
static SWiT mul_prim (int argc, SWiT *argv);
static SWiT div_prim (SWiT arg, int argc, SWiT *argv);
static SWiT max_prim (SWiT arg, int argc, SWiT *argv);
static SWiT min_prim (SWiT arg, int argc, SWiT *argv);
static SWiT abs_prim (SWiT arg);
static SWiT expt_prim (SWiT arg1, SWiT arg2);
static SWiT floor_prim (SWiT arg);
static SWiT ceiling_prim (SWiT arg);
static SWiT truncate_prim (SWiT arg);
static SWiT round_prim (SWiT arg);

/*
 * Functions |N x ... -> |N
 *  (returns an inexact result if one of their argument is inexact)
 */
static SWiT bin_gcd (SWiT n1, SWiT n2);
static SWiT bin_lcm (SWiT n1, SWiT n2);
static SWiT gcd_prim (int argc, SWiT *argv);
static SWiT lcm_prim (int argc, SWiT *argv);
static SWiT quotient_prim (SWiT arg1, SWiT arg2);
static SWiT rem_prim (SWiT arg1, SWiT arg2);
static SWiT modulo_prim (SWiT arg1, SWiT arg2);
static SWiT fact_prim (SWiT arg);

/* Functions |R -> |R (always returns an inexact result) */
static SWiT exp_prim (SWiT arg);
static SWiT log_prim (SWiT arg);
static SWiT sin_prim (SWiT arg);
static SWiT cos_prim (SWiT arg);
static SWiT tan_prim (SWiT arg);
static SWiT asin_prim (SWiT arg);
static SWiT acos_prim (SWiT arg);
static SWiT cbrt_prim (SWiT arg);
static SWiT log10_prim (SWiT arg);
static SWiT expm1_prim (SWiT arg);
static SWiT log1p_prim (SWiT arg);
static SWiT sinh_prim (SWiT arg);
static SWiT cosh_prim (SWiT arg);
static SWiT tanh_prim (SWiT arg);
static SWiT asinh_prim (SWiT arg);
static SWiT acosh_prim (SWiT arg);
static SWiT atanh_prim (SWiT arg);

/* Functions |R x |R -> |R (always returns an inexact result) */
static SWiT atan2_prim (SWiT arg1, SWiT arg2);
static SWiT hypot_prim (SWiT arg1, SWiT arg2);

/*
 * Function that can return an exact result
 *   when its argument is exact (but not always).
 */
static SWiT sqrt_prim (SWiT arg);

/* Misc. */
static SWiT rand_prim (void);
static SWiT srand_prim (SWiT arg);
static SWiT exact_to_inexact (SWiT arg);
static SWiT inexact_to_exact (SWiT arg);
static SWiT string_to_number_prim (SWiT arg1, SWiT arg2);
static SWiT fixnum_to_string (int i, int radix);
static SWiT bignum_to_string (MP_INT *i, int radix);
static SWiT inexact_to_string (double d);

/*========================
 *      PRINTING NUMBERS
 *========================
 */

void
print_inexact (double x)
{
  int i, len, no_pt = 1, no_exp = 1;
  char numstr[32];
  sprintf (numstr, "%1.10g", x);
  /* add a '.0' if neither decimal point nor exponent mark : */
  len = strlen (numstr);
  for (i = 0; (i < len) && (no_pt && no_exp); i++) {
    no_pt = (numstr[i] == '.') ? 0 : 1;
    no_exp = ((numstr[i] == 'e') || (numstr[i] == 'n')) ? 0 : 1;
  }
  numstr[len] = '\0';
  str2buff(numstr);
  if (no_pt && no_exp) str2buff(".0");
}

static void
print_number (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  switch (SWIT_TYPE(obj) & OBJ)
    {    
    case BIGNUM: {
      char *numstr = mpz_get_str(NULL, 10, SWIT_BIGNUM(obj));
      str2buff(numstr);
      break;
    }
    case  FIXNUM: {
      char numstr[10];
      sprintf (numstr, "%ld", SWIT_FIXNUM(obj));
      str2buff(numstr);
      break;
    }
    case INEXACT:
      print_inexact(SWIT_DBL_VAL(obj));
      break;
    }
}      

/* exported to switTk.c */
SWiT number_to_string (SWiT arg1, SWiT arg2);

/* Some conversion functions */
SWiT int_to_integer (int n);  /* exported */
static SWiT big_to_integer (MP_INT *x);
static SWiT inexact_to_integer (double r);

SWiT
swit_make_bignum (MP_INT *big)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = BIGNUM;
  (MP_INT *)(obj->data) = big;
  return (obj);
}

SWiT
swit_make_fixnum (int fix)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = FIXNUM;
  (int)(obj->data) = fix;
  return (obj);
}


SWiT
swit_make_inexact (double d)
{
  SWiT obj = swit_alloc_object();
  double *dbl = (double *)swit_malloc(sizeof(double));
  SWIT_TYPE(obj) = INEXACT;
  *dbl = d ;
  SWIT_DBL(obj) = dbl;
  return (obj);
}


/*----------------------------------------------------------------------
 *                   Conversions functions
 *----------------------------------------------------------------------
 */

double /* exported to switLex.c and to switQuad.c */
big_to_inexact(MP_INT *x)
{
  char *s = mpz_get_str(NULL, 10, x);
  double res = atof(s);
  return (res);
}

SWiT
int_to_integer (int n)
{
  unsigned long int m = abs(n);
  if (m >> 15) { 
    MP_INT *x = (MP_INT *)swit_malloc(sizeof(MP_INT));
    mpz_init_set_si(x, n);
    return swit_make_bignum(x);
  } else {
    return swit_make_fixnum (n);
  }
}

static SWiT
big_to_integer (MP_INT *x)
{
  if (mpz_size(x) > 1)
    return (swit_make_bignum(x));
  else {
    unsigned long int m = mpz_get_ui(x);
    if (m >> 15) {
      return swit_make_bignum(x);
    } else {
      return swit_make_fixnum (mpz_get_si(x));
    }
  }
}

/*
 * The first 16 decimals of r are kept,
 *  (the first 15 may be right, others are not significant).
 * It is slow when the number is large.
 */
static SWiT
inexact_to_integer (double r)
{
  MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
  unsigned long int e = log10(fabs(r));
  if (e>8) {
    double y,p;
    int n;
    char s[32],s1[16];
    e -= 8;
    p = pow(10,e);
    y = r/p;
    n = (int)floor(y);
    if ( e > 6 ) {
      p = 1e+7;
      e = e-7;
    } else {
      e = 0;
    }
    sprintf(s,"%d",n);
    sprintf(s1,"%d",(int)rint((y - n+1) * p));
    strcat(s,s1+1);
    if (e) {
      MP_INT tmp1,tmp2;
      char *s2 = s;
      mpz_init(res);
      mpz_init_set_ui(&tmp1, 10);
      mpz_init(&tmp2);
      mpz_pow_ui(&tmp2,&tmp1,e);
      mpz_init_set_str(&tmp1, s2, 10);
      mpz_mul(res, &tmp1, &tmp2);
      mpz_clear(&tmp1);
      mpz_clear(&tmp2);
    } else {
      mpz_init_set_str(res, s, 10);
    }
  } else {
    mpz_init_set_si(res, (int)rint(r));
  }
  return (big_to_integer(res));
}


/*---------------------------------------------------------------
 *   General predicates <obj> -> <bool>
 *---------------------------------------------------------------
 */

static SWiT
number_p_prim (SWiT arg)
{
  return (SWIT_NUMBERP(arg) ? swit_true : swit_false);
}

static SWiT
complex_p_prim (SWiT arg)
{
  return (SWIT_NUMBERP(arg) ? swit_true : swit_false);
}

static SWiT
real_p_prim (SWiT arg)
{
  return (SWIT_NUMBERP(arg) ? swit_true : swit_false);
}

/* inline */
int
isinteger(double x) 
{
  return(x == floor(x));
}

 
static int
integer_p (SWiT arg)
{
  return (SWIT_INTP(arg) ? 1
	  : (SWIT_INEXACTP(arg) && isinteger(SWIT_DBL_VAL(arg)))
	  ? 1 : 0);
}

static SWiT
integer_p_prim (SWiT arg)
{
  return (integer_p(arg) ? swit_true : swit_false);
}

static SWiT
fixnum_p_prim (SWiT arg)
{
  return (SWIT_FIXNUMP(arg) ? swit_true : swit_false);
}

static SWiT
bignum_p_prim (SWiT arg)
{
  return (SWIT_BIGNUMP(arg) ? swit_true : swit_false);
}

static SWiT
rational_p_prim (SWiT arg)
{
  return (SWIT_EXACTP(arg) ? swit_true : swit_false);
}

static SWiT
exact_p_prim (SWiT arg)
{
  return (SWIT_INTP(arg) ? swit_true : swit_false);
}

static SWiT
inexact_p_prim (SWiT arg)
{
  return (SWIT_INEXACTP(arg) ? swit_true : swit_false);
}

/*---------------------------------------------------------------
 *   Overloaded comparison predicates <num> x ... -> <bool>
 *---------------------------------------------------------------
 */
#define GEN_BIN_COMP(name, op) \
int \
bin_ ## name (SWiT n1, SWiT n2) \
{ \
   if (SWIT_FIXNUMP(n1)) { \
      if (SWIT_FIXNUMP(n2)) \
        return (SWIT_FIXNUM(n1) op SWIT_FIXNUM(n2)); \
      else if (SWIT_INEXACTP(n2)) \
        return (SWIT_FIXNUM(n1) op SWIT_DBL_VAL(n2)); \
      else /* (SWIT_BIGNUMP(n2)) */ \
        return (0 op mpz_cmp_si(SWIT_BIGNUM(n2),SWIT_FIXNUM(n1)));\
   } else if (SWIT_INEXACTP(n1)) { \
      if (SWIT_FIXNUMP(n2)) \
        return (SWIT_DBL_VAL(n1) op SWIT_FIXNUM(n2)); \
      else if (SWIT_INEXACTP(n2)) \
        return (SWIT_DBL_VAL(n1) op SWIT_DBL_VAL(n2)); \
      else /* (SWIT_BIGNUMP(n2)) */ \
        return (SWIT_DBL_VAL(n1) op big_to_inexact(SWIT_BIGNUM(n2)));\
   } else /* (SWIT_BIGNUMP(n1)) */ { \
      if (SWIT_FIXNUMP(n2)) \
        return (mpz_cmp_si(SWIT_BIGNUM(n1),SWIT_FIXNUM(n2)) op 0);\
      else if (SWIT_INEXACTP(n2)) \
        return (big_to_inexact(SWIT_BIGNUM(n1)) op SWIT_DBL_VAL(n2));\
      else /* (SWIT_BIGNUMP(n2)) */ \
	return (mpz_cmp(SWIT_BIGNUM(n1),SWIT_BIGNUM(n2)) op 0);\
   } \
}

GEN_BIN_COMP(eq, ==)         /* exported for eqv? */
static GEN_BIN_COMP(lt, <)
static GEN_BIN_COMP(gt, >)
static GEN_BIN_COMP(le, <=)
static GEN_BIN_COMP(ge, >=)

#define GEN_NARY_COMP(name) \
static SWiT  \
name ## _prim(SWiT arg1, SWiT arg2, int argc, SWiT *argv) \
{ \
   if (! bin_ ## name(arg1, arg2)) { \
       return (swit_false); \
   } else if (argc) { \
      SWiT x = arg2;\
      int i; \
      for (i=0;i<argc;i++) { \
         SWiT y = x;\
         x = argv[i];\
         if (! bin_ ## name(y,x)) { \
             return (swit_false); \
         } \
      } \
   } \
   return (swit_true); \
}

GEN_NARY_COMP(eq)
GEN_NARY_COMP(lt)
GEN_NARY_COMP(gt)
GEN_NARY_COMP(le)
GEN_NARY_COMP(ge)


#define GEN_SIGNP(name, op) \
static SWiT \
name ## _p_prim (SWiT arg)\
{\
    if (SWIT_FIXNUMP(arg)) {\
	return (SWIT_FIXNUM(arg) op 0 ? swit_true : swit_false);\
    } else if (SWIT_INEXACTP(arg)) {\
	return (SWIT_DBL_VAL(arg) op 0 ? swit_true : swit_false);\
    } else /* (SWIT_BIGNUMP(arg)) */ {\
	return (mpz_cmp_si(SWIT_BIGNUM(arg),0) op 0 ? swit_true : swit_false);\
    }\
}

GEN_SIGNP(zero, ==)
GEN_SIGNP(positive, >)
GEN_SIGNP(negative, <)


/*-------------------------------------------------------------------
 *        Predicates |N -> <bool>
 *-------------------------------------------------------------------
 */
static SWiT
odd_p_prim (SWiT arg)
{
  if (SWIT_FIXNUMP(arg)) {
    int n = SWIT_FIXNUM(arg);
    return (n % 2 != 0 ? swit_true : swit_false);
  } else if (SWIT_BIGNUMP(arg)) {
    int n = mpz_get_si(SWIT_BIGNUM(arg));
    return (n % 2 != 0 ? swit_true : swit_false);
  } else {
    double dbl = SWIT_DBL_VAL(arg);
    if (dbl == floor(dbl)) {
      return odd_p_prim(inexact_to_integer(dbl));
    } else {
      exn = exn_type;
      prim_err_msg = wrong_type;
      exn_val = arg;
      return NULL;
    }
  }
}

static SWiT
even_p_prim (SWiT arg)
{
  if (SWIT_FIXNUMP(arg)) {
    int n = SWIT_FIXNUM(arg);
    return (n % 2 == 0 ? swit_true : swit_false);
  } else if (SWIT_BIGNUMP(arg)) {
    int n = mpz_get_si(SWIT_BIGNUM(arg));
    return (n % 2 == 0 ? swit_true : swit_false);
  } else /* inexact integer */ {
    double dbl = SWIT_DBL_VAL(arg);
    if (dbl == floor(dbl)) {
      return even_p_prim(inexact_to_integer(dbl));
    } else {
      exn = exn_type;
      prim_err_msg = wrong_type;
      exn_val = arg;
      return NULL;
    }
  }
}

static SWiT
prime_p_prim (SWiT arg)
{
  if SWIT_FIXNUMP(arg) {
    MP_INT tmp;
    int res;
    mpz_init_set_si(&tmp, SWIT_FIXNUM(arg));
    res = mpz_probab_prime_p(&tmp, 25);
    mpz_clear(&tmp);
    return (res ? swit_true : swit_false);
  } else if SWIT_BIGNUMP(arg) {
    int n = mpz_probab_prime_p(SWIT_BIGNUM(arg), 25);
    return (n ? swit_true : swit_false);
  } else /* mustbe an inexact integer */ {
    double dbl = SWIT_DBL_VAL(arg);
    if (dbl == floor(dbl)) {
      return prime_p_prim(inexact_to_integer(dbl));
    } else {
      exn = exn_type;
      prim_err_msg = wrong_type;
      exn_val = arg;
      return NULL;
    }
  }
}


/*---------------------------------------------------------------
 *        Overloaded functions T x ... -> T
 *---------------------------------------------------------------
 */
#define GEN_BIN_OP(name,c_op) \
static SWiT \
bin_ ## name (SWiT arg1, SWiT arg2) \
{ \
  if SWIT_FIXNUMP(arg1) { \
    if SWIT_FIXNUMP(arg2) \
      return (int_to_integer(SWIT_FIXNUM(arg1) c_op SWIT_FIXNUM(arg2))); \
    else if SWIT_INEXACTP(arg2) \
      return (swit_make_inexact(SWIT_FIXNUM(arg1) c_op SWIT_DBL_VAL(arg2))); \
    else /* (SWIT_BIGNUMP(arg2)) */ { \
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); \
      int n = SWIT_FIXNUM(arg1); \
      mpz_init_set_si(res, n); \
      mpz_ ## name(res, res, SWIT_BIGNUM(arg2)); \
      return (big_to_integer(res)); \
    } \
  } else if SWIT_INEXACTP(arg1) { \
    if SWIT_FIXNUMP(arg2) { \
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) c_op SWIT_FIXNUM(arg2))); \
    } else if (SWIT_INEXACTP(arg2)) { \
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) c_op SWIT_DBL_VAL(arg2))); \
    } else /* (SWIT_BIGNUMP (arg2)) */ { \
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) c_op big_to_inexact(SWIT_BIGNUM(arg2)))); \
    } \
  } else /* (SWIT_BIGNUMP(arg1)) */ { \
    if SWIT_FIXNUMP(arg2) { \
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); \
      int n = SWIT_FIXNUM(arg2); \
      mpz_init_set_si(res, n); \
      mpz_ ## name(res, SWIT_BIGNUM(arg1), res); \
      return (big_to_integer(res)); \
    } else if SWIT_INEXACTP(arg2) { \
      return (swit_make_inexact (big_to_inexact(SWIT_BIGNUM(arg1)) c_op SWIT_DBL_VAL(arg2))); \
    } else /* (SWIT_BIGNUMP (arg2))*/ { \
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); \
      mpz_init(res); \
      mpz_ ## name(res, SWIT_BIGNUM(arg1), SWIT_BIGNUM(arg2)); \
      return (big_to_integer(res)); \
    } \
  } \
}

GEN_BIN_OP(add,+)
GEN_BIN_OP(sub,-)

static SWiT
bin_mul (SWiT arg1, SWiT arg2)
{ 
  if SWIT_FIXNUMP(arg1) { 
    int n1 = SWIT_FIXNUM(arg1);
    if SWIT_FIXNUMP(arg2) {
      int n2 = SWIT_FIXNUM(arg2);
      int n3 = n1 * n2;
      return (int_to_integer(n3)); 
      /*      if (n3 == 0) 
	      MP_INT *b1 = (MP_INT *)swit_malloc(sizeof(MP_INT)); 
	      MP_INT *b2 = (MP_INT *)swit_malloc(sizeof(MP_INT)); 
	      mpz_init_set_si(b1, n1); 
	      mpz_init_set_si(b2, n2); 
	      mpz_mul(b1, b2, b1); 
	      return (big_to_integer(b1)); 
	      } else {
	      return (int_to_integer(n3)); 
	      } 
      */
    } else if SWIT_INEXACTP(arg2) 
      return (swit_make_inexact(n1 * SWIT_DBL_VAL(arg2))); 
    else /* (SWIT_BIGNUMP(arg2)) */ { 
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); 
      int n = SWIT_FIXNUM(arg1); 
      mpz_init_set_si(res, n); 
      mpz_mul(res, res, SWIT_BIGNUM(arg2)); 
      return (big_to_integer(res)); 
    } 
  } else if SWIT_INEXACTP(arg1) { 
    if SWIT_FIXNUMP(arg2) { 
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) * SWIT_FIXNUM(arg2))); 
    } else if (SWIT_INEXACTP(arg2)) { 
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) * SWIT_DBL_VAL(arg2))); 
    } else /* (SWIT_BIGNUMP (arg2)) */ { 
      return (swit_make_inexact (SWIT_DBL_VAL(arg1) * big_to_inexact(SWIT_BIGNUM(arg2)))); 
    } 
  } else /* (SWIT_BIGNUMP(arg1)) */ { 
    if SWIT_FIXNUMP(arg2) { 
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); 
      int n = SWIT_FIXNUM(arg2); 
      mpz_init_set_si(res, n); 
      mpz_mul(res, SWIT_BIGNUM(arg1), res); 
      return (big_to_integer(res)); 
    } else if SWIT_INEXACTP(arg2) { 
      return (swit_make_inexact (big_to_inexact(SWIT_BIGNUM(arg1)) * SWIT_DBL_VAL(arg2))); 
    } else /* (SWIT_BIGNUMP (arg2))*/ { 
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)); 
      mpz_init(res); 
      mpz_mul(res, SWIT_BIGNUM(arg1), SWIT_BIGNUM(arg2)); 
      return (big_to_integer(res)); 
    } 
  } 
}

static SWiT
bin_div (SWiT arg1, SWiT arg2)
{
  if SWIT_FIXNUMP(arg1) {
    if SWIT_FIXNUMP(arg2) {
      int n1 = SWIT_FIXNUM(arg1), n2 = SWIT_FIXNUM(arg2);
      if (n1 % n2)
	return (swit_make_inexact((double)n1 / n2));
      else
	return (int_to_integer(n1 / n2));
    }
    else if SWIT_INEXACTP(arg2)
      return (swit_make_inexact(SWIT_FIXNUM(arg1) / SWIT_DBL_VAL(arg2)));
    else /* (SWIT_BIGNUMP (arg2)) */ {
      return (swit_make_inexact(SWIT_FIXNUM(arg1) / big_to_inexact(SWIT_BIGNUM(arg2))));
    }
  } else if SWIT_INEXACTP(arg1) {
    if SWIT_FIXNUMP(arg2) {
      return (swit_make_inexact(SWIT_DBL_VAL(arg1) / SWIT_FIXNUM(arg2)));
    } else if SWIT_INEXACTP(arg2) {
      return (swit_make_inexact(SWIT_DBL_VAL(arg1) / SWIT_DBL_VAL(arg2)));
    } else /* (SWIT_BIGNUMP (arg2)) */ {
      return (swit_make_inexact(SWIT_DBL_VAL(arg1) / big_to_inexact(SWIT_BIGNUM(arg2))));
    }
  } else /* (SWIT_BIGNUMP(arg1)) */ {
    if SWIT_FIXNUMP(arg2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT)), rem;
      int n = SWIT_FIXNUM(arg2);
      mpz_init_set_si(res, n);
      mpz_init(&rem);
      mpz_divmod(res, &rem, SWIT_BIGNUM(arg1), res);
      if (mpz_cmp_si(&rem, 0)) {
	return (swit_make_inexact(big_to_inexact(res)));
      } else {
	return (big_to_integer(res));
      }
    } else if SWIT_INEXACTP(arg2) {
      return (swit_make_inexact(big_to_inexact(SWIT_BIGNUM(arg1)) / SWIT_DBL_VAL(arg2)));
    } else /* (SWIT_BIGNUMP (arg2))*/ {
      return (swit_make_inexact(big_to_inexact(SWIT_BIGNUM(arg1)) / big_to_inexact(SWIT_BIGNUM(arg2))));
    }
  }
}

#define GEN_ZEROPLUS_OP(name, ident) \
static SWiT  \
name ## _prim (int argc, SWiT *argv) \
{ \
    switch (argc) {\
    case 0:\
	return (ident);\
    case 1: {\
        return (argv[0]);\
    }\
    case 2:\
	return (bin_ ## name(argv[0],argv[1]));\
    default: {\
	SWiT ret = ident;\
	int i;\
	for (i = 0; i < argc; i++) {\
	    ret = bin_ ## name (ret, argv[i]);\
	}\
	return (ret);\
    }\
    }\
}

GEN_ZEROPLUS_OP(add, swit_zero)
GEN_ZEROPLUS_OP(mul, swit_one)

#define GEN_ONEPLUS_OP(name, ident) \
static SWiT  \
name ## _prim (SWiT arg, int argc, SWiT *argv) \
{ \
    switch (argc) {\
    case 0:\
	return (bin_ ## name(ident,arg));\
    case 1: {\
	return (bin_ ## name(arg,argv[0]));\
    }\
    default: {\
	SWiT ret = arg;\
	int i;\
	for (i = 0; i < argc; i++) {\
	    ret = bin_ ## name (ret, argv[i]);\
	}\
	return (ret);\
    }\
    }\
}

GEN_ONEPLUS_OP(sub, swit_zero)
GEN_ONEPLUS_OP(div, swit_one)

#define GEN_BIN_MAX_MIN(name,op)\
static SWiT \
bin_ ## name (SWiT arg1, SWiT arg2) \
{ \
  if SWIT_FIXNUMP(arg1) { \
      int n1 = SWIT_FIXNUM(arg1); \
      if SWIT_FIXNUMP(arg2) { \
	  int n2 = SWIT_FIXNUM(arg2); \
	  return ((n1 op n2) ? arg1 : arg2); \
      } else if SWIT_INEXACTP(arg2) { \
	  double n2 = SWIT_DBL_VAL(arg2); \
          if ((double)n1 == n2) \
	     return (arg2);\
	  else \
             return ((n1 op n2) ? swit_make_inexact((double)n1) : arg2); \
      } else /* bignum */ { \
         return (arg2); \
      } \
  } else if SWIT_INEXACTP(arg1) { \
      double n1 = SWIT_DBL_VAL(arg1); \
      if SWIT_FIXNUMP(arg2) { \
	  double n2 = (double)SWIT_FIXNUM(arg2); \
          if (n1 == n2) \
	     return (arg1);\
	  else \
             return ((n1 op n2) ? arg1 : swit_make_inexact(n2)); \
      } else if SWIT_INEXACTP(arg2) { \
	  double n2 = SWIT_DBL_VAL(arg2); \
	  return ((n1 op n2) ? arg1 : arg2); \
      } else /* bignum */ { \
	  double n2 = big_to_inexact(SWIT_BIGNUM(arg2)); \
          if (n1 == n2) \
	     return (arg1);\
	  else \
             return ((n1 op n2) ? arg1 : arg2); \
      } \
  } else /* bignum */ { \
      if SWIT_FIXNUMP(arg2) { \
	  return (arg1); \
      } else if SWIT_INEXACTP(arg2) { \
	  double n1 = big_to_inexact(SWIT_BIGNUM(arg1)); \
	  double n2 = SWIT_DBL_VAL(arg2); \
          if (n1 == n2) \
	     return (arg2);\
	  else \
             return ((n1 op n2) ? swit_make_inexact(n1) : arg2); \
      } else /* bignum */ { \
          int n = mpz_cmp(SWIT_BIGNUM(arg1),SWIT_BIGNUM(arg2)); \
          return ((n op 0) ? arg1 : arg2); \
      } \
  } \
}

GEN_BIN_MAX_MIN(max,>)
GEN_BIN_MAX_MIN(min,<)

#define GEN_NARY_MAX_MIN(name, bin_name) \
static SWiT  \
name ## _prim(SWiT arg, int argc, SWiT *argv) \
{ \
    switch (argc) {\
    case 0:\
	return (arg);\
    case 1:\
	return (bin_name (arg,argv[0]));\
    default: {\
	SWiT ret = arg;\
	int i;\
	for (i = 0; i < argc; i++) {\
	    ret = bin_name (ret, argv[i]);\
	}\
	return (ret);\
    }\
    } \
}


GEN_NARY_MAX_MIN(max, bin_max)
GEN_NARY_MAX_MIN(min, bin_min)

/* Special treatment because of restriction on type */
static SWiT
expt_prim (SWiT n1, SWiT n2)
{
  if SWIT_FIXNUMP(n1) {
    int n = SWIT_FIXNUM(n1);
    if SWIT_FIXNUMP(n2) {
      int m = SWIT_FIXNUM(n2);
      if (m == 0) return swit_one;
      if (m > 0) {
	MP_INT *res1 = (MP_INT *)swit_malloc(sizeof(MP_INT)), res;
	mpz_init(res1);
	mpz_init_set_si(&res, n);
	mpz_pow_ui(res1, &res, m);
	return (big_to_integer(res1));
      } else {
	return (swit_make_inexact(pow(n, (double)m)));
      }
    } else if (SWIT_INEXACTP(n2)) {
      return (swit_make_inexact(pow (n, SWIT_DBL_VAL(n2))));
    } else /* bignum */ {
      prim_err_msg = too_big;
      exn = exn_math;
      exn_val = n1;
      return NULL;
    }
  } else if (SWIT_INEXACTP(n1)) {
    if (SWIT_FIXNUMP(n2)) {
      return (swit_make_inexact(pow(SWIT_DBL_VAL(n1), SWIT_FIXNUM(n2))));
    } else if (SWIT_INEXACTP(n2)) {
      return (swit_make_inexact(pow(SWIT_DBL_VAL(n1), SWIT_DBL_VAL(n2))));
    } else /* bignum */ {
      return (swit_make_inexact(pow(SWIT_DBL_VAL(n1), big_to_inexact(SWIT_BIGNUM(n2)))));
    }
  } else /* bignum */ {
    if (SWIT_FIXNUMP(n2)) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      int n = SWIT_FIXNUM(n2);
      mpz_init(res);
      mpz_pow_ui(res, SWIT_BIGNUM(n1), n);
      return (swit_make_bignum(res));
    } else if (SWIT_INEXACTP(n2)) {
      return (swit_make_inexact(pow (big_to_inexact(SWIT_BIGNUM(n1)), SWIT_DBL_VAL(n2))));
    } else /* (SWIT_BIGNUMP(n2))*/ {
      prim_err_msg = too_big;
      exn = exn_math;
      exn_val = n2;
      return NULL;
    }
  }
}

static SWiT
abs_prim (SWiT arg)
{
  if (SWIT_FIXNUMP(arg)) {
    int n = SWIT_FIXNUM(arg);
    return (swit_make_fixnum(abs(n)));
  } else if (SWIT_INEXACTP(arg)) {
    return (swit_make_inexact(fabs(SWIT_DBL_VAL(arg))));
  } else /* (SWIT_BIGNUMP(arg)) */ {
    MP_INT *pos = (MP_INT *)swit_malloc(sizeof(MP_INT));
    mpz_init(pos);
    mpz_abs(pos, SWIT_BIGNUM(arg));
    return (swit_make_bignum(pos));
  }
}

static SWiT
floor_prim (SWiT arg)
{
  if (SWIT_INTP (arg)) {
    return (arg);
  } else /* (SWIT_INEXACTP (arg)) */ {
    return (swit_make_inexact (floor (SWIT_DBL_VAL(arg))));
  }
}

static SWiT
ceiling_prim (SWiT arg)
{
  if (SWIT_INTP (arg)) {
    return (arg);
  } else /* (SWIT_INEXACTP (arg)) */ {
    return (swit_make_inexact (ceil (SWIT_DBL_VAL(arg))));
  }
}

static SWiT
truncate_prim (SWiT arg)
{
  if (SWIT_INTP (arg)) {
    return (arg);
  } else /* (SWIT_INEXACTP (arg)) */ {
    return (swit_make_inexact ((int)(SWIT_DBL_VAL(arg))));
  }
}

static SWiT
round_prim (SWiT arg)
{
  if (SWIT_INTP (arg)) {
    return (arg);
  } else /* (SWIT_INEXACTP (arg)) */ {
    return (swit_make_inexact (rint(SWIT_DBL_VAL(arg))));
  }
}

/*-------------------------------------------------------------------
 *        Functions |N x ... -> |N
 *        (inexact result if one argument is inexact)
 *-------------------------------------------------------------------
 */

#define MAX(n1,n2) (((n1)>(n2)) ? (n1) : (n2))
#define MIN(n1,n2) (((n1)<(n2)) ? (n1) : (n2))

static int
fix_gcd(int n1, int n2)
{
  return ((n2 == 0)? n1 : fix_gcd(n2, n1 % n2));
}

static SWiT
bin_gcd (SWiT arg1, SWiT arg2)
{
  if SWIT_FIXNUMP(arg1) {
    int n1 = SWIT_FIXNUM(arg1);
    if SWIT_FIXNUMP(arg2) {
      /* the result is certainly a fixnum */
      return swit_make_fixnum(fix_gcd(abs(n1),abs(SWIT_FIXNUM(arg2))));
    } else if SWIT_BIGNUMP(arg2) {
      MP_INT bn1, *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(res);
      mpz_init_set_si(&bn1, n1);
      mpz_gcd(res, &bn1, SWIT_BIGNUM(arg2));
      return big_to_integer(res);
    } else { /* arg2 : inexact int */
      double dbl = SWIT_DBL_VAL(arg2);
      if (dbl == floor(dbl)) {
	MP_INT bn1, bn2, *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
	mpz_init(res);
	mpz_init_set_si(&bn1, n1);
	mpz_init_set_si(&bn2, (int)dbl);
	mpz_gcd(res, &bn1, &bn2);
	return swit_make_inexact(big_to_inexact(res));
      } else {
	exn = exn_type;
	exn_val = arg2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else if SWIT_BIGNUMP(arg1) {
    if SWIT_BIGNUMP(arg2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(res);
      mpz_gcd(res, SWIT_BIGNUM(arg1), SWIT_BIGNUM(arg2));
      return (big_to_integer(res));
    } else { /* arg2 : inexact int */
      return bin_gcd(arg2, arg1);
    }
  } else { /* arg1 : inexact int */
    double dbl = SWIT_DBL_VAL(arg1);
    if (dbl == floor(dbl)) {
      return bin_gcd(inexact_to_exact(arg1), arg2);
    } else {	
      exn = exn_type;
      prim_err_msg = wrong_type;
      exn_val = arg1;
      return NULL;
    }
  }
}

static SWiT
bin_lcm (SWiT n1, SWiT n2)
{
  SWiT d = bin_gcd(n1, n2);
  if (zero_p_prim(d) == swit_true) {
    return (d);
  } else {
    return abs_prim(bin_mul(n1, quotient_prim(n2, d)));
  }
}

GEN_ZEROPLUS_OP(gcd, swit_zero)
GEN_ZEROPLUS_OP(lcm, swit_one)

static SWiT
quotient_prim (SWiT n1, SWiT n2)
{
  if SWIT_FIXNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      return swit_make_fixnum (SWIT_FIXNUM(n1) / SWIT_FIXNUM(n2));
    } else if SWIT_BIGNUMP(n2) {
      return swit_zero;
    } else /* n2 : double */ {
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(quotient_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else if SWIT_BIGNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init_set_si(res, SWIT_FIXNUM(n2));
      mpz_mdiv(res, SWIT_BIGNUM(n1), res);
      return(big_to_integer(res));
    } else if SWIT_BIGNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(res);
      mpz_mdiv(res, SWIT_BIGNUM(n1), SWIT_BIGNUM(n2));
      return big_to_integer(res);
    } else /* n2 : double */{
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(quotient_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else /* n1 : double */ {
    double dbl = SWIT_DBL_VAL(n1);
    if (dbl == floor(dbl)) {
      SWiT res = quotient_prim(inexact_to_integer(dbl), n2);
      return (SWIT_INTP(res) ? exact_to_inexact(res) : res);
    } else {
      exn = exn_type;
      exn_val = n1;
      prim_err_msg = wrong_type;
      return NULL;
    }
  }
}

static SWiT
rem_prim (SWiT n1, SWiT n2)
{
  if SWIT_FIXNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      return swit_make_fixnum (SWIT_FIXNUM(n1) % SWIT_FIXNUM(n2));
    } else if SWIT_BIGNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init_set_si(res, SWIT_FIXNUM(n1));
      mpz_mod(res, res,SWIT_BIGNUM(n2));
      return big_to_integer(res);
    } else /* n2 : double */ {
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(rem_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else if SWIT_BIGNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init_set_si(res, SWIT_FIXNUM(n2));
      mpz_mod(res,SWIT_BIGNUM(n1), res);
      return big_to_integer(res);
    } else if SWIT_BIGNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(res);
      mpz_mod(res, SWIT_BIGNUM(n1), SWIT_BIGNUM(n2));
      return big_to_integer(res);
    } else /* n2 : double */{
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(rem_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else /* n1 : double */ {
    double dbl = SWIT_DBL_VAL(n1);
    if (dbl == floor(dbl)) {
      SWiT res = rem_prim(inexact_to_integer(dbl), n2);
      return (SWIT_INTP(res) ? exact_to_inexact(res) : res);
    } else {
      exn = exn_type;
      exn_val = n1;
      prim_err_msg = wrong_type;
      return NULL;
    }
  }
}

static SWiT
modulo_prim (SWiT n1, SWiT n2)
{
  if SWIT_FIXNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      int i1, i2, i;
      i1 = SWIT_FIXNUM(n1);
      i2 = SWIT_FIXNUM(n2);
      i = i1 % i2;
      return swit_make_fixnum (((i2 < 0) ? (i > 0) : (i < 0)) ? i+i2 : i);
    } else if SWIT_BIGNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init_set_si(res, SWIT_FIXNUM(n1));
      mpz_mmod(res, res, SWIT_BIGNUM(n2));
      return big_to_integer(res);
    } else /* n2 : double */ {
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(modulo_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else if SWIT_BIGNUMP(n1) {
    if SWIT_FIXNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init_set_si(res, SWIT_FIXNUM(n2));
      mpz_mmod(res, SWIT_BIGNUM(n1), res);
      return big_to_integer(res);
    } else if SWIT_BIGNUMP(n2) {
      MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(res);
      mpz_mmod(res, SWIT_BIGNUM(n1), SWIT_BIGNUM(n2));
      return big_to_integer(res);
    } else /* n2 : double */{
      double dbl = SWIT_DBL_VAL(n2);
      if (dbl == floor(dbl)) {
	return exact_to_inexact(modulo_prim(n1, inexact_to_integer(dbl)));
      } else {
	exn = exn_type;
	exn_val = n2;
	prim_err_msg = wrong_type;
	return NULL;
      }
    }
  } else /* n1 : double */ {
    double dbl = SWIT_DBL_VAL(n1);
    if (dbl == floor(dbl)) {
      SWiT res = modulo_prim(inexact_to_integer(dbl) ,n2);
      return (SWIT_INTP(res) ? exact_to_inexact(res) : res);
    } else {
      exn = exn_type;
      exn_val = n1;
      prim_err_msg = wrong_type;
      return NULL;
    }
  }
}

static SWiT
fact_prim (SWiT arg)
{
  int n = SWIT_FIXNUM(arg);
  if (n < 0) {
    exn = exn_type;
    exn_val = arg;
    prim_err_msg = wrong_type;
    return NULL;
  } else {
    MP_INT *res = (MP_INT *)swit_malloc(sizeof(MP_INT));
    mpz_init(res);
    mpz_fac_ui(res,(unsigned long int)n);
    return big_to_integer(res);
  }
}


/*-------------------------------------------------------------------
 *        Functions |R -> |R and |R x |R -> |R
 *         always returning an inexact result.
 *-------------------------------------------------------------------
 */
#define GEN_UNARY_FUN(name) \
static SWiT  \
name ## _prim (SWiT arg) \
{ \
    if (SWIT_FIXNUMP (arg)) \
        return (swit_make_inexact (name (SWIT_FIXNUM (arg)))); \
    else if (SWIT_INEXACTP (arg)) \
        return (swit_make_inexact (name (SWIT_DBL_VAL(arg)))); \
    else /* (SWIT_BIGNUMP (arg)) */ \
	return (swit_make_inexact (name (big_to_inexact(SWIT_BIGNUM(arg)))));\
}


GEN_UNARY_FUN(exp)
GEN_UNARY_FUN(log)
GEN_UNARY_FUN(sin)
GEN_UNARY_FUN(cos)
GEN_UNARY_FUN(tan)
GEN_UNARY_FUN(asin)
GEN_UNARY_FUN(acos)
/* extensions : */
GEN_UNARY_FUN(cbrt)
GEN_UNARY_FUN(log10)
GEN_UNARY_FUN(expm1)
GEN_UNARY_FUN(log1p)
GEN_UNARY_FUN(sinh)
GEN_UNARY_FUN(cosh)
GEN_UNARY_FUN(tanh)
GEN_UNARY_FUN(asinh)
GEN_UNARY_FUN(acosh)
GEN_UNARY_FUN(atanh)

#define GEN_BINARY_FUN(name) \
static SWiT \
name ## _prim (SWiT arg1, SWiT arg2)\
{\
  if (SWIT_FIXNUMP(arg1))\
    if (SWIT_FIXNUMP(arg2))\
      return (swit_make_inexact (name (SWIT_FIXNUM(arg1), SWIT_FIXNUM(arg2))));\
    else if (SWIT_INEXACTP(arg2))\
      return (swit_make_inexact (name (SWIT_FIXNUM(arg1), SWIT_DBL_VAL(arg2))));\
    else /* (SWIT_BIGNUMP (arg2)) */ \
      return (swit_make_inexact (name (SWIT_FIXNUM(arg1), big_to_inexact(SWIT_BIGNUM(arg2)))));\
  else if (SWIT_INEXACTP(arg1))\
    if (SWIT_FIXNUMP(arg2))\
      return (swit_make_inexact (name (SWIT_DBL_VAL(arg1), SWIT_FIXNUM(arg2))));\
    else if (SWIT_INEXACTP(arg2))\
      return (swit_make_inexact (name (SWIT_DBL_VAL(arg1), SWIT_DBL_VAL(arg2))));\
    else /* (SWIT_BIGNUMP (arg2)) */ \
      return (swit_make_inexact (name (SWIT_DBL_VAL(arg1), big_to_inexact(SWIT_BIGNUM(arg2)))));\
  else /* (SWIT_BIGNUMP(arg1)) */ \
    if (SWIT_FIXNUMP(arg2))\
      return (swit_make_inexact (name (big_to_inexact(SWIT_BIGNUM(arg1)), SWIT_FIXNUM(arg2))));\
    else if (SWIT_INEXACTP(arg2))\
      return (swit_make_inexact (name (big_to_inexact(SWIT_BIGNUM(arg1)), SWIT_DBL_VAL(arg2))));\
    else /* (SWIT_BIGNUMP (arg2)) */ \
      return (swit_make_inexact (name (big_to_inexact(SWIT_BIGNUM(arg1)), big_to_inexact(SWIT_BIGNUM(arg2)))));\
}

GEN_BINARY_FUN(atan2)

GEN_BINARY_FUN(hypot)


/*-------------------------------------------------------------------
 *      Function 'sqrt' that can return an exact result
 *        when its argument is exact (but not always).
 *-------------------------------------------------------------------
 */
static SWiT
sqrt_prim (SWiT arg)
{
  if (SWIT_FIXNUMP (arg)) {
    int n = SWIT_FIXNUM (arg);
    double root = sqrt (n);
    if (((int)root * (int)root) == n)
      return (swit_make_fixnum ((int)root));
    else
      return (swit_make_inexact (root));
  } else if (SWIT_INEXACTP (arg)) {
    return (swit_make_inexact (sqrt (SWIT_DBL_VAL(arg))));
  } else /* (SWIT_BIGNUMP (arg)) */ {
    MP_INT *big = SWIT_BIGNUM(arg);
    if (mpz_perfect_square_p(big)) {
      MP_INT *sqroot = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(sqroot);
      mpz_sqrt(sqroot, big);
      /* Should be coerced to a fixnum whenever possible */
      return (big_to_integer(sqroot));
    } else {
      return (swit_make_inexact(sqrt(big_to_inexact(big))));
    }
  }
}

/*-------------------------------------------------------------------
 *                     Random
 *-------------------------------------------------------------------
 */
static SWiT
rand_prim (void)
{
  return (int_to_integer(rand()));
}

static SWiT
srand_prim (SWiT arg)
{
  int seed;
  if (SWIT_FIXNUMP(arg))
    seed = SWIT_FIXNUM(arg);
  else
    seed = mpz_get_ui(SWIT_BIGNUM(arg));
  srand(seed);
  return (int_to_integer(seed));
}


/*-------------------------------------------------------------------
 *           exact <-> inexact conversion functions
 *-------------------------------------------------------------------
 */
static SWiT
exact_to_inexact (SWiT arg)
{
  if SWIT_FIXNUMP(arg) {
    return (swit_make_inexact((double)SWIT_FIXNUM(arg)));
  } else if SWIT_BIGNUMP(arg) {
    return (swit_make_inexact(big_to_inexact(SWIT_BIGNUM(arg))));
  } else {
    return arg;
  }
}

static SWiT
inexact_to_exact (SWiT arg)
{
  if (SWIT_FIXNUMP(arg) || SWIT_BIGNUMP(arg)) {
    /* already exact */
    return arg;
  } else {
    double r = SWIT_DBL_VAL(arg);
    if ((r < 32767.5) && (r > -32768.5)) {
      return (swit_make_fixnum ((int)rint(r)));
    } else {
      SWiT n = inexact_to_integer(r);
      return (n);
    }
  }
}

/*-------------------------------------------------------------------
 *           string <-> number conversion functions
 *-------------------------------------------------------------------
 */

/* erreur si une base != 10 est donne avec un dbl ??? */

SWiT
number_to_string (SWiT arg1, SWiT arg2)
{
  int radix = SWIT_FIXNUM(arg2);

  if (SWIT_FIXNUMP(arg1)) {
    return (fixnum_to_string (SWIT_FIXNUM(arg1), radix));
  } else if (SWIT_INEXACTP(arg1)) {
    if (radix != 10) {
      prim_err_msg = wrong_radix;
      exn = exn_math;
      exn_val = arg1;
      return NULL;
    }
    return (inexact_to_string (SWIT_DBL_VAL(arg1)));
  } else /* (SWIT_BIGNUMP(arg1)) */ {
    return (bignum_to_string(SWIT_BIGNUM(arg1), radix));
  }
}

static SWiT
fixnum_to_string (int i, int radix)
{
  char buf[36];
  int len;
  char *res;

  switch (radix) {
  case 2:
    if (i == 0)  {
      buf[0] = '\0';
    } else {
      unsigned int k = 32, mask = 0x80000000, l = 0;
      if (i<0) {
	i = -i;
	buf[l++] = '-';
	k++;
      }
      for(;(mask & i) == 0; mask >>= 1) k--;
      for(; l < k; l++ , mask >>= 1) buf[l] = (mask & i) ? '1' : '0';
      buf[l] = '\0';
    }
    break;
  case 8:
    sprintf (buf, "%o", i);
    break;
  case 10:
    sprintf (buf, "%d", i);
    break;
  case 16:
    sprintf (buf, "%x", i);
    break;
  default:
    prim_err_msg = wrong_radix;
    exn = exn_math;
    exn_val = swit_make_fixnum(radix);
    return NULL;
  }

  len = strlen(buf);
  res = (char *)swit_malloc_atomic(len);
  memcpy(res, buf, len);
  return swit_make_string(res, len);
}


static SWiT
bignum_to_string (MP_INT *i, int radix)
{
  switch (radix) {
  case 2:
  case 8:
  case 10:
  case 16:
    {
      char *str = mpz_get_str(NULL, radix, i);
      return (swit_make_string(str, strlen(str)));
    }
  default:
    prim_err_msg = wrong_radix;
    exn = exn_math;
    exn_val = swit_make_fixnum(radix);
    return NULL;
  }
}

static SWiT
inexact_to_string (double d)
{
  char buf[32];
  int i, len, no_pt = 1, no_exp = 1;
  int l;
  char *res;

  sprintf (buf, "%1.10g", d);
  /* add a '.0' if neither decimal point nor exponent mark : */
  len = strlen (buf);
  for (i = 0; i < len && no_pt && no_exp; i++) {
    no_pt = (buf[i] == '.') ? 0 : 1;
    no_exp = ((buf[i] == 'e') || (buf[i] == 'n')) ? 0 : 1;
  }
  if (no_pt && no_exp) {
    buf[i++] = '.';
    buf[i++] = '0';
    buf[i] = '\0';
  }

  l = strlen(buf);
  res = (char *)swit_malloc_atomic(l);
  memcpy(res, buf, l);
  return swit_make_string(res, l);
}

/* faux depuis les nouveaux string (n'tant pas des C-string) */
static SWiT
string_to_number(char *str, int len, int base, int exactness)
{
  int is_float, i;
  char *ptr;
  double d;
  
  if (*str == '+') {
    str++; len--;
    if (!isdigit(*str)) return swit_false;
  }
  if (!len || ((len == 1) && (!isdigit(*str)))) {
    return swit_false;
  }

  is_float = 0;
  if (base == 10) {
    for (i = 0 ; i < len ; ++i ) {
      int ch = str[i];
      if ((ch == '.') || (ch == 'e') || (ch == 'E') || (ch == '#')) {
	is_float = 1;
	if (ch == '#') str[i] = '0';
      } else if (ch == '\0') {
	return swit_false;
      }
    }
    if (is_float || (exactness == -1)) {
      if (exactness == 1) return swit_false;
      d = strtod(str, &ptr);
      if (*ptr) {
	  return (swit_false);
      }
      return swit_make_inexact (d);
    } else {
      MP_INT *val = (MP_INT *)swit_malloc(sizeof(MP_INT));
      mpz_init(val);
      if (mpz_init_set_str(val, str, 10) == -1){
	return swit_false;
      } else {
	return big_to_integer(val);
      }
    }
  } else {
    MP_INT *val = (MP_INT *)swit_malloc(sizeof(MP_INT));
    mpz_init(val);
    if (mpz_init_set_str(val, str, base) == -1){
      return swit_false;
    } else {
      return big_to_integer(val);
    }
  }
}

static SWiT
string_to_number_prim (SWiT arg1, SWiT arg2)
{
  if (arg1 != swit_null_string) {
    int base = 0;
    int exactness = 0;
    char *str = STRING_ELS(arg1);
    /* radix : 10 is the default, but can be overriden in string */
    if (str[0]=='#') {
      switch (str[1]) {
      case 'b':
      case 'B':
	base = 2;
	exactness = 1;
	break;
      case 'o':
      case 'O':
	base = 8;
	exactness = 1;
	break;
      case 'x':
      case 'X':
	base = 16;
	exactness = 1;
	break;
      case 'd':
      case 'D':
	base = 10;
	exactness = 1;
	break;
      case 'e':
      case 'E':
	exactness = 1;
	break;
      case 'i':
      case 'I':
	exactness = -1;
	break;
      default:
	return (swit_false);
      }
      str += 2;
    }
    if (str[0] == '#') {
      switch (str[1]) {
      case 'b':
      case 'B':
	if ((exactness == -1) || base)
	  return (swit_false);
	exactness = 1;
	base = 2;
	break;
      case 'o':
      case 'O':
	if ((exactness == -1) || base)
	  return (swit_false);
	exactness = 1;
	base = 8;
	break;
      case 'x':
      case 'X':
	if ((exactness == -1) || base)
	  return (swit_false);
	exactness = 1;
	base = 16;
	break;
      case 'd':
      case 'D':
	if ((exactness == -1) || base)
	  return (swit_false);
	exactness = 1;
	base = 10;
	break;
      case 'e':
      case 'E':
	if (exactness && (base == 0))
	  return (swit_false);
	exactness = 1;
	break;
      case 'i':
      case 'I':
	if (exactness)
	  return (swit_false);
	base = 10;
	exactness = -1;
	break;
      default:
	return (swit_false);
      }
      str += 2;
    } else {
      if (!base) base = SWIT_FIXNUM (arg2);
    }
    return(string_to_number(str, STRING_LENGTH(arg1), base, exactness));
  } else {
    return swit_false;
  }
}
  

/*
 * The number 10, used as the default base for 'string->number'
 *  and 'number->string' (exported to switPrint.c).
 */

void
swit_init_number (void)
{  
  print_table[NUMBER >> 1] = print_number;
  swit_ten = swit_make_fixnum(10);
  swit_make_prim1 ("number?", number_p_prim,
		   ANY);
  swit_make_prim1 ("complex?", complex_p_prim,
		   ANY);
  swit_make_prim1 ("real?", real_p_prim,
		   ANY);
  swit_make_prim1 ("rational?", rational_p_prim,
		   ANY);
  swit_make_prim1 ("integer?", integer_p_prim,
		   ANY);
  swit_make_prim1 ("fixnum?", fixnum_p_prim,
		   ANY);
  swit_make_prim1 ("bignum?", bignum_p_prim,
		   ANY);
  swit_make_prim1 ("exact?", exact_p_prim,
		   NUM_M);
  swit_make_prim1 ("inexact?", inexact_p_prim,
		   NUM_M);
  swit_make_prim2n ("=", eq_prim, 
		    NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim2n ("<", lt_prim, 
		    NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim2n (">", gt_prim, 
		    NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim2n ("<=", le_prim,
		    NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim2n (">=", ge_prim,
		    NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim1 ("zero?", zero_p_prim,
		   NUM_M);
  swit_make_prim1 ("positive?", positive_p_prim,
		   NUM_M);
  swit_make_prim1 ("negative?", negative_p_prim,
		   NUM_M);
  swit_make_prim1 ("odd?", odd_p_prim,
		   NUM_M);
  swit_make_prim1 ("even?", even_p_prim,
		   NUM_M);
  swit_make_prim1n ("max", max_prim,
		    NUM_M, NUM_M, NUM_M);
  swit_make_prim1n ("min", min_prim,
		    NUM_M, NUM_M, NUM_M);
  swit_make_prim0n ("+", add_prim,
		    NUM_M, NUM_M);
  swit_make_prim1n ("-", sub_prim,
		    NUM_M, NUM_M, NUM_M);
  swit_make_prim0n ("*", mul_prim,
		    NUM_M, NUM_M);
  swit_make_prim1n ("/", div_prim,
		    NUM_M, NUM_M, NUM_M);
  swit_make_prim1 ("abs", abs_prim,
		   NUM_M);
  swit_make_prim2 ("quotient", quotient_prim,
		   NUM_M, NUM_M);
  swit_make_prim2 ("remainder", rem_prim,
		   NUM_M, NUM_M);
  swit_make_prim2 ("modulo", modulo_prim,
		   NUM_M, NUM_M);
  swit_make_prim0n ("gcd", gcd_prim,
		    NUM_M, NUM_M);
  swit_make_prim0n ("lcm", lcm_prim,
		    NUM_M, NUM_M);
  swit_make_prim1 ("floor", floor_prim,
		   NUM_M);
  swit_make_prim1 ("ceiling", ceiling_prim,
		   NUM_M);
  swit_make_prim1 ("truncate", truncate_prim,
		   NUM_M);
  swit_make_prim1 ("round", round_prim,
		   NUM_M);
  swit_make_prim1 ("exp", exp_prim,
		   NUM_M);
  swit_make_prim1 ("log", log_prim,
		   NUM_M);
  swit_make_prim1 ("sin", sin_prim,
		   NUM_M);
  swit_make_prim1 ("cos", cos_prim,
		   NUM_M);
  swit_make_prim1 ("tan", tan_prim,
		   NUM_M);
  swit_make_prim1 ("asin", asin_prim,
		   NUM_M);
  swit_make_prim1 ("acos", acos_prim,
		   NUM_M);
  swit_make_prim12 ("atan", atan2_prim, 
		    NUM_M, NUM_M, swit_one);
  swit_make_prim2 ("hypot", hypot_prim, 
		   NUM_M, NUM_M);
  swit_make_prim1 ("sqrt", sqrt_prim,
		   NUM_M);
  swit_make_prim1 ("cbrt", cbrt_prim,
		   NUM_M);
  swit_make_prim1 ("log10", log10_prim,
		   NUM_M);
  swit_make_prim1 ("exp-1", expm1_prim,
		   NUM_M);
  swit_make_prim1 ("log1+", log1p_prim,
		   NUM_M);
  swit_make_prim1 ("sinh", sinh_prim,
		   NUM_M);
  swit_make_prim1 ("cosh", cosh_prim,
		   NUM_M);
  swit_make_prim1 ("tanh", tanh_prim,
		   NUM_M);
  swit_make_prim1 ("asinh", asinh_prim,
		   NUM_M);
  swit_make_prim1 ("acosh", acosh_prim,
		   NUM_M);
  swit_make_prim1 ("atanh", atanh_prim,
		   NUM_M);
  swit_make_prim1 ("fact", fact_prim,
		   FIXNUM);
  swit_make_prim1 ("prime?", prime_p_prim,
		   NUM_M);
  swit_make_prim0 ("random", rand_prim);
  swit_make_prim1 ("srandom", srand_prim,
		   FIXNUM);
  swit_make_prim2 ("expt", expt_prim,
		   NUM_M, NUM_M);
  swit_make_prim1 ("exact->inexact", exact_to_inexact,
		   NUM_M);
  swit_make_prim1 ("inexact->exact", inexact_to_exact,
		   NUM_M);
  swit_make_prim12 ("number->string", number_to_string,
		    NUM_M, FIXNUM,
		    swit_ten);
  swit_make_prim12 ("string->number", string_to_number_prim,
		    STRING_M, FIXNUM,
		    swit_ten);
}
