/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*-----------------------------------------------------------
 *                     6.10.3 OUTPUT
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *
 *  write                      <OBJ> X [<PORT>] -> <OBJ>
 *  display                    <OBJ> X [<PORT>] -> <OBJ>
 *  newline                    [<PORT>] -> <OBJ>
 *  write-char                 <CHAR> X [<PORT>] -> <OBJ>
 *
 *-----------------------------------------------------------
 */

#include "swit.h"
#include <sys/socket.h>
#include <errno.h>

#define current_op SWIT_PORT(cur_out_port)
#define out_descr PORT_DESCR(current_op)
#define print_buffer PORT_BUFFER(current_op)
#define print_cur_idx PORT_INDEX(current_op)
#define print_base_idx PORT_EOB(current_op)
#define write_fun PORT_WRITE_FUN(current_op)

/* Forward declarations */
static SWiT write_prim ( SWiT arg1, SWiT arg2);
static SWiT display_prim (SWiT arg1, SWiT arg2);
static SWiT newline_prim (SWiT arg);
static SWiT write_char_prim ( SWiT arg1, SWiT arg2);
static int print_swit (SWiT obj, int esc);

/* imported */
extern SWiT cur_out_port;
extern SWiT null_out_port;
extern int issocket(int des);
extern void print_pair (SWiT expr, int esc, int tcl, SWiT fv, int depth);
extern int print_depth;

/* exported */
SWiT big; /* big = 1000 */
void swit_printl (SWiT obj, SWiT fv);
void print_to_buffer (SWiT obj, int esc, int tcl, SWiT fv, int depth);
void str2buff (char *str);
void space2buff (void);
void *print_table[128];

typedef SWiT(* Print_Fun)(SWiT, int, int, SWiT, int);
/* to be called as: ((Print_Fun)(print_table[ TYPE & MASK>> 1]))(...) */

/* for debugging C-code  */
void
debug(SWiT obj)
{
  print_swit(obj, 1);
  fprintf(stdout,"\n");
}

/* Only for interactive output */
void
swit_printl (SWiT obj, SWiT fv)
{
  int len;
  print_cur_idx = 0;
  print_to_buffer(obj, 1, 0, fv, 0);
  len = print_cur_idx;
  /* flush buffer */
  if ((*write_fun)(out_descr, print_buffer, len) != len) {
    swit_signal_error(swit_else, out_err);
  }
  PORT_EOB(current_op) = 0;
  PORT_INDEX(current_op) = 0;
}

/* All printing use this function */
void
mem2buff (char *str, int len)
{
  int idx = print_cur_idx;
  int free = IO_BUFFER_SIZE - idx;

 start_over:
  if (len <= free) {
    register i;
    for(i = 0; i < len; i++) {
      print_buffer[idx++] = *str++;
    }
  } else { /* buffer is smaller than string */
    register i;
    for(i = 0; i < free; i++) {
      print_buffer[idx++] = *str++;
    }

    /* flush the buffer */
    if ((*write_fun)(out_descr, print_buffer, IO_BUFFER_SIZE)
	  != IO_BUFFER_SIZE) {
      PORT_EOB(current_op) = 0;
      PORT_INDEX(current_op) = 0;
      swit_signal_error(swit_else, out_err);
    }

    /* an empty buffer */
    PORT_EOB(current_op) = 0;
    PORT_INDEX(current_op) = 0;
    idx = 0;
    len -= free;
    free = IO_BUFFER_SIZE;
    goto start_over;
  }

  /* idx has been incremented and hold the current pos. */
  PORT_INDEX(current_op) = idx;
  PORT_EOB(current_op) = idx;
}


void
str2buff (char *str)
{
  mem2buff(str, strlen(str));
}

void
char2buff(char c)
{
  mem2buff(&c, 1);
}

void
space2buff (void)
{
  mem2buff(" ", 1);
}


void
print_to_buffer (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  if (depth > 10*print_depth) {
    mem2buff("..", 2);
  } else if SWIT_PAIRP(obj) {
    print_pair(obj, esc, tcl, fv, depth);
  } else {
    unsigned int idx = (SWIT_TYPE(obj) & TYPE_MASK1) >> 1;
    if ((idx < 128) && print_table[idx]) {
      ((Print_Fun)(print_table[idx]))(obj, esc, tcl, fv, depth);
    } else {
      if SWIT_VECTORP(obj) { /* a marked vector */
	((Print_Fun)(print_table[VECT >> 1]))(obj, esc, tcl, fv, depth);
      } else if SWIT_STRINGP(obj) {  /* a very long string */
	((Print_Fun)(print_table[STRING >> 1]))(obj, esc, tcl, fv, depth);
      } else {
	str2buff("#<error>");
      }
    }
  }
}

/*
 *================================================================
 *                   6.10.3.   OUTPUT PRIMITIVES
 *================================================================
 */

/* print obj into the current output-port buffer and then flush */
static int
print_swit (SWiT obj, int esc)
{
  SWiT_Port p = SWIT_PORT(cur_out_port);
  int descr = PORT_DESCR(p);
  int eob = PORT_EOB(p);
  int idx = PORT_INDEX(p);

  if (descr == -1) {
    prim_err_msg = closed;
    exn = exn_io;
    return 0;
  }

  if (!write_fun) {
    prim_err_msg = wrong_type;
    exn = exn_io;
    return 0;
  }

  /* adjust file position (to ignore any previous reading) */
  if (eob != idx) {
    if (!issocket(descr)
	&& ((lseek(descr, idx - eob, SEEK_CUR)) == -1)) {
      exn = exn_io;
      prim_err_msg = sys_errlist[errno];
      return 0;
    }

    /* now, it's an empty buffer */
    PORT_EOB(p) = 0;
    PORT_INDEX(p) = 0;
  }

  /* write into buffer, if obj is bigger, 
     buffer is flushed (see mem2buff) */
  print_to_buffer(obj, esc, 0, swit_null_list, 0);

  /* idx is the # of char not yet flushed */
  idx = PORT_INDEX(p);

  /* flush them */
  if ((*write_fun)(descr, print_buffer, idx) != idx) {
    prim_err_msg = sys_errlist[errno]; /* not always the best msg */
    exn = exn_io;
    PORT_EOB(p) = 0;
    PORT_INDEX(p) = 0;
    return 0;
  }

  PORT_EOB(p) = 0;
  PORT_INDEX(p) = 0;

  return 1;
}

static SWiT
write_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_out_port;
  int noerror;
  if (arg2 != null_out_port) {
    cur_out_port = arg2;
  }
  noerror = print_swit(arg1, 1);
  cur_out_port = old_port;
  return noerror ? swit_true : NULL;
}

static SWiT
display_prim (SWiT arg1, SWiT arg2)
{
  SWiT old_port = cur_out_port;
  int noerror;
  if (arg2 != null_out_port) {
    cur_out_port = arg2;
  }
  noerror = print_swit(arg1, 0);
  cur_out_port = old_port;
  return noerror ? swit_true : NULL;
}

static SWiT
write_char (int c, SWiT arg)
{
  SWiT old_port = cur_out_port;
  SWiT_Port p;
  int descr;

  if (arg != null_out_port) {
    cur_out_port = arg;
  }

  p = SWIT_PORT(cur_out_port);
  descr = PORT_DESCR(p);

  if (descr == -1) {
    cur_out_port = old_port;
    prim_err_msg = closed;
    exn = exn_io;
    return NULL;
  }

  if (PORT_EOB(p) != PORT_INDEX(p)) {
    /* was reading => adjust position */
    if (!issocket(descr)
	&& ((lseek(descr, PORT_INDEX(p) - PORT_EOB(p), SEEK_CUR)) == -1)) {
      exn = exn_io;
      prim_err_msg = sys_errlist[errno];
      return 0;
    }
    PORT_EOB(p) = 0;
    PORT_INDEX(p) = 0;
  }

  if (c < 256) {
    unsigned char c1 = c;
    if ((*write_fun)(descr, &c1, 1) != -1) {
      cur_out_port = old_port;
      return swit_true;
    } else {
      goto error;
      }
  } else if (c < 65536) {
    unsigned char c1 = c % 256;
    unsigned char c2 = c / 256;
    if (((*write_fun)(descr, &c1, 1) != -1)
	&& ((*write_fun)(descr, &c2, 1) != -1)) {
      cur_out_port = old_port;
      return swit_true;
    } else {
      goto error;
    }
  } else {
    goto error;
  }
error:
  cur_out_port = old_port;
  prim_err_msg = cannot_output;
  exn = exn_io;
  return NULL;
}

static SWiT
newline_prim (SWiT arg)
{
  return (write_char('\n', arg));
}

static SWiT
write_char_prim (SWiT arg1, SWiT arg2)
{
  return (write_char(SWIT_CHAR(arg1), arg2));
}

/*
 *================================================================
 *                  INITIALISATION
 *================================================================
 */
void
swit_init_print (void)
{
  big = swit_make_fixnum(1000);
  swit_make_prim12 ("write", write_prim,
		    ANY, PORT, null_out_port);
  swit_make_prim12 ("display", display_prim,
		    ANY, PORT, null_out_port);
  swit_make_prim12 ("write-char", write_char_prim,
		    CHAR, PORT, null_out_port);
  swit_make_prim01 ("newline", newline_prim,
		    PORT, null_out_port);
}
