/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*-----------------------------------------------------------------------
 *   Quads : used for 3D graphics as points (with homogenous coordinates),
 *   as vectors when the fourth coordinate is zero (= points at infinity),
 *   and as quaternions (= rotations in space)
 *-----------------------------------------------------------------------
 */

/*==============================================================
 *                      POINTS, VECT3 and QUATERNIONS
 *==============================================================
 *   quad?                  OBJ  ->  BOOLEAN
 *   quad                   REAL X REAL [ X REAL X REAL ] -> QUAD
 *   quad-ref               QUAD X INTEGER -> REAL
 *   quad-set!              QUAD X INTEGER X REAL   -> QUAD
 *   quad-copy              QUAD -> QUAD
 *   quad-length            QUAD -> REAL
 *   quad-equal?            QUAD X QUAD -> BOOLEAN
 *   quad-scale             REAL X QUAD -> QUAD
 *
 *   point-distance         POINT X POINT -> REAL
 *   point-translate        VECT3 X POINT -> POINT
 *   points-translate!      VECT3 X POINT X POINT ... -> #t
 *   point-rotate           QUATER X POINT -> POINT
 *   points-rotate!         QUATER X POINT X POINT ... -> #t
 *   point-line             POINT X POINT X NUMBER -> POINT
 *   points->vect3          POINT X POINT -> VECT3
 *
 *   vect3-mult             VECT3 X VECT3 -> VECT3
 *   vect3-add              VECT3 X VECT3 -> VECT3
 *   vect3-sub              VECT3 X VECT3 -> VECT3
 *   vect3-normalize        VECT3 -> VECT3
 *   vect3-dot              VECT3 X VECT3 -> REAL
 *==================================================================
 */

#include "swit.h"
#include "quad.h"
#include <math.h>

/* imported */
extern double big_to_inexact(MP_INT *x); /* defined in switNumber.c */

/* locally defined */
static SWiT make_quad_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);

/*==================================================================
 *                            TYPE
 *==================================================================
 */
SWiT swit_quad; 

/*==============================================================
 *  SYNTAX : <quad> == { <real> <real> }
 *                  | { <real> <real> <real> }
 *                  | { <real> <real> <real> <real> }
 *==============================================================
 */

extern SWiT cur_in_port;
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
static SWiT
read_quad (void)
{
  int token;
  SWiT zw[2], y, x = rd_datum(get_tok());

  if (!SWIT_NUMBERP(x)) {
    exn = exn_read;
    prim_err_msg = wrong_quad_coord;
    longjmp(*read_jbuf, 2);
    return NULL;
  }
  y = rd_datum(get_tok());
  if (!SWIT_NUMBERP(y)) {
    exn = exn_read;
    prim_err_msg = wrong_quad_coord;
    longjmp(*read_jbuf, 2);
    return NULL;
  }
  token = get_tok();
  if (token == '}') {
    return make_quad_prim(x, y, 0, NULL);
  } else {
    zw[0] = rd_datum(token);
    if (!SWIT_NUMBERP(zw[0])) {
      exn = exn_read;
      prim_err_msg = wrong_quad_coord;
      longjmp(*read_jbuf, 2);
      return NULL;
    }
    token = get_tok();
    if (token == '}') {
      return make_quad_prim(x, y, 1, zw);
    }
    zw[1] = rd_datum(token);
    if (!SWIT_NUMBERP(zw[1])) {
      exn = exn_read;
      prim_err_msg = wrong_quad_coord;
      longjmp(*read_jbuf, 2);
      return NULL;
    }
    if (get_tok() != '}') {
      exn = exn_read;
      prim_err_msg = unexpected_token;
      longjmp(*read_jbuf, 2);
      return NULL;
    } else {
      return make_quad_prim(x, y, 2, zw);
    }
  }
}


extern void print_inexact (double x);

static void
print_double (double x)
{
  print_inexact(x);
}

static void
print_quad (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  SWiT_Quad p = SWIT_QUAD(obj);
  double z = QUAD_Z(p), w = QUAD_W(p);
  depth++;
  char2buff('{');
  print_double(QUAD_X(p));
  space2buff();
  print_double(QUAD_Y(p));
  if (w == 1.0) {
    if (z == 0.0) {
      char2buff('}');
    } else {
      space2buff();
      print_double(QUAD_Z(p));
      char2buff('}');
    }
  } else {
    space2buff();
    print_double(QUAD_Z(p));
    space2buff();
    print_double(QUAD_W(p));
    char2buff('}');
  }
}



/*===========================================================
 *                        QUADS
 *===========================================================
 */

double
to_double (SWiT arg)
{
  if (SWIT_INEXACTP(arg)) {
    return SWIT_DBL_VAL(arg);
  } else if (SWIT_FIXNUMP (arg)) {
    return ((double)SWIT_FIXNUM(arg));
  } else /* SWIT_BIGNUMP(arg) */ {
    return big_to_inexact(SWIT_BIGNUM(arg));
  }
}


SWiT_Quad
make_quad (double x, double y, double z, double w)
{
  SWiT_Quad p = SWIT_ALLOC(Quad);
  QUAD_X(p) = x;
  QUAD_Y(p) = y;
  QUAD_Z(p) = z;
  QUAD_W(p) = w;
  return p;
}


SWiT
swit_make_quad (SWiT_Quad q)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = QUAD;
  SWIT_QUAD(obj) = q;
  return(obj);
}



/*------------------------------------------------------------------
 * quad --
 *
 * Syntax : (quad x y z w)
 * ^^^^^^   where x,y,z,w are real numbers (world coordinates),
 *          z and w are optional : if w is omitted then w = 1,
 *          if z is also omitted then z = 0.
 *
 * Result : a point which homogeneous coordinates are (x,y,z,w)
 * ^^^^^^   If w = 0, a vect3 (i.e a point at infinity) is defined.
 *------------------------------------------------------------------
 */

static SWiT
make_quad_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv)
{
  double x = to_double(arg1);
  double y = to_double(arg2);
  double z,w;

  switch (argc) {
  case 0:
    z = 0.0;
    w = 1.0;
    break;
  case 1:
    z = to_double(argv[0]);
    w = 1.0;
    break;
  case 2:
    z = to_double(argv[0]);
    w = to_double(argv[1]);
    /*    if (w<0) {
      w = -w;
      z = -z;
      x = -x;
      y = -y;
    } */
    break;
  default:
    prim_err_msg = too_many_args;
    exn = exn_arity;
    return NULL;
  }
  return swit_make_quad(make_quad(x, y, z, w));
}

static SWiT
quad_p_prim (SWiT arg)
{
  return (SWIT_QUADP(arg) ? swit_true : swit_false);
}

/* also for equal? primitive */
int
equal_quad (SWiT arg1, SWiT arg2)
{
  SWiT_Quad p = SWIT_QUAD(arg1);
  SWiT_Quad q = SWIT_QUAD(arg2);
  return((QUAD_X(p) == QUAD_X(q)) &&
	 (QUAD_Y(p) == QUAD_Y(q)) &&
	 (QUAD_Z(p) == QUAD_Z(q)) &&
	 (QUAD_W(p) == QUAD_W(q)));
}

static SWiT
quad_equal_prim (SWiT arg1, SWiT arg2)
{
  return (equal_quad(arg1, arg2) ? swit_true : swit_false);
}

static SWiT
quad_ref_prim (SWiT point, SWiT index)
{
  int n = SWIT_FIXNUM(index);
  SWiT_Quad p = SWIT_QUAD(point);
  double d;
  switch (n) {
  case 0:
    d = QUAD_X(p);
    break;
  case 1:
    d = QUAD_Y(p);
    break;
  case 2:
    d = QUAD_Z(p);
    break;
  case 3:
    d = QUAD_W(p);
    break;
  default:
    prim_err_msg = wrong_index;
    exn = exn_range;
    return NULL;
  }
  return (swit_make_inexact(d));
}

static SWiT
quad_set_prim (SWiT point, SWiT index, SWiT val)
{
  int n = SWIT_FIXNUM(index);
  SWiT_Quad p = SWIT_QUAD(point);
  double d = to_double(val);
  switch (n)
    {
    case 0:
      QUAD_X(p) = d;
      break;
    case 1:
      QUAD_Y(p) = d;
      break;
    case 2:
      QUAD_Z(p) = d;
      break;
    case 3:
      QUAD_W(p) = d;
      break;
    default:
      prim_err_msg = wrong_index;
      exn = exn_range;
      return NULL;
    }
  return (point);
}

/* exported to switModel.c */
SWiT
quad_copy_prim (SWiT arg)
{
  SWiT_Quad p = SWIT_QUAD(arg);
  return swit_make_quad(make_quad(QUAD_X(p),QUAD_Y(p),
				  QUAD_Z(p),QUAD_W(p)));
}


/* Euclidian length (meaningful for vect3 or quaternions) */
static SWiT
quad_length_prim (SWiT arg)
{
  SWiT_Quad q = SWIT_QUAD(arg);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  return swit_make_inexact(sqrt(x*x+y*y+z*z+w*w));
}

/* returns q1 = k*q */
static SWiT
quad_nscale_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv)
{
  double k = to_double(arg1);
  int i;
  SWiT_Quad q = SWIT_QUAD(arg2);
  QUAD_X(q) *= k;
  QUAD_Y(q) *= k;
  QUAD_Z(q) *= k;
  QUAD_W(q) *= k;
  for (i = 0; i < argc; i++) {
    q = SWIT_QUAD(*argv++);
    QUAD_X(q) *= k;
    QUAD_Y(q) *= k;
    QUAD_Z(q) *= k;
    QUAD_W(q) *= k;
  }
  return swit_true;
}


/* returns q1 = k*q */
static SWiT
quad_scale_prim (SWiT arg1, SWiT arg2)
{
  double k = to_double(arg1);
  SWiT_Quad q = SWIT_QUAD(arg2);
  return swit_make_quad(make_quad(k*QUAD_X(q), k*QUAD_Y(q), k*QUAD_Z(q), k*QUAD_W(q)));
}

static SWiT
set_qmark_prim (SWiT arg1, SWiT arg2)
{
  unsigned int m = (SWIT_FIXNUM(arg2) % 16) << 16;
  SWIT_TYPE(arg1) &=  ~0xf0000;
  SWIT_TYPE(arg1) |=  m;
  return swit_true;
}

static SWiT
get_qmark_prim (SWiT arg)
{
  return swit_make_fixnum((SWIT_TYPE(arg) & 0x000f0000) >> 16);
}

/*===========================================================
 *                         POINTS
 *===========================================================
 */

static void *
wt ()
{
  exn = exn_type;
  prim_err_msg = wrong_type;
  return NULL;
}

/* returns the distance between p1 and p2 */
static SWiT
distance_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad p1 = SWIT_QUAD(arg1);
  SWiT_Quad p2 = SWIT_QUAD(arg2);
  double w1 = QUAD_W(p1);
  double w2 = QUAD_W(p2);
  double w = w1*w2;
  double x = QUAD_X(p1)*w2-QUAD_X(p2)*w1;
  double y = QUAD_Y(p1)*w2-QUAD_Y(p2)*w1;
  double z = QUAD_Z(p1)*w2-QUAD_Z(p2)*w1;
  if (w == 0) {
    return wt();
  }
  return swit_make_inexact(sqrt(x*x+y*y+z*z)/w);
}

static SWiT
/* Returns p' = p+v */
translate_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad v = SWIT_QUAD(arg1);
  SWiT_Quad p = SWIT_QUAD(arg2);
  double w = QUAD_W(p);
  if ((w == 0) || (QUAD_W(v) != 0)) {
    return wt();
  }
  return swit_make_quad(make_quad(QUAD_X(p)+QUAD_X(v)*w,
		    QUAD_Y(p)+QUAD_Y(v)*w,
		    QUAD_Z(p)+QUAD_Z(v)*w,
		    w));
}

static SWiT
/* Returns p = p+v */
ntranslate_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv)
{
  int i;
  SWiT_Quad v = SWIT_QUAD(arg1);
  double dx = QUAD_X(v);
  double dy = QUAD_Y(v);
  double dz = QUAD_Z(v);
  SWiT_Quad p = SWIT_QUAD(arg2);
  double w = QUAD_W(p);
  if ((w == 0) || (QUAD_W(v) != 0)) {
    return wt();
  } else {
    QUAD_X(p) += dx*w;
    QUAD_Y(p) += dy*w;
    QUAD_Z(p) += dz*w;
    for (i = 0; i < argc; i++) {
      p = SWIT_QUAD(*argv++);
      w = QUAD_W(p);
      QUAD_X(p) += dx*w;
      QUAD_Y(p) += dy*w;
      QUAD_Z(p) += dz*w;
    }
    return swit_true;
  }
}

/* returns (1-t)*p1+t*p2, i.e a point that lies on the line p1p2 */
SWiT_Quad
pt_line (SWiT_Quad p1, SWiT_Quad p2, double t)
{
  double w1 = QUAD_W(p1);
  double w2 = QUAD_W(p2);
  if ((w1 == 0) || (w2 == 0)) {
    return wt();
  } else {
    double k1 = (1-t)*w2;
    double k2 = t*w1;
    return make_quad(QUAD_X(p1)*k1+QUAD_X(p2)*k2,
		     QUAD_Y(p1)*k1+QUAD_Y(p2)*k2,
		     QUAD_Z(p1)*k1+QUAD_Z(p2)*k2,
		     w1*w2);
  }
}

static SWiT
point_line_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  return swit_make_quad(pt_line(SWIT_QUAD(arg1),
				SWIT_QUAD(arg2),
				to_double(arg3)));
}

/* compute q*p*q' */
static SWiT
rotate_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad q = SWIT_QUAD(arg1);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  double x2 = x*x;
  double y2 = y*y;
  double z2 = z*z;
  double w2 = w*w;
  double n = (x2+y2+z2+w2)/2;
  double xy = x*y;
  double yz = y*z;
  double zx = x*z;
  double wx = x*w;
  double wy = y*w;
  double wz = z*w;
  SWiT_Quad p = SWIT_QUAD(arg2);
  double px = QUAD_X(p);
  double py = QUAD_Y(p);
  double pz = QUAD_Z(p);
  double pw = QUAD_W(p);
  /*  if (pw == 0) {
      return wt();
    } */
  if (n == 0) {
    prim_err_msg = null_quad;
    exn = exn_math;
    return NULL;
  }
  return swit_make_quad(make_quad((px*(n-y2-z2) + py*(xy-wz) + pz*(zx+wy))/n,
		    (px*(xy+wz) + py*(n-z2-x2) + pz*(yz-wx))/n,
		    (px*(zx-wy) + py*(yz+wx) + pz*(n-x2-y2))/n,
		    pw));
}

/* assign  q*p*q' to p for each argument p */
static SWiT
nrotate_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv)
{
  int i;
  SWiT_Quad q = SWIT_QUAD(arg1);
  double x = QUAD_X(q);
  double y = QUAD_Y(q);
  double z = QUAD_Z(q);
  double w = QUAD_W(q);
  double x2 = x*x;
  double y2 = y*y;
  double z2 = z*z;
  double w2 = w*w;
  double n = (x2+y2+z2+w2)/2;
  double xy = x*y;
  double yz = y*z;
  double zx = x*z;
  double wx = x*w;
  double wy = y*w;
  double wz = z*w;
  double A = n-y2-z2;
  double B = xy-wz;
  double C = zx+wy;
  double D = xy+wz;
  double E = n-z2-x2;
  double F = yz-wx;
  double G = zx-wy;
  double H = yz+wx;
  double I = n-x2-y2;
  SWiT_Quad p = SWIT_QUAD(arg2);
  double px = QUAD_X(p);
  double py = QUAD_Y(p);
  double pz = QUAD_Z(p);
  /*  if (pw == 0) {
      return wt();
    } */
  if (n == 0) {
    prim_err_msg = null_quad;
    exn = exn_math;
    return NULL;
  }
  QUAD_X(p) = (px*A + py*B + pz*C)/n;
  QUAD_Y(p) = (px*D + py*E + pz*F)/n;
  QUAD_Z(p) = (px*G + py*H + pz*I)/n;
  for (i = 0; i < argc; i++) {
    SWiT_Quad p = SWIT_QUAD(*argv++);
    double px = QUAD_X(p);
    double py = QUAD_Y(p);
    double pz = QUAD_Z(p);
    QUAD_X(p) = (px*A + py*B + pz*C)/n;
    QUAD_Y(p) = (px*D + py*E + pz*F)/n;
    QUAD_Z(p) = (px*G + py*H + pz*I)/n;
  }
  return swit_true;
}


/*-------------------------------------------------------------------------
 *  Predicate : true if a point (arg3) lies inside the semi-space defined
 *  by a normal (arg1) and a point (arg2), returns the null list if on the
 *  separation plane, and false otherwise (i.e. inside the other semi-plane).
 *  Used for culling.
 *-------------------------------------------------------------------------
 */
static SWiT
externp_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  SWiT_Quad N = SWIT_QUAD(arg1);
  SWiT_Quad A = SWIT_QUAD(arg2);
  SWiT_Quad B = SWIT_QUAD(arg3);
  if ((QUAD_W(N) != 0) || (QUAD_W(A) == 0) || (QUAD_W(B) == 0)) {
    return wt();
  } else {
    double w0 = QUAD_W(A);
    double w1 = QUAD_W(B);
    double p = QUAD_X(N)*(w0*QUAD_X(B)-w1*QUAD_X(A)) +
      QUAD_Y(N)*(w0*QUAD_Y(B)-w1*QUAD_Y(A)) +
      QUAD_Z(N)*(w0*QUAD_Z(B)-w1*QUAD_Z(A));
    if (w0*w1 < 0) p = -p;
    return (p > 0 ? swit_true : (p == 0 ? swit_null_list : swit_false));
  }
}



/*===========================================================
 *                          QUATERNIONS
 *===========================================================
 */

static SWiT
quat_mult_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad q1 = SWIT_QUAD(arg1);
  double x1 = QUAD_X(q1);
  double y1 = QUAD_Y(q1);
  double z1 = QUAD_Z(q1);
  double w1 = QUAD_W(q1);
  SWiT_Quad q2 = SWIT_QUAD(arg2);
  double x2 = QUAD_X(q2);
  double y2 = QUAD_Y(q2);
  double z2 = QUAD_Z(q2);
  double w2 = QUAD_W(q2);
  return swit_make_quad(make_quad(x1*x2-y1*y2-z1*z2-w1*w2,
		    x1*y2+y1*x2+z1*w2-z2*w1,
		    x1*z2+z1*x2-y1*w2+y2*w1,
		    x1*w2+w1*x2+y1*z2-y2*z1));
}

static SWiT
quat_conj_prim (SWiT arg)
{
  SWiT_Quad q = SWIT_QUAD(arg);
  return swit_make_quad(make_quad(QUAD_X(q), -QUAD_Y(q),
				  -QUAD_Z(q), -QUAD_W(q)));
}

static SWiT
mk_rotation_prim (SWiT arg1, SWiT arg2)
{
  double t = to_double(arg1);
  double s = sin(t);
  SWiT_Quad v = SWIT_QUAD(arg2);
  if (QUAD_W(v) != 0) {
    return wt();
  } else {
    double x = QUAD_X(v);
    double y = QUAD_Y(v);
    double z = QUAD_Z(v);
    double n = sqrt(x*x+y*y+z*z);
    if (n == 0) {
      prim_err_msg = null_vect3;
      exn = exn_math;
      return NULL;
    }
    return swit_make_quad(make_quad(n*cos(t), s*x, s*y, s*z));
  }
}

/*===========================================================
 *                         3D VECTORS
 *===========================================================
 */

/* Returns v = p2-p1 */
static SWiT_Quad
points_to_vect3 (SWiT_Quad p1, SWiT_Quad p2)
{
  double w1 = QUAD_W(p1);
  double w2 = QUAD_W(p2);
  double w = w1*w2;
  if (w == 0) {
    return wt();
  }
  return make_quad((QUAD_X(p2)*w1-QUAD_X(p1)*w2)/w,
		   (QUAD_Y(p2)*w1-QUAD_Y(p1)*w2)/w,
		   (QUAD_Z(p2)*w1-QUAD_Z(p1)*w2)/w,
		   0);
}

static SWiT
points_to_vect3_prim (SWiT arg1, SWiT arg2)
{
  return swit_make_quad(points_to_vect3(SWIT_QUAD(arg1), SWIT_QUAD(arg2)));
}

/* Normalize a vect3 */
static SWiT
vect3_norm_prim (SWiT arg)
{
  SWiT_Quad v = SWIT_QUAD(arg);
  if (QUAD_W(v) != 0) {
    return wt();
  } else {
    double x = QUAD_X(v);
    double y = QUAD_Y(v);
    double z = QUAD_Z(v);
    double n = sqrt(x*x+y*y+z*z);
    if (n == 0) {
      prim_err_msg = null_vect3;
      exn = exn_math;
      return NULL;
    } else {
      return swit_make_quad(make_quad(x/n, y/n, z/n, 0));
    }
  }
}

/* Normalize a vect3 (destructive) */
static SWiT
vect3_nnorm_prim (SWiT arg)
{
  SWiT_Quad v = SWIT_QUAD(arg);
  if (QUAD_W(v) != 0) {
    return wt();
  } else {
    double x = QUAD_X(v);
    double y = QUAD_Y(v);
    double z = QUAD_Z(v);
    double n = sqrt(x*x+y*y+z*z);
    if (n == 0) {
      prim_err_msg = null_vect3;
      exn = exn_math;
      return NULL;
    }
    QUAD_X(v) /= n;
    QUAD_Y(v) /= n;
    QUAD_Z(v) /= n;
    return arg;
  }
}

/* returns the dot product of v1 and v2 */
static double
vect3_dot (SWiT_Quad v1, SWiT_Quad v2)
{
  return ((QUAD_X(v1)*QUAD_X(v2)+
	   QUAD_Y(v1)*QUAD_Y(v2)+
	   QUAD_Z(v1)*QUAD_Z(v2)));
}

static SWiT
vect3_dot_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad v1 = SWIT_QUAD(arg1);
  SWiT_Quad v2 = SWIT_QUAD(arg2);
  if ((QUAD_W(v1) != 0) || (QUAD_W(v2) != 0)) {
    return wt();
  }
  return swit_make_inexact(vect3_dot(v1, v2));
}


/* returns the vector product of v1 and v2 */
static SWiT
vect3_mult_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad v1 = SWIT_QUAD(arg1);
  SWiT_Quad v2 = SWIT_QUAD(arg2);
  if ((QUAD_W(v1) != 0) || (QUAD_W(v2) != 0)) {
    return wt();
  } else {
    double x1 = QUAD_X(v1);
    double x2 = QUAD_X(v2);
    double y1 = QUAD_Y(v1);
    double y2 = QUAD_Y(v2);
    double z1 = QUAD_Z(v1);
    double z2 = QUAD_Z(v2);
    return swit_make_quad(make_quad(y1*z2 - z1*y2, x2*z1 - x1*z2,
				    x1*y2 - x2*y1, 0));
  }
}

/* returns v = v2 - v1 */
static SWiT
vect3_sub_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad v1 = SWIT_QUAD(arg1);
  SWiT_Quad v2 = SWIT_QUAD(arg2);
  if ((QUAD_W(v1) != 0) || (QUAD_W(v2) != 0)) {
    return wt();
  }
  return swit_make_quad(make_quad(QUAD_X(v2) - QUAD_X(v1),
		    QUAD_Y(v2) - QUAD_Y(v1),
		    QUAD_Z(v2) - QUAD_Z(v1), 0));
}

/* returns v = v1 + v2 */
static SWiT
vect3_add_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Quad v1 = SWIT_QUAD(arg1);
  SWiT_Quad v2 = SWIT_QUAD(arg2);
  if ((QUAD_W(v1) != 0) || (QUAD_W(v2) != 0)) {
    return wt();
  }
  return swit_make_quad(make_quad(QUAD_X(v2) + QUAD_X(v1),
		    QUAD_Y(v2) + QUAD_Y(v1),
		    QUAD_Z(v2) + QUAD_Z(v1), 0));
}

/*=====================================================
 *                 INITIALISATION
 *=====================================================
 */
void
swit_init_quad (void)
{
  read_table['{']= read_quad;
  print_table[QUAD>>1] = print_quad;
  equal_table[QUAD>>1] = equal_quad;
  swit_quad = make_type("<quad>");

  /* quads */
  swit_make_prim1("quad?", quad_p_prim,
		  ANY);
  swit_make_prim2n("quad", make_quad_prim,
		   NUM_M, NUM_M, NUM_M, NUM_M);
  swit_make_prim2("quad-equal?", quad_equal_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2("quad-ref", quad_ref_prim,
		  QUAD_M, FIXNUM);
  swit_make_prim3("quad-set!", quad_set_prim,
		  QUAD_M, FIXNUM, NUM_M);
  swit_make_prim1("quad-copy", quad_copy_prim,
		  QUAD_M);
  swit_make_prim1("quad-length", quad_length_prim,
		  QUAD_M);
  swit_make_prim2("quad-scale", quad_scale_prim,
		  NUM_M, QUAD_M);
  swit_make_prim2n("quad-scale!", quad_nscale_prim,
		   NUM_M, QUAD_M, QUAD_M, QUAD_M);
  swit_make_prim1 ("get-qmark", get_qmark_prim,
		   QUAD_M);
  swit_make_prim2 ("set-qmark!", set_qmark_prim,
		   QUAD_M, FIXNUM);

  /* points */
  swit_make_prim2("point-distance", distance_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2n("point-translate!", ntranslate_prim,
		   QUAD_M, QUAD_M, QUAD_M, QUAD_M);
  swit_make_prim2("point-translate", translate_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2("point-rotate", rotate_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2n("point-rotate!", nrotate_prim,
		   QUAD_M, QUAD_M, QUAD_M, QUAD_M);
  swit_make_prim2("points->vect3", points_to_vect3_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim3("point-line", point_line_prim,
		  QUAD_M, QUAD_M, NUM_M);

  /* 3D-vectors */
  swit_make_prim2("vect3-add", vect3_add_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2("vect3-sub", vect3_sub_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim2("vect3-mult", vect3_mult_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim1("vect3-normalize", vect3_norm_prim,
		  QUAD_M);
  swit_make_prim1("vect3-normalize!", vect3_nnorm_prim,
		  QUAD_M);
  swit_make_prim2("vect3-dot", vect3_dot_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim3("extern?", externp_prim,
		  QUAD_M, QUAD_M, QUAD_M);

  /* quaternions */
  swit_make_prim2("quater-mult", quat_mult_prim,
		  QUAD_M, QUAD_M);
  swit_make_prim1("quater-conjugate", quat_conj_prim,
		  QUAD_M);
  swit_make_prim2("make-rotation", mk_rotation_prim,
		  NUM_M, QUAD_M);
}
