/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *                            QUEUES
 *==================================================================
 *   queue?                   OBJ                     ->  BOOLEAN
 *   queue                    OBJ * ...               -> QUEUE
 *   queue-empty?             QUEUE                   -> BOOLEAN
 *   queue-front              QUEUE                   -> OBJ
 *   queue-rear               QUEUE                   -> OBJ
 *   queue-pop!, dequeue      QUEUE                   -> OBJ
 *   enqueue!                 QUEUE X OBJ X OBJ X ... -> QUEUE
 *   queue-push!              QUEUE X OBJ X OBJ X ... -> QUEUE
 *   queue-copy               QUEUE                   -> QUEUE
 *   queue-rotate!            QUEUE X NUMBER          -> QUEUE
 *   queue-length             QUEUE                   -> NUMBER
 *   queue-ref                QUEUE X NUMBER          -> OBJ
 *   queue-set!               QUEUE X NUMBER X OBJ    -> OBJ
 *   queue-search             QUEUE X PROCEDURE       -> OBJ
 *   queue->insert!           QUEUE X LIST            -> QUEUE
 *   queue->list              QUEUE                   -> LIST
 *   queue->list!             QUEUE                   -> LIST
 *   list->queue              LIST                    -> QUEUE
 *   queue-append!            QUEUE X QUEUE X ...     -> QUEUE
 *   queue-reverse!           QUEUE                   -> QUEUE
 *==================================================================
 */

#include "swit.h"

/*==================================================================
 * forward declarations of procedures defined in this file
 *==================================================================
 */
static SWiT queue_p_prim (SWiT arg);
static SWiT queue_prim (int argc, SWiT *argv);
static SWiT queue_void_p_prim (SWiT arg);
static SWiT queue_first_prim (SWiT arg);
static SWiT queue_last_prim (SWiT arg);
static SWiT queue_pop_prim (SWiT arg);
static SWiT queue_push_front_prim (SWiT arg1, SWiT arg2, int argv, SWiT* arg3);
static SWiT queue_push_rear_prim (SWiT arg1, SWiT arg2, int argv, SWiT* arg3);
static SWiT queue_copy_prim (SWiT arg);
static SWiT queue_rotate_prim (SWiT arg1, SWiT arg2);
static SWiT queue_length_prim (SWiT arg);
static SWiT queue_ref_prim (SWiT arg1, SWiT arg2);
static SWiT queue_set_prim (SWiT arg1, SWiT arg2, SWiT arg3);
static SWiT queue_search_prim (SWiT arg1, SWiT arg2);
static SWiT queue_to_list_prim (SWiT arg);
static SWiT list_to_queue_prim (SWiT arg);
static SWiT queue_nreverse_prim (SWiT arg);
static SWiT queue_append_prim (SWiT arg, int argc, SWiT *argv);

/*==================================================================
 *           DEFINITION AND CONSTRUCTION OF QUEUES
 *==================================================================
 */
#define SWIT_QUEUE(obj) ((struct SWiT_Queue *)((obj)->data))
#define QUEUE_HEAD(obj) ((obj)->head)
#define QUEUE_TAIL(obj) ((obj)->tail)

struct SWiT_Queue
{
  SWiT head;
  SWiT tail;
};
typedef struct SWiT_Queue *SWiT_Queue;

static SWiT
make_queue (SWiT head, SWiT tail)
{
  SWiT obj = swit_alloc_object ();
  SWiT_Queue queue = SWIT_ALLOC(Queue);
  QUEUE_HEAD(queue) = head;
  QUEUE_TAIL(queue) = tail;
  SWIT_TYPE(obj) = QUEUE;
  SWIT_QUEUE (obj) = queue;
  return (obj);
}

/*==============================================================
 *                          TYPE
 *==============================================================
 */
SWiT swit_queue;


/*==================================================================
 *                     SYNTAX : #[ <datum>* ]
 *==================================================================
 */
SWiT
read_queue ()
{
  int token = get_tok();
  SWiT head = NULL, tail = NULL;
  if (token != ']') {
    head = swit_make_pair(rd_datum(token), NULL);
    tail = head;
    while ((token = get_tok()) != ']') {
      SWiT new = swit_make_pair(rd_datum(token), NULL);
      SET_CDR(tail, new);
      tail = new;
    }
  }
  return (make_queue(head, tail));
}

static int
equal_queue (SWiT q1, SWiT q2)
{
  SWiT lst1 = QUEUE_HEAD(SWIT_QUEUE(q1));
  SWiT lst2 = QUEUE_HEAD(SWIT_QUEUE(q2));
  while (lst1) {
    if (lst2) {
      if (swit_equal(SWIT_CAR(lst1), SWIT_CAR(lst2))) {
	lst1 = SWIT_CDR(lst1);
	lst2 = SWIT_CDR(lst2);
      } else {
	return 0;
      }
    } else {
      return 0;
    }
  }
  return (lst2 ? 0 : 1);
}

static void
print_queue (SWiT queue, int esc, int tcl, SWiT fv, int depth)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(queue));
  depth++;
  str2buff("#[");
  if (!car) {
    char2buff(']');
  } else {
    print_to_buffer(SWIT_CAR(car), esc, tcl, fv, depth);
    car = SWIT_CDR(car);
    while (car) {
      space2buff();
      print_to_buffer(SWIT_CAR(car), esc, tcl, fv, depth);
      car = SWIT_CDR(car);
    }
    char2buff(']');
  }
}


/*==================================================================
 *                         PRIMITIVES
 *==================================================================
 */
static SWiT
queue_p_prim (SWiT arg)
{
    return (SWIT_QUEUEP(arg) ? swit_true : swit_false);
}

static SWiT
make_queue_prim (SWiT len, SWiT fill)
{
  int l = SWIT_FIXNUM(len);
  if (l > 0) {
    int i;
    SWiT last, first = swit_make_pair(fill, NULL);
    last = first;
    for (i = 1; i < l; i++) {
      SWiT new = swit_make_pair(fill, NULL);
      SET_CDR(last, new);
      last = new;
    }
    return (make_queue(first, last));
  } else if (l == 0) {
    return (make_queue(NULL, NULL));
  } else {
    exn = exn_type;
    prim_err_msg = wrong_type;
    return NULL;
  }
}

static SWiT
queue_prim (int argc, SWiT *argv)
{
  if (argc != 0) {
    int i;
    SWiT last, first = swit_make_pair(argv[0], NULL);
    last = first;
    for (i = 1; i < argc; i++) {
      SWiT new = swit_make_pair(argv[i], NULL);
      SET_CDR(last, new);
      last = new;
    }
    return (make_queue(first, last));
  } else {
    return (make_queue(NULL, NULL));
  }
}

static SWiT
queue_void_p_prim (SWiT arg)
{
  return (QUEUE_HEAD(SWIT_QUEUE(arg)) ? swit_false : arg);
}

static SWiT
queue_first_prim (SWiT obj)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(obj));
  if (car) {
    return (SWIT_CAR(car));
  } else {
    prim_err_msg = empty_queue;
    exn = exn_type;
    return NULL;
  }
}

static SWiT
queue_last_prim (SWiT obj)
{
  SWiT cdr = QUEUE_TAIL(SWIT_QUEUE(obj));
  if (cdr) {
    return (SWIT_CAR(cdr));
  } else {
    prim_err_msg = empty_queue;
    exn = exn_type;
    return NULL;
  }
}

static SWiT
queue_pop_prim (SWiT arg)
{
  SWiT_Queue queue = SWIT_QUEUE(arg);
  SWiT car = QUEUE_HEAD(queue);
  if (car == QUEUE_TAIL(queue)) {
    if (car) {
      QUEUE_HEAD(queue) = NULL;
      QUEUE_TAIL(queue) = NULL;
    } else {
      prim_err_msg = empty_queue;
      exn = exn_type;
      return NULL;
    }
  } else {
    QUEUE_HEAD(queue) = SWIT_CDR(car);
  }
  return (SWIT_CAR(car));
}

static SWiT
queue_push_front_prim (SWiT arg, SWiT obj, int argc, SWiT *argv)
{
  SWiT_Queue queue = SWIT_QUEUE(arg);
  SWiT new = swit_make_pair(obj, QUEUE_HEAD(queue));
  int i;
  QUEUE_HEAD(queue) = new;
  if (!QUEUE_TAIL(queue)) {
    QUEUE_TAIL(queue) = new;
  }
  for (i = 0; i < argc; i++) {
    SWiT new = swit_make_pair(argv[i], QUEUE_HEAD(queue));
    QUEUE_HEAD(queue) = new;
  }
  return (arg);
}

static SWiT
queue_push_rear_prim (SWiT arg, SWiT obj, int argc, SWiT *argv)
{
  SWiT_Queue queue = SWIT_QUEUE(arg);
  SWiT cdr = QUEUE_TAIL(queue);
  SWiT new = swit_make_pair(obj, NULL);
  int i;
  if (cdr) {
    SET_CDR(cdr, new);
  } else {
    QUEUE_HEAD(queue) = new;
  }
  for (i = 0; i < argc; i++) {
    cdr = new;
    new = swit_make_pair(argv[i], NULL);
    SET_CDR(cdr, new);
  }
  QUEUE_TAIL(queue) = new;
  return (arg);
}

static SWiT
queue_copy_prim (SWiT arg)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(arg));
  if (car) {
    SWiT current, rest, first = swit_make_pair(SWIT_CAR(car), NULL);
    current = first;
    rest = SWIT_CDR(car);
    while (rest != NULL) {
      SWiT new = swit_make_pair(SWIT_CAR(car), NULL);
      rest = SWIT_CDR(rest);
      SET_CDR(current, new);
      current = new;
    }
    return (make_queue(first, current));
  } else {
    return (make_queue(NULL, NULL));
  }
}


static void
queue_frotate (SWiT_Queue queue, int n)
{
  SWiT car = QUEUE_HEAD(queue);
  SWiT cdr = QUEUE_TAIL(queue);
  if (car != cdr) {
    int i;
    SET_CDR(cdr, car);
    for (i = 1; i < n; i++) {
      car = SWIT_CDR(car);
    }
    QUEUE_TAIL(queue) = car;
    QUEUE_HEAD(queue) = SWIT_CDR(car);
    SET_CDR(car, NULL);
  }
}

static void
queue_irotate (SWiT_Queue queue, int n)
{
  SWiT car = QUEUE_HEAD(queue);
  SWiT cdr = QUEUE_TAIL(queue);
  if (car != cdr) {
    int i,len = 2;
    SWiT tmp;
    SET_CDR(cdr, car);
    /* compute the length */
    while ((tmp = SWIT_CDR(car)) != cdr) {
      len++;
      car = tmp;
    }
    for (i = 0; i < (len-n+1)%len; i++) {
      car = SWIT_CDR(car);
    }
    QUEUE_TAIL(queue) = car;
    QUEUE_HEAD(queue) = SWIT_CDR(car);
    SET_CDR(car, NULL);
  }
}

static SWiT
queue_rotate_prim (SWiT arg1, SWiT arg2)
{
  SWiT_Queue queue = SWIT_QUEUE(arg1);
  int n = SWIT_FIXNUM(arg2);
  if (n > 0) {
    queue_frotate(queue, n);
  } else {
    queue_irotate(queue, -n);
  }
  return arg1;
}

static SWiT
queue_length_prim (SWiT arg)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(arg));
  int length = 0;
  while (car) {
    length++;
    car = SWIT_CDR(car);
  }
  return (swit_make_fixnum(length));
}

static SWiT
queue_ref_prim (SWiT arg1, SWiT arg2)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(arg1));
  int i;
  int idx = SWIT_FIXNUM(arg2);
  if (idx < 0) {
    goto err;
  }
  if (car) {
    for (i = 0; i < idx; i++) {
      car = SWIT_CDR (car);
      if (!car) {
	goto err;
      }
    }
    return (SWIT_CAR (car));
  } else {
    prim_err_msg = empty_queue;
    exn = exn_type;
    return NULL;
  }
 err:
  exn = exn_range;
  prim_err_msg = index_out_of_range;
  return NULL;
}

static SWiT
queue_set_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  SWiT car = QUEUE_HEAD(SWIT_QUEUE(arg1));
  if SWIT_FIXNUMP(arg2) {
    int i, idx = SWIT_FIXNUM(arg2);
    if (idx < 0) {
      prim_err_msg = negative_index;
      exn = exn_range;
      return NULL;
    }
    if (car) {
      for (i = 0; i < idx; i++) {
	car = SWIT_CDR (car);
	if (!car) {
	  prim_err_msg = index_too_large;
	  exn = exn_range;
	  return NULL;
	}
      }
      SET_CAR(car, arg3);
      return (arg3);
    } else {
      prim_err_msg = empty_queue;
      exn = exn_type;
      return NULL;
    }
  } else if SWIT_PROCP(arg2) {
    while (car) {
      SWiT val = SWIT_CAR(car);
      if (swit_apply (arg2, 1, &val) != swit_false) {
	SET_CAR(car, arg3);
	return arg3;
      }
      car = SWIT_CDR(car);
    }
    return swit_false;
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
}

static SWiT
queue_search_prim (SWiT arg1, SWiT arg2)
{
  SWiT lst = QUEUE_HEAD(SWIT_QUEUE(arg1));
  while (lst) {
    SWiT val = SWIT_CAR(lst);
    SWiT true;
    if ((true = swit_apply (arg2, 1, &val)) != swit_false) {
      return (true ? val : NULL);
    }
    lst = SWIT_CDR(lst);
  }
  return swit_false;
}

static SWiT
queue_to_nlist_prim (SWiT arg)
{
  SWiT_Queue q = SWIT_QUEUE(arg);
  SWiT lst = QUEUE_HEAD(q);
  if (lst) {
    SET_CDR(QUEUE_TAIL(q), swit_null_list);
    QUEUE_HEAD(q) = NULL;
    QUEUE_TAIL(q) = NULL;
    return lst;
  } else {
    return swit_null_list;
  }
}

static SWiT
queue_insert_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  SWiT_Queue q2 = SWIT_QUEUE(arg2);
  SWiT_Queue q1 = SWIT_QUEUE(arg1);
  SWiT l2 = QUEUE_HEAD(q2);
  if SWIT_FIXNUMP(arg3) {
    if (l2) {
      SWiT l1 = QUEUE_HEAD(q1);
      if (l1) {
	int n = SWIT_FIXNUM(arg3);
	if (n == 0) {
	  QUEUE_HEAD(q1) = l2;
	  SET_CDR(QUEUE_TAIL(q2), l1);
	} else {
	  register i;
	  SWiT l3;
	  for (i = 0; i < n-1; i++) {
	    if (l1) {
	      l1 = SWIT_CDR (l1);
	    } else {
	      prim_err_msg = index_too_large;
	      exn = exn_range;
	      return NULL;
	    }
	  }
	  l3 = SWIT_CDR(l1);
	  SET_CDR(l1, l2);
	  SET_CDR(QUEUE_TAIL(q2), l3);
	}
      } else {
	QUEUE_HEAD(q1) = l2;
	QUEUE_TAIL(q1) = QUEUE_TAIL(q2);
      }
      QUEUE_HEAD(q2) = NULL;
      QUEUE_TAIL(q2) = NULL;
    }
    return arg1;
  } else if SWIT_PROCP(arg3) {
    SWiT l1 = QUEUE_HEAD(q1);
    SWiT l0 = l1;
    while (l1) {
      SWiT val = SWIT_CAR(l1);
      SWiT true;
      if ((true = swit_apply (arg3, 1, &val)) != swit_false) {
	if (!true) return NULL;
	if (l1 == l0) {
	  QUEUE_HEAD(q1) = l2;
	  SET_CDR(QUEUE_TAIL(q2), l1);
	} else {
	  SWiT l3 = SWIT_CDR(l0);
	  SET_CDR(l0, l2);
	  SET_CDR(QUEUE_TAIL(q2), l3);
	}
	QUEUE_HEAD(q2) = NULL;
	QUEUE_TAIL(q2) = NULL;
	return arg1;
      }
      l0 = l1;
      l1 = SWIT_CDR(l1);
    }
    SET_CDR(QUEUE_TAIL(q1), l2);
    QUEUE_TAIL(q1) = QUEUE_TAIL(q2);
    QUEUE_HEAD(q2) = NULL;
    QUEUE_TAIL(q2) = NULL;
    return arg1;
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    return NULL;
  }
}


static SWiT
queue_to_list_prim (SWiT arg)
{
  SWiT last, first;
  SWiT lst = QUEUE_HEAD(SWIT_QUEUE(arg));
  if (lst) {
    first = last = swit_make_pair(SWIT_CAR(lst), swit_null_list);
    while ((lst = SWIT_CDR(lst)) != NULL) {
      SWiT new = swit_make_pair(SWIT_CAR(lst), swit_null_list);
      SET_CDR(last, new);
      last = new;
    }
    return (first);
  } else {
    return swit_null_list;
  }
}

static SWiT
list_to_queue_prim (SWiT lst)
{
  if (lst == swit_null_list) {
    return (make_queue(NULL, NULL));
  } else {
    int len = swit_list_length(lst);
    if (len < 0) {
#ifdef STRICT
      prim_err_msg = proper_list;
      exn = exn_type;
      return NULL;
#endif
      len = -len;
    }
    if (len == 0) {
      prim_err_msg = finite_list;
      exn = exn_type;
      return NULL;
    } else {
      SWiT last, first = swit_make_pair(SWIT_CAR(lst), NULL);
      last = first;
      while (SWIT_PAIRP(lst = SWIT_CDR(lst))) {
	SWiT new = swit_make_pair(SWIT_CAR(lst), NULL);
	SET_CDR(last, new);
	last = new;
      }
      return (make_queue(first, last));
    }
  }
}

static SWiT
queue_append_prim (SWiT arg, int argc, SWiT *argv)
{
  if (argc != 0) {
    int i;
    SWiT_Queue queue = SWIT_QUEUE(arg);
    SWiT last = QUEUE_TAIL(queue);
    for (i = 0; i < argc; i++) {
      SWiT_Queue q = SWIT_QUEUE(argv[i]);
      SET_CDR(last, QUEUE_HEAD(q));
      last = QUEUE_TAIL(q);
    }
    QUEUE_TAIL(queue) = last;
    /* make all intermediate queues void */
    for (i = 0; i < argc; i++) {
      SWiT_Queue q = SWIT_QUEUE(argv[i]);
      QUEUE_HEAD(q) = NULL;
      QUEUE_TAIL(q) = NULL;
    }
  }
  return (arg);
}

static SWiT
queue_nreverse_prim (SWiT arg)
{
  SWiT_Queue queue = SWIT_QUEUE(arg);
  SWiT lst = QUEUE_HEAD(queue);
  if (lst) {
    SWiT first = lst;
    SWiT next = SWIT_CDR(lst);
    SET_CDR(lst, NULL);
    while (next != NULL) {
      SWiT tmp = SWIT_CDR(next);
      SET_CDR(next, first);
      first = next;
      next = tmp;
    }
    QUEUE_HEAD(queue) = first;
    QUEUE_TAIL(queue) = lst;
  }
  return arg;
}




/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_queue (void)
{
  read_table[RD_QUEUE] = read_queue;
  print_table[QUEUE>>1] = print_queue;
  equal_table[QUEUE>>1] = equal_queue;
  swit_queue = make_type("<queue>");

  swit_make_prim1 ("queue?", queue_p_prim,
		   ANY);
  swit_make_prim12 ("make-queue", make_queue_prim,
		    FIXNUM, ANY, swit_null_list);
  swit_make_prim0n ("queue", queue_prim,
		    ANY, ANY);
  swit_make_prim1 ("queue-empty?", queue_void_p_prim,
		   QUEUE);
  swit_make_prim1 ("queue-front", queue_first_prim,
		   QUEUE);
  swit_make_prim1 ("queue-rear", queue_last_prim,
		   QUEUE);
  swit_make_prim1 ("queue-pop!", queue_pop_prim,
		   QUEUE);
  swit_make_prim1 ("dequeue!", queue_pop_prim,
		   QUEUE);
  swit_make_prim2n ("queue-push!", queue_push_front_prim,
		    QUEUE, ANY, ANY, ANY);
  swit_make_prim2n ("enqueue!", queue_push_rear_prim,
		    QUEUE, ANY, ANY, ANY);
  swit_make_prim1 ("queue-copy", queue_copy_prim,
		    QUEUE);
  swit_make_prim12 ("queue-rotate!", queue_rotate_prim,
		    QUEUE, FIXNUM, swit_one);
  swit_make_prim1 ("queue-length", queue_length_prim,
		   QUEUE);
  swit_make_prim2 ("queue-ref", queue_ref_prim,
		   QUEUE, FIXNUM);
  swit_make_prim3 ("queue-set!", queue_set_prim,
		   QUEUE, ANY, ANY); /* snd arg : index or predicate */
  swit_make_prim3 ("queue-insert!", queue_insert_prim,
		   QUEUE, QUEUE, ANY);
  swit_make_prim2 ("queue-search", queue_search_prim,
		   QUEUE, PROC_M);
  swit_make_prim1 ("queue->list", queue_to_list_prim,
		   QUEUE);
  swit_make_prim1 ("queue->list!", queue_to_nlist_prim,
		   QUEUE);
  swit_make_prim1 ("list->queue", list_to_queue_prim,
		   LIST);
  swit_make_prim1 ("queue-reverse!", queue_nreverse_prim,
		   QUEUE);
  swit_make_prim1n ("queue-append!", queue_append_prim,
		    QUEUE, QUEUE, ANY);
}
