/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*-----------------------------------------------------------
 *                 6.10.2  INPUT
 *
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *    read                     INPUT-PORT -> OBJ
 *    read-char                INPUT-PORT -> CHAR
 *    peek-char                INPUT-PORT -> CHAR
 *    eof-object?              INPUT-PORT -> BOOLEAN
 *    char-ready?              INPUT-PORT -> BOOLEAN
 *
 * Extensions :
 * ^^^^^^^^^^
 *    read-line                INPUT-PORT -> STRING
 *-----------------------------------------------------------
 */


#include "swit.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <fcntl.h>
#include "../readline/readline.h"
#include "../readline/history.h"


extern SWiT cur_in_port;
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
#define in_port PORT_DESCR(current_ip)
#define read_buffer PORT_BUFFER(current_ip)
#define read_cur_idx PORT_INDEX(current_ip)
#define read_top_idx PORT_EOB(current_ip)
#define read_fun PORT_READ_FUN(current_ip)
#define lineno PORT_LINENO(current_ip)
#define ERROR 0

extern SWiT swit_prompt;
extern double big_to_inexact(MP_INT *x); /* defined in switNumber.c */
extern SWiT null_in_port;
extern int transcript_descr;
extern SWiT search_port(int descr);
extern int issocket(int des);


#define MAX_READ_STRING_SIZE 46340  /* MAX_FIXNUM */

/* Input */
static SWiT read_prim (SWiT arg);
static SWiT read_line_prim (SWiT arg);
static SWiT read_char_prim (SWiT arg);
static SWiT peek_char_prim (SWiT arg);
static SWiT eof_object_p_prim (SWiT arg);
static SWiT char_ready_p_prim (SWiT arg);

static SWiT tok_val;
static char text_buffer[MAX_READ_STRING_SIZE];

static int text_index;
static int isseparator (int c);
static void read_eol (void);
static void read_text (char c);
static void read_string (void);
static int read_number (char c);
static void read_exact (char c);
static int read_base (void);
static int read_exactness (void);
static int parse_integer (int base);
static int parse_inexact (int base);

static int read_identifier (int ch);

static unsigned char *line_read = (char *)NULL;
static int nchar = 0;
static line_index = -1;
static unsigned char get_char (void);
static char get_nonsep_char (void);
static char peek_char (void);
static int read_octal (int n);
static int read_hexa (void);
static fd_set rdfds;
static char *default_prompt;

/* globals */
fd_set fds; /* for char-ready? */
int prompt = 1; /* 1 : display the prompt, 0 = no prompt displayed */
char *tok_text = text_buffer;
int read_paren = 0;
int start_string;
int inside_string;

/* (some) keywords */
SWiT swit_quote;
SWiT swit_quasiquote;
SWiT swit_unquote;
SWiT swit_unquote_splicing;
SWiT swit_else;


/*---------------------------------------------------------------------
 * get_tok() returns the next token read (an integer that corresponds
 * to the lexical class), and puts the corresponding scheme object in
 * the (exported) variable 'tok_val'. The string read for this token
 * is sometimes stored in the local buffer 'text_buffer', and can be
 * accessible through the global pointer variable 'tok_text'.
 *
 * It works by dispatching with respect to the the next
 * character of the input (this dispatching mechanism should be
 * improved with syntax-tables, as found in Common Lisp).
 *---------------------------------------------------------------------
 */
int
get_tok (void)
{
  char ch;

 start_over:
  ch = get_char();
  while (isspace(ch)) {
    ch = get_char();
  }
  text_buffer[0] = ch;
  text_buffer[1] = '\0';
  text_index = 1;
  switch (ch)
    {
    case ')':case '[':case '{':case ']':
    case '}':case '(':case '`':case '\'':
      tok_val = swit_make_char(ch);
      return (ch);

      /* string */
    case '"': 
      read_string ();
      if (text_index) {
	char *dup = (char *)swit_malloc_atomic(text_index+1);
	memcpy(dup, tok_text, text_index);
	dup[text_index] = '\0';
	tok_val = swit_make_constant_string(dup, text_index);
      } else {
	tok_val = swit_null_string;
      }
      return LITERAL;
    

      /* '.' can appear in pair syntax if alone,
             can start an inexact number e.g. .12 is valid
	     can start an identifier (if not parse as a number) */
    case '.':
      if (isseparator (peek_char())) {
	tok_val = swit_make_char(ch);
	return (ch);
      } else if (isdigit (peek_char())) {
	read_number(ch);
	return (parse_inexact(10));
      } else {
	int tok = read_identifier(ch);
	return tok;
      }

      /* unquote or unquote-splicing */
    case ',':
      if (peek_char() == '@') {
	ch = get_char();
	return SPLICE;
      } else {
	return (ch);
      }

      /* comment */
    case ';':
      read_eol();
      goto start_over;

      /* the 'dispatching' character */
    case '#':
      if (isspace(peek_char())) {
	/* system command */
	read_eol();
	if (*text_buffer) {
	  tok_val = swit_make_constant_string(tok_text, strlen(tok_text));
	} else {
	  prim_err_msg = wrong_shell_command;
	  exn = exn_read;
	  longjmp(*read_jbuf, 2);
	  return ERROR;
	}
	return SYSTEM;
      } else {
	ch = get_char();
	text_buffer[1] = ch;
	text_buffer[2] = '\0';
	text_index = 2;
	switch (ch)
	  {
	    /* vector */
	  case '(':
	    return VECTOR;

	    /* primitive */
	  case '\'':
	    read_identifier(get_char());
	    return PRIMITIVE;

	    /* special constants: ##eof, ##unbound */
	  case '#': {
	    read_text(get_char());
	    if (strcmp(text_buffer, "eof") == 0) {
	      tok_val = swit_eof;
	      return LITERAL;
	    } else if (strcmp(text_buffer, "unbound") == 0) {
	      tok_val = swit_unbound;
	      return LITERAL;
	    } else {
	      prim_err_msg = bad_literal;
	      exn = exn_read;
	      longjmp(*read_jbuf, 2);
	      return ERROR;
	    }
	  }
	    /* Queues */
	  case '[':
	    return RD_QUEUE;

	    /* #| is 'super' comment terminated with a |# */
	  case '|':
	    inside_string = 1;
	  encore:
	    while (get_char() != '|') {};
	    if (get_char() == '#') {
	      inside_string = 0;
	      goto start_over;
	    }
	    goto encore;

	    /* #! is also a comment (up to end of line) */
	  case '!':
	    read_eol();
	    goto start_over;

	    /* #.<expr> : eval at read-time */
	  case '.':
	    return READ_EVAL;

	    /* character */
	  case '\\': {
	    unsigned char ch1, ch = get_char();
	    if (isseparator(peek_char())) {
	      tok_val = swit_make_char(ch);
	      if (ch == ')') read_paren++;
	      if (ch == '(') read_paren--;
	      return LITERAL;
	    }
	    ch1 = get_char();
	    if (isseparator(peek_char())) {
	      tok_val = swit_make_char(ch*256+ch1);
	      if (ch == ')') read_paren++;
	      if (ch == '(') read_paren--;
	      if (ch1 == ')') read_paren++;
	      if (ch1 == '(') read_paren--;
	      return LITERAL;
	    }
	    read_text(ch1);
	    switch (ch)
	      {
	      case '0':case '1':case '2':case '3': {
		int n;
		/* read an octal character */
		if ((n = read_octal(ch - '0')) != -1) {
		  tok_val = swit_make_char(n);
		  return LITERAL;
		} else {
		  prim_err_msg = bad_octal;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      }
	      case 'x': {
		int n;
		/* read char in hexadecimal (2 or 4 hexa) */
		if ((n = read_hexa()) != -1) {
		  tok_val = swit_make_char(n);
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      }		
	      case 's':
	      case 'S':
		if (strcmp(text_buffer, "pace") == 0) {
		  tok_val = swit_space;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 'r':
	      case 'R':
		if (strcmp(text_buffer, "eturn") == 0) {
		  tok_val = swit_return;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 'f':
	      case 'F':
		if (strcmp(text_buffer, "ormfeed") == 0) {
		  tok_val = swit_formfeed;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 't':
	      case 'T':
		if (strcmp(text_buffer, "ab") == 0) {
		  tok_val = swit_tab;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 'e':
	      case 'E':
		if (strcmp(text_buffer,"scape") == 0) {
		  tok_val = swit_escape;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 'n':
	      case 'N':
		if (strcmp(text_buffer,"ewline") == 0) {
		  tok_val = swit_newline;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      case 'b':
	      case 'B':
		if (strcmp(text_buffer,"ackspace") == 0) {
		  tok_val = swit_backspace;
		  return LITERAL;
		} else {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      default: {
		  prim_err_msg = bad_char_name;
		  exn = exn_read;
		  longjmp(*read_jbuf, 2);
		  return ERROR;
		}
	      }

	    /* boolean */
	  case 't':
	  case 'T':
	    if (isseparator (peek_char())) {
	      tok_val = swit_true;
	      return LITERAL;
	    } else {
	      line_index = nchar;
	      exn = exn_read;
	      prim_err_msg = bad_bool;
	      longjmp(*read_jbuf, 2);
	    }
	  case 'f':
	  case 'F':
	    if (isseparator (peek_char())) {
	      tok_val = swit_false;
	      return LITERAL;
	    } else {
	      line_index = nchar;
	      exn = exn_read;
	      prim_err_msg = bad_bool;
	      longjmp(*read_jbuf, 2);
	    }
	  }
	    /* number */
	  case 'e':
	  case 'E':
	    {
	      int base = read_base();
	      read_exact(get_nonsep_char());
	      return (parse_integer(base));
	    }
	  case 'i':
	  case 'I':
	    {
	      int base = read_base();
	      read_number(get_nonsep_char());
	      return (parse_inexact(base));
	    }
	  case 'D':
	  case 'd':
	    if (read_exactness()) {
	      read_exact(get_nonsep_char());
	      return (parse_integer(10));
	    } else {
	      read_number(get_nonsep_char());
	      return (parse_inexact(10));
	    }
	  case 'X':
	  case 'x':
	    if (read_exactness()) {
	      read_exact(get_nonsep_char());
	      return (parse_integer(16));
	    } else {
	      read_number(get_nonsep_char());
	      return (parse_inexact(16));
	    }
	  case 'O':
	  case 'o':
	    if (read_exactness()) {
	      read_exact(get_nonsep_char());
	      return (parse_integer(8));
	    } else {
	      read_number(get_nonsep_char());
	      return (parse_inexact(8));
	    }
	  case 'B':
	  case 'b':
	    if (read_exactness()) {
	      read_exact(get_nonsep_char());
	      return (parse_integer(2));
	    } else {
	      read_number(get_nonsep_char());
	      return (parse_inexact(2));
	    }

	    /* error */
	  default:
	    line_index = nchar;
	    exn = exn_read;
	    prim_err_msg = unexpected_numbersign;
	    longjmp(*read_jbuf, 2);
	  } /* end of '#' dispatch */

	/* numbers */
      case '+':
      case '-':
	if (isseparator(peek_char())) {
	  tok_val = swit_make_symbol(&ch, 1, swit_unbound);
	  return LITERAL;
	} else {
	  int exact = read_number(ch);
	  if (exact) {
	    return (parse_integer(10));
	  } else {
	    return (parse_inexact(10));
	  }
	}
      default:
	if (isdigit(ch)) {
	  int exact = read_number(ch);
	  if (exact) {
	    return (parse_integer(10));
	  } else {
	    return (parse_inexact(10));
	  }
	} else {
	  return (read_identifier(ch));
	}
      }
    }
}

/* for comments, system and load commands */
static void
read_eol (void)
{
  text_index = 0;
  inside_string = 1; /* don't count parens */
  while (peek_char() != '\n') {
    if (text_index == MAX_READ_STRING_SIZE) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = num_too_long;
      longjmp(*read_jbuf, 2);
    }
    text_buffer[text_index++] = get_char();
  }
  inside_string = 0;
  text_buffer[text_index]='\0';
}

static int
isseparator (int ch)
{
  return ((ch == ' ')
	  || (ch == '(')
	  || (ch == ')')
	  || (ch == '}')
	  || (ch == '{')
	  || (ch == ']')
	  || (ch == '[')
	  || (ch == '"')
	  || (ch == ';')
	  || (ch == '\n')
	  || (ch == '\r')
	  || (ch == '\f')
	  || (ch == '\t')
	  || (ch == '\v')
	  || (ch == '\0') /* for string ports */
	  || (ch == -1));
}

/*
 * `read_text' reads input (by calling get_char()) until a separator
 *  is found and puts the result in text_buffer.
 * The letters are converted to lowercase.
 * The number of characters read (+ 1) is in text_index.
 * This function is called when an identifier or a character name is expected.
 */
static void
read_text (char c)
{
  text_buffer[0] = tolower(c);
  text_index = 1;
  while (!isseparator(peek_char())) {
    if (text_index == MAX_READ_STRING_SIZE) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = num_too_long;
      longjmp(*read_jbuf, 2);
    }
    text_buffer[text_index++] = tolower(get_char());
  }
  text_buffer[text_index] = '\0';
}


/* for reading characters given in octal */
static int
read_octal (int n)
{
  char ch;
  int i;
  for (i = 0; i < 2; i++) {
    ch = text_buffer[i];
    if ((ch >= '0') && (ch < '8')) {
      n = 8*n + ch - '0';
    } else {
      return -1;
    }
  }
  if (text_buffer[2] != '\0') {
    return -1;
  } else {
    return n;
  }
}


/* for reading characters given in hexa */
static int
read_hexa (void)
{
  char ch;
  int i;
  int n = 0;
  for (i = 0; i < 2; i++) {
    ch = text_buffer[i];
    if ((ch >= '0') && (ch <= '9')) {
      n = 16*n + ch - '0';
    } else if ((ch >= 'a') && (ch <= 'f')) {
      n = 16*n + 10 + ch - 'a';
    } else {
      return -1;
    }
  }
  if (text_buffer[2] != '\0') {
    for (i = 2; i < 4; i++) {
      ch = text_buffer[i];
      if ((ch >= '0') && (ch <= '9')) {
	n = 16*n + ch - '0';
      } else if ((ch >= 'a') && (ch <= 'f')) {
	n = 16*n + 10 + ch - 'a';
      } else {
	return -1;
      }
    }
    if (text_buffer[4] != '\0') {
      return -1;
    } else {
      return n;
    }
  } else {
    return n;
  }
}


static int
read_base (void)
{
  if (isseparator (peek_char())) {
    line_index = nchar;
    exn = exn_read;
    prim_err_msg = bad_num;
    longjmp(*read_jbuf, 2);
  } else if (peek_char() == '#') {
    int ch = get_char();
    text_buffer[text_index++] = ch;
    if (isseparator (peek_char())) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = bad_num;
      longjmp(*read_jbuf, 2);
    } else {
      ch = get_char();
      text_buffer[text_index++] = ch;
      text_buffer[text_index] = '\0';
      switch (ch)
	{
	case 'b':
	  return 2;
	case 'o':
	  return 8;
	case 'x':
	  return 16;
	case 'd':
	  return 10;
	default:
	  line_index = nchar;
	  exn = exn_read;
	  prim_err_msg = bad_num;
	  longjmp(*read_jbuf, 2);
	}
    }
  } else {
    return 10;
  }
  return 0; /* avoids a compiler warning */
}

static int
read_exactness (void)
{
  if (isseparator (peek_char())) {
    line_index = nchar;
    exn = exn_read;
    prim_err_msg = bad_num;
    longjmp(*read_jbuf, 2);
  } else if (peek_char() == '#') {
    int ch = get_char();
    text_buffer[text_index++] = ch;
    if (isseparator (peek_char())) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = bad_num;
      longjmp(*read_jbuf, 2);
    } else {
      ch = get_char();
      text_buffer[text_index++] = ch;
      text_buffer[text_index] = '\0';
      switch (ch)
	{
	case 'e':
	  return 1;
	case 'i':
	  return 0;
	default:
	  line_index = nchar;
	  exn = exn_read;
	  prim_err_msg = bad_num;
	  longjmp(*read_jbuf, 2);
	}
    }
  } else {
    return 1;
  }
  return 0; /* avoids a compiler warning */
}

static int
read_identifier (int ch)
{
  char *dup;
  read_text(ch);
  dup = (char *)swit_malloc_atomic(text_index+1);
  memcpy(dup, tok_text, text_index);
  dup[text_index] = '\0';
  tok_val = swit_make_symbol(dup, text_index, swit_unbound);
  return LITERAL;
}


/*---------------------------------------------
 * Refer to R4RS (p32) for the syntax.
 * (rational and complex numbers not yet parsed)
 *---------------------------------------------
 */
#define EXPONENT_MARKER \
 case 'e':case 'E':case 's':case 'S':case 'f':case 'F':case 'd':case 'D':case 'l':case 'L':

static int
read_number (char c)
{
  int state = 0;
  if (c == '+') {
    text_buffer[0] = get_char();
  } else {
    text_buffer[0] = c;
  }
  text_index = 1;
  while (!isseparator(peek_char())) {
    char d = get_char();
    if (text_index == MAX_READ_STRING_SIZE) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = number_too_long;
      longjmp(*read_jbuf, 2);
    }
    switch (state) {
    case 0:
      switch (d) {
      case '#':
        state = 1;
	text_buffer[text_index++] = '0' + rand() % 10;
	break;
      case '.':
	text_buffer[text_index++] = '.';
        state = 4;
	break;
      EXPONENT_MARKER
	text_buffer[text_index++] = 'e';
        state = 3;
	break;
      default: /* stay in state 0 */
	text_buffer[text_index++] = d;
      }
      break;
    case 1:
      switch (d) {
      case '#': /* stay in state 1 */
	text_buffer[text_index++] = '0' + rand() % 10;
	break;
      case '.':
	text_buffer[text_index++] = '.';
        state = 2;
	break;
      EXPONENT_MARKER
	text_buffer[text_index++] = 'e';
        state = 3;
	break;
      default:
	goto error;
      }
      break;
    case 2:
      switch (d) {
      case '#': /* stay in state 2 */
	text_buffer[text_index++] = '0' + rand() % 10;
	break;
      EXPONENT_MARKER
	text_buffer[text_index++] = 'e';
        state = 3;
	break;
      default:
	goto error;
      }
      break;
    case 3:
      if (d != '#') {
	text_buffer[text_index++] = d;
      } else {
	goto error;
      }
      break;
    case 4:
      switch (d)
	{
        EXPONENT_MARKER
	  text_buffer[text_index++] = 'e';
	  state = 3;
	  break;
	case '#':
	  text_buffer[text_index++] = '0' + rand() % 10;
	  state = 2;
	  break;
	default:
	  text_buffer[text_index++] = d;
	}
      break;
    default:
      goto error;
    }
  }
  text_buffer[text_index] = '\0';
  return (!state);
 error:
  text_buffer[text_index] = '\0';
  line_index = nchar;
  exn = exn_read;
  prim_err_msg =  bad_num;
  longjmp(*read_jbuf, 2);
  return 0;
}


/*
 * 'read_exact' reads a sequence of characters
 *  and checks if no '#' are present.
 * The rest of parsing is done in the function 'parse_integer'.
 */
static void
read_exact (char c)
{
  if (c == '+')
    text_buffer[0] = get_char();
  else
    text_buffer[0] = c;
  text_index = 1;
  while (!isseparator(peek_char())) {
    if (text_index == MAX_READ_STRING_SIZE) {
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = number_too_long;
      longjmp(*read_jbuf, 2);
    }
    if ((text_buffer[text_index++] = get_char()) == '#') {
      text_buffer[text_index]='\0';
      line_index = nchar;
      exn = exn_read;
      prim_err_msg = bad_exact_num;
      longjmp(*read_jbuf, 2);
    }
  }
  text_buffer[text_index]='\0';
}


/*
 * A first double quote has been read when this function is called.
 *  A \0 terminated string is returned without its leading double quote.
 */
static void
read_string (void)
{
  start_string = lineno;
  inside_string = 1;
  text_index = 0;
  while (inside_string) {
    char ch;
    if (text_index == MAX_READ_STRING_SIZE) {
      line_index = nchar;
      inside_string = 0;
      exn = exn_read;
      prim_err_msg = string_too_long;
      longjmp(*read_jbuf, 2);
    }
    ch = get_char();
    switch (ch)
      {
      case '\\':
	ch = get_char();
	switch (ch)
	  {
	  case '0':case '1':case '2':case '3':
	    /* read an octal character */
	    {
	      int i, n = ch - '0';
	      for (i = 1; i < 3; i++) {
		ch = get_char();
		if ((ch >= '0') && (ch < '8')) {
		  n = 8*n + ch - '0';
		} else {
		  inside_string = 0;
		  line_index = nchar;
		  exn = exn_read;
		  prim_err_msg = bad_octal;
		  longjmp(*read_jbuf, 2);
		}
	      }
	      text_buffer[text_index++] = n;
	      break;
	    }
	  case 'n' :
	    text_buffer[text_index++] = '\n'; /* newline */
	    break;
	  case 't' :
	    text_buffer[text_index++] = '\t'; /* tab */
	    break;
	  case 'f' :
	    text_buffer[text_index++] = '\f'; /* formfeed */
	    break;
	  case 'r' :
	    text_buffer[text_index++] = '\r'; /* return */
	    break;
	  case 'b' :
	    text_buffer[text_index++] = '\b'; /* backspace */
	    break;
	  case 'e' :
	    text_buffer[text_index++] = '\e'; /* escape */
	    break;
	  case '\n' :  /* a '\' followed with a newline = ignored  */
	    break;     /*   (useful to break a long line)          */
	  default:
	    text_buffer[text_index++] = ch;
	  }
	break;
      case '"':
	inside_string = 0;
	break;
      default:
	text_buffer[text_index++] = ch;
      }
  }
  text_buffer[text_index]='\0';
}


/*
 * Tries to parse an integer (fixnum or bignum) from text_buffer.
 */
static int
parse_integer (int base)
{
  MP_INT *val = (MP_INT *)swit_malloc(sizeof(MP_INT));
  if (mpz_init_set_str(val, text_buffer, base) == -1){
    mpz_clear(val);
    line_index = nchar;
    exn = exn_read;
    prim_err_msg = int_expected;
    longjmp(*read_jbuf, 2);
    return 0;
  } else {
    if (mpz_size(val) > 1) {
      tok_val = swit_make_bignum(val);
    } else {
      unsigned long int m = mpz_get_ui(val);
      if (m >> 15) {
	tok_val = swit_make_bignum(val);
      } else {
	tok_val = swit_make_fixnum(mpz_get_si(val));
	mpz_clear(val);
      }
    }
    return LITERAL;
  }
}

static int
parse_inexact (int base)
{
  MP_INT val;
  if (mpz_init_set_str(&val, text_buffer, base) == -1){
    mpz_clear(&val);
    if (base == 10) {
      char *tail;
      /* strtod does not work with on SUN(bsd) : use LINUX !!! */
#ifdef LINUX
      double d = strtod(text_buffer, &tail);
      if (*tail) {
	goto error;
      } else {
	tok_val = swit_make_inexact(d);
	return LITERAL;
      }
/* BSD : a poor replacement of the above code */
#else
      double d = atof(text_buffer);
      if (d == 0.0) {
	double d1;
	char c = text_buffer[0];
	text_buffer[0]='1';
	d1 = atof(text_buffer);
	if (d1 == 0.0) {
	  text_buffer[0]=c;
	  goto error;
	}
	tok_val = swit_make_inexact(d);
	return LITERAL;
      } else {
	tok_val = swit_make_inexact(d);
	return LITERAL;
      }
#endif
    } else {
      goto error;
    }
  } else {
    double x = big_to_inexact(&val);
    mpz_clear(&val);
    tok_val = swit_make_inexact(x);
    return LITERAL;
  }
 error:
  line_index = nchar;
  exn = exn_read;
  prim_err_msg = bad_num;
  longjmp(*read_jbuf, 2);
  return 0;
}

static char
get_nonsep_char (void)
{
  if (isseparator(peek_char())) {
    line_index = nchar;
    exn = exn_read;
    prim_err_msg = bad_num;
    longjmp(*read_jbuf, 2);
  } else {
    return get_char();
  }
  return 0;
}

/*
 * Get the next character of the input stream.
 * When input is interactive, readline() is used,
 *   otherwise, input is buffered.
 */
int raw_mode;
extern int history_length;
static unsigned char
get_char (void)
{
  int descr = in_port;
  char c;
  if (isatty(descr)) {
    if (raw_mode) {
      char d;
      read(descr, &c, 1);
      d = c;
      while ((char)c != '\n') {
	read(descr, &c, 1);
      }
      return d;
    }
    if ((line_index == nchar)
	|| (line_read == (unsigned char *) NULL)) {
      if (line_read) {
	free(line_read);
	line_read = (char *) NULL;
      } else {
	putchar('\n');
      }
      if (read_paren > 0) {
	fprintf(stdout, "%c", read_paren + '0');
      }
      if (read_paren < 0) {
	read_paren = 0 ;
      }
      line_read = "\n";
      line_index = -1;
      nchar = 0;
      return ('\n');
    } else {
      if (line_index == -1) {
	SWiT prompt_val = SYMBOL_VAL(SWIT_SYM(swit_prompt));
	char *prompt_str;
	if SWIT_STRINGP(prompt_val) {
	  prompt_str = STRING_ELS(prompt_val);
	} else {
	  prompt_str = "> ";
	} 
	
	if (prompt) {
	  line_read = readline(prompt_str);
	} else {
	  line_read = readline("");
	}

	if (transcript_descr != -1) {
	  if (prompt) write(transcript_descr, prompt_str, strlen(prompt_str));
	  write(transcript_descr, line_read, strlen(line_read));
	  write(transcript_descr, "\n", 1);
	}
	if (line_read && *line_read) {
	  if ((history_length == 0)
	      || (history_search_prefix(line_read, -1) !=  history_length-1)) {
	    add_history(line_read);
	  }

	  nchar = strlen(line_read);
	  line_read[nchar] = '\n';
	  line_index = 0;
	  c = line_read[line_index++];
	  if ((c == '(') && !inside_string) read_paren++;
	  if ((c == ')') && !inside_string) read_paren--;
	  return (c);
	} else {
	  if (read_paren > 0) {
	    fprintf(stdout, "%c", read_paren + '0');
	  }
	  if (read_paren < 0) {
	    read_paren = 0 ;
	  }
	  return ('\n');
	}
      } else {
	c = line_read[line_index++];
	if ((c == '(') && !inside_string) read_paren++;
	if ((c == ')') && !inside_string) read_paren--;
	return (c);
      }
    }
  } else {
    if (descr == -1) {
      prim_err_msg = closed_port;
      exn = exn_read;
      longjmp(*read_jbuf, 2); /* closed port */ 
    }
    if (read_cur_idx == read_top_idx) {
      /* read more char into read_buffer */
      int cnt;
      if ((cnt = (*read_fun)(descr, read_buffer, IO_BUFFER_SIZE)) != -1) {
	if (cnt) {
	  read_top_idx = cnt;
	  read_cur_idx = 0;
	} else {
	  longjmp(*read_jbuf, 1); /* EOF */
	}
      } else {
	prim_err_msg = sys_errlist[errno];
	exn = exn_read;
	longjmp(*read_jbuf, 2); /* closed port */ 
      }
    }
    c = read_buffer[read_cur_idx++];
    if ((c == '(') && !inside_string) read_paren++;
    if ((c == ')') && !inside_string) read_paren--;
    if (c == '\n') lineno++;
    return c;
  }
}

static char
peek_char (void)
{
  int descr = in_port;
  if (descr == -1) {
    exn = exn_read;
    prim_err_msg = closed_port;
    longjmp(*read_jbuf, 2);
  }
  if (isatty (descr)) {
    return line_read[line_index];
  }
  if (read_cur_idx == read_top_idx) {
    int cnt;
    /* socket ? */
    if (issocket(descr)) {
      exn = exn_read;
      prim_err_msg = cannot_peek;
      longjmp(*read_jbuf, 2);
    }
    /* read more char into read_buffer */
    if ((cnt = (*read_fun)(descr, read_buffer, IO_BUFFER_SIZE)) != -1) {
      if (cnt) {
	read_top_idx = cnt;
	read_cur_idx = 0;
      } else {
	longjmp(*read_jbuf, 1); /* EOF */
      }
    } else {
      prim_err_msg = sys_errlist[errno];
      exn = exn_read;
      longjmp(*read_jbuf, 2);
    }
  }
  return read_buffer[read_cur_idx];
}



typedef SWiT(* Read_Fun)(void);

static SWiT
read_literal (void)
{
  return tok_val;
}

static SWiT
read_quote (void)
{
  return swit_make_pair(swit_quote,
			swit_make_pair(rd_datum(get_tok()),
				       swit_null_list));
}

static SWiT
read_quasiquote (void)
{
  return swit_make_pair(swit_quasiquote,
			swit_make_pair(rd_datum(get_tok()),
				       swit_null_list));
}

static SWiT
read_unquote (void)
{
  return swit_make_pair(swit_unquote,
			swit_make_pair(rd_datum(get_tok()),
				       swit_null_list));
}

static SWiT
read_splice (void)
{
  return swit_make_pair(swit_unquote_splicing,
			swit_make_pair(rd_datum(get_tok()),
				       swit_null_list));
}

SWiT
rd_datum (int token)
{
  return ((Read_Fun)(read_table[token]))();
}



/*
 *================================================================
 *                   6.10.2.   INPUT PRIMITIVES
 *================================================================
 */

static SWiT
read_prim (SWiT arg)
{
  SWiT ret;
  int token, err;
  SWiT old_port = cur_in_port;

  if (arg != null_in_port) {
    cur_in_port = arg;
  }
  if ((err = setjmp(*read_jbuf))) {
    cur_in_port = old_port;
    prompt = 1;
    if (err == 1) {
      return swit_eof;
    } else {
      return NULL;
    }
  }
  prompt = 0;
  token = get_tok();
  /*  if (token == MY_EOF) return swit_eof;*/
  ret = rd_datum(token);
  cur_in_port = old_port;
  prompt = 1;
  return(ret);
}

extern SWiT_Ccont done_cont;
static SWiT
read_eval (void)
{
  return swit_eval(read_program(get_tok()), NULL, done_cont);
}

static SWiT
read_char_prim (SWiT arg)
{
  unsigned char ch;
  int err;
  SWiT old_port = cur_in_port;
  if (arg != null_in_port) {
    cur_in_port = arg;
  }
  if ((err = setjmp(*read_jbuf))) {
    prompt = 1;
    raw_mode = 0;
    cur_in_port = old_port;
    if (err == 1) {
      return swit_eof;
    } else {
      return NULL;
    }
  }
  prompt = 0;
  line_read = "\n";
  line_index = -1;
  nchar = 0;
  raw_mode = 1;
  ch = get_char();
  prompt = 1;
  raw_mode = 0;
  cur_in_port = old_port;
  return swit_make_char(ch);
}

static SWiT
peek_char_prim (SWiT arg)
{
  unsigned char ch;
  int err;
  SWiT old_port = cur_in_port;
  if (arg != null_in_port) {
    cur_in_port = arg;
  }
  if ((err = setjmp(*read_jbuf))) {
    cur_in_port = old_port;
    if (err == 1) {
      return swit_eof;
    } else {
      return NULL;
    }
  }
  ch = peek_char();
  cur_in_port = old_port;
  return swit_make_char(ch);
}

static SWiT
eof_object_p_prim (SWiT arg)
{
  return (SWIT_EOFP(arg) ? swit_true : swit_false);
}


static struct timeval zero;

static SWiT
char_ready_p_prim (SWiT arg)
{
  SWiT_Port p;
  int des;
  if (arg == null_in_port) arg = cur_in_port;
  p = SWIT_PORT(arg);
  des = PORT_DESCR(p);
  if (isatty(des)) return swit_false; /* no raw mode */
  if (des == -1) {
    exn = exn_read;    
    prim_err_msg = closed_port;
    return NULL;    
  }
  /* some unread characters are still in the buffer ? */
  if (PORT_INDEX(p) < PORT_EOB(p)) {
    return swit_true;
  } else {
    rdfds = fds;
    if (select(FD_SETSIZE, &rdfds, NULL, NULL, &zero) == -1) {
      prim_err_msg = sys_errlist[errno];
      exn = exn_read;
      return NULL;
    }
    return (FD_ISSET(des, &rdfds) ? swit_true :  swit_false);
  }
}
  

#define LINE_LENGTH 0xfff /* 4095 */

static SWiT
read_line_prim(SWiT arg)
{
  SWiT old_port = cur_in_port;
  int descr;
  char *line = NULL;
  int l = 0;
  if (arg != null_in_port) {
    cur_in_port = arg;
  }
  prompt = 1;
  descr = in_port;
  if (descr == -1) {
    exn = exn_read;    
    prim_err_msg = closed_port;
    cur_in_port = old_port;
    prompt = 1;
    return NULL;    
  }
  if (isatty (descr)) {
    SWiT prompt_val = SYMBOL_VAL(SWIT_SYM(swit_prompt));
    char *prompt_str;
    if SWIT_STRINGP(prompt_val) {
      prompt_str = STRING_ELS(prompt_val);
    } else {
      prompt_str = "> ";
    }
    if (prompt) {
      line = readline(prompt_str);
    } else {
      line = readline("");
    }
    if ((history_length == 0)
	|| (history_search_prefix(line, -1) !=  history_length-1)) {
      add_history(line);
    }
    l = strlen(line);
    if (transcript_descr != -1) {
      if (prompt) write(transcript_descr, prompt_str, strlen(prompt_str));
      write(transcript_descr, line, l);
      write(transcript_descr, "\n", 1);
    }
  } else {
    int c;
    int line_length = LINE_LENGTH;
    line = (char *)swit_malloc_atomic(LINE_LENGTH);
    for (;;) {
      if (read_cur_idx == read_top_idx) {
	/* read more char into read_buffer */
	int cnt;
	if ((cnt = (*read_fun)(descr, read_buffer, IO_BUFFER_SIZE)) != -1) {
	  if (cnt) {
	    read_top_idx = cnt;
	    read_cur_idx = 0;
	  } else {
	    if (l == 0) line = NULL;
	    break;
	  }
	} else {
	  exn = exn_read;
	  prim_err_msg = sys_errlist[errno];
	  cur_in_port = old_port;
	  prompt = 1;
	  return NULL;
	}
      }
      c = read_buffer[read_cur_idx++];
      if (c == '\n') {
	break;
      }
      line[l++] = c;
      if (l == LINE_LENGTH) {
	/* realloc line with 2*line_length */
	line_length = 2*line_length;
	line = (char *)swit_realloc(line, line_length);
      }
    }
  }
  prompt = 1;
  cur_in_port = old_port;
  if (!line) return swit_eof;
  if (l == 0) return swit_null_string;
  return swit_make_string(line, l);
}




static SWiT
make_keyword(char *keyword_name)
{
  SWiT sym = swit_make_symbol(keyword_name, strlen(keyword_name), swit_unbound);
  SET_SYMBOL_FLAG(SWIT_SYM(sym), SYMBOL_KEYWORD_MASK);
  return sym;
}


static SWiT system_prim_sym;
static SWiT
read_system (void)
{
  return swit_make_pair(system_prim_sym, 
			swit_make_pair(tok_val, swit_null_list));
}



/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */


void
swit_init_read (void)
{
  default_prompt = "> ";
  read_table[LITERAL] = read_literal;
  read_table['\''] = read_quote;
  read_table['`'] = read_quasiquote;
  read_table[','] = read_unquote;
  read_table[SPLICE] = read_splice;

  /* for #: syntax : */
  system_prim_sym = swit_make_symbol("system", 6, swit_unbound);

  read_table[SYSTEM] = read_system;
  read_table[READ_EVAL] = read_eval;


  /* a 'zero' timeout for char-ready? */
  zero.tv_sec = 0;
  zero.tv_usec = 0;

  /*------------------------------------------------------------------------
   *   making swit keywords
   *
   * Note : defining a keyword is not just adding a line s.t.
   *
   *       make_keyword("my_new_keyword")
   *
   *  these steps are also needed :
   *   0) declare an instruction code in swit.h
   *   1) the compiler must be modified (to take it into account)
   *   2) add functions (make_ and code: ) in switCode.c to make code with it
   *   3) switEval.c must have a new case to evaluate the new special form
   *   4) add a function in switDisassemble.c to be able to print it
   *   5) add a case in disassamble.swt
   *   6) recompile compiler.swt and disassemble.swt
   *   7) finally, type-alpha.swt must also be able to used it
   *------------------------------------------------------------------------
   */
  make_keyword("lambda");
  make_keyword("if");
  make_keyword("try");
  make_keyword("protect");
  make_keyword("set!");
  make_keyword("begin");
  make_keyword("cond");
  make_keyword("and");
  make_keyword("or");
  make_keyword("case");
  make_keyword("let");
  make_keyword("let*");
  make_keyword("letrec");
  make_keyword("do");
  make_keyword("delay");
  make_keyword("=>");
  make_keyword("define");
  make_keyword("undefine");
  make_keyword("defmacro");
  make_keyword("macrolet");
  make_keyword("step-eval");
  make_keyword("declare");
  make_keyword("signature");

  swit_quote = make_keyword("quote");
  swit_quasiquote = make_keyword("quasiquote");

  /* `contextual keywords'
     (are keywords only in some context (checked by the compiler) */
  swit_else = make_keyword("else");
  swit_unquote = make_keyword("unquote");
  swit_unquote_splicing = make_keyword("unquote-splicing");


/* Input */
  swit_make_prim01 ("read", read_prim,
		   PORT, null_in_port);
  swit_make_prim01 ("read-char", read_char_prim,
		   PORT, null_in_port);
  swit_make_prim01 ("peek-char", peek_char_prim,
		   PORT, null_in_port);
  swit_make_prim1 ("eof-object?", eof_object_p_prim,
		   ANY);
  swit_make_prim01 ("char-ready?", char_ready_p_prim,
		    PORT, null_in_port);
  swit_make_prim01 ("read-line", read_line_prim,
		    PORT, null_in_port);
}

