/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *                         6.7 STRINGS
 *==================================================================
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *    string?              OBJ            -> BOOLEAN
 *    make-string          INTEGER X CHAR -> STRING
 *    string               CHAR LIST -> STRING
 *    string-length        STRING -> INTEGER
 *    string-ref           STRING X INTEGER -> CHAR
 *    string-set!          STRING X INTEGER X CHAR -> CHAR
 *    string=?             STRING X STRING -> BOOLEAN
 *    string-ci=?          STRING X STRING -> BOOLEAN
 *    string<?          STRING X STRING -> BOOLEAN
 *    string>?           STRING X STRING -> BOOLEAN
 *    string<=?           STRING X STRING -> BOOLEAN
 *    string>=?           STRING X STRING -> BOOLEAN
 *    string-ci<?           STRING X STRING -> BOOLEAN
 *    string-ci>?           STRING X STRING -> BOOLEAN
 *    string-ci<=?           STRING X STRING -> BOOLEAN
 *    string-ci>=?           STRING X STRING -> BOOLEAN
 *    substring           STRING X INTEGER X INTEGER -> STRING
 *    string-append      STRING LIST -> STRING
 *    string->list       STRING -> CHAR LIST
 *    list->string       CHAR LIST -> STRING
 *    string-copy        STRING -> STRING
 *    string-fill!       STRING X CHAR -> STRING
 *
 * Extensions :
 * ^^^^^^^^^^ 
 *    string-match               STRING X STRING X FIXNUM -> FIXNUM
 *    string-pos                 STRING X CHAR X FIXNUM -> FIXNUM
 *    string-blt!                STRING X FIXNUM X STRING -> STRING
 *    constant-string?           STRING -> BOOLEAN
 *    primitive->string          PRIMITIVE -> STRING
 *    string-ref2                STRING x FIXNUM -> CHAR (16 bits)
 *    string2->list              STRING -> CHAR LIST   
 *
 *==================================================================
 */

#include "swit.h"

static int memcicmp (char *a1, char *a2, size_t size);

/* globals */
SWiT swit_null_string;


/*====================================
 *           PRINTING
 *======================================
 */
extern void mem2buff();

/* exported to switSymbol.c and switControl.c */
void
print_str (char *str, int len, int esc, int tcl)
{
  if (esc) {
    register i;
    str2buff("\"");
    for (i = 0; i < len; i++, str++) {
      unsigned int c = (unsigned int)*str;
      switch (c)
	{
	case '"': str2buff("\\\"");break;
	case '\\': str2buff("\\\\");break;
	case '{': str2buff("\\{");break; /* tcl */
	case '}': str2buff("\\}");break; /* tcl */
	default: 
	  {
	    if (c < 32) {
	      switch (c)
		{
		case '\n': str2buff("\n");break;
		case '\t': str2buff("\t");break;
		case '\f': str2buff("\f");break;
		case '\r': str2buff("\r");break;
		case '\b': str2buff("\b");break;
		case '\e': str2buff("\e");break;
		default:
		  {
		    char car[4];
		    str2buff("\\");
		    car[0] = '0';
		    car[1] = 48 + c / 8;
		    car[2] = 48 + c % 8;
		    car[3] = '\0';
		    str2buff(car);
		  }
		}
	    } else {
	      char2buff((char)c);
	    }
	  }
	}
    }
    str2buff("\"");
  } else {
    mem2buff(str, len);
  }
}

static void
print_string (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  print_str(STRING_ELS(obj), STRING_LENGTH(obj), esc, tcl);
}


SWiT
swit_make_string (void *chars, unsigned int size)
{
  SWiT str = swit_alloc_object();
  SWIT_TYPE(str) = STRING + (size << 8);
  STRING_ELS(str) = chars;
  return (str);
}

SWiT
swit_make_constant_string (void *chars, unsigned int size)
{
  SWiT str = swit_alloc_object();
  SWIT_TYPE(str) = CONST_STRING + (size << 8);
  STRING_ELS(str) = chars;
  return (str);
}

/*==================================================================
 *                         PRIMITIVES
 *==================================================================
 */


static SWiT
string_p_prim (SWiT arg)
{
  return (SWIT_STRINGP(arg) ? swit_true : swit_false);
}

static SWiT
make_string_prim (SWiT arg1, SWiT arg2)
{
  int size = SWIT_FIXNUM(arg1);
  if (size == 0) {
    return swit_null_string;
  } else if (size > 0) {
    int ch = SWIT_CHAR(arg2);
    if (ch < 256) {
      unsigned char *new = (unsigned char *)swit_malloc_atomic(size+1);
      memset(new, ch, size);
      new[size] = '\0';
      return swit_make_string(new, size);
    } else if (ch < 65536) {
      unsigned char *new = (unsigned char *)swit_malloc_atomic(2*size+1);
      int i;
      unsigned char c1 = ch / 256; 
      unsigned char c2 = ch % 256; 
      for (i = 0; i < 2*size;) {
	new[i++] = c1;
	new[i++] = c2;
      }
      new[i] = '\0';
      return swit_make_string(new, 2*size);
    } else {
      prim_err_msg = wrong_type;
      exn = exn_type;
      exn_val = arg2;
      return NULL;
    }
  } else {
    exn = exn_type;
    prim_err_msg = wrong_type; /* size is negative */
    exn_val = arg1;
    return NULL;
  }
}

static SWiT
string_prim (int argc, SWiT *argv)
{
  if (argc == 0) {
    return swit_null_string;
  } else {
    unsigned char *str = (unsigned char *)swit_malloc_atomic(2*argc+1);
    unsigned char *new = str;
    int len = 0;
    register i;
    for (i = 0; i < argc; i++, len++) {
      int ch = SWIT_CHAR(*argv);
      if (ch < 256) {
	new[i] = ch;
      } else if (ch < 65536) {
	unsigned char c1 = ch / 256; 
	unsigned char c2 = ch % 256; 
	new[i] = c1; new++;
	new[i] = c2;
	len++;
      } else {
	exn = exn_type;
	prim_err_msg = wrong_type;
	exn_val = *argv;
	return NULL;
      }
      argv++;
    }
    str[len] = '\0';
    return swit_make_string(str, len);
  }
}

static SWiT
string_length_prim (SWiT arg)
{
  return swit_make_fixnum(STRING_LENGTH(arg));
}

static SWiT
string_ref_prim (SWiT arg1, SWiT arg2)
{
  unsigned char *str = STRING_ELS(arg1);
  int i = SWIT_FIXNUM(arg2);
  if ((i >= STRING_LENGTH(arg1)) || (i < 0)) {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  } else {
    unsigned char c = str[i];
    return swit_make_char(c);
  }
}

static SWiT
string_ref2_prim (SWiT arg1, SWiT arg2)
{
  unsigned char *str = STRING_ELS(arg1);
  int i = SWIT_FIXNUM(arg2), len = STRING_LENGTH(arg1);
  if ((i >= len-1) || (i < 0)) {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  } else {
    unsigned char c1 = str[i];
    unsigned char c2 = str[i+1];
    return swit_make_char(c1*256+c2);
  }
}

static SWiT
string_set_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  if SWIT_CONST_STRINGP(arg1) {
    exn = exn_type;
    exn_val = arg1;
    prim_err_msg = wrong_type;
    return NULL;
  } else {
    int len = STRING_LENGTH(arg1);
    unsigned char *s = STRING_ELS(arg1);
    int i = SWIT_FIXNUM(arg2);
    int ch = SWIT_CHAR(arg3);
    if ((i >= len) || (i < 0) || ((ch > 255) && (i >= len-1))) {
      prim_err_msg = index_out_of_range;
      exn = exn_range;
      exn_val = arg2;
      return NULL;
    }
    if (ch < 256) {
      s[i] = ch;
      return (arg1);
    } else if (ch < 65536) {
      s[i] = ch / 256;
      s[i+1] = ch % 256;
      return (arg1);
    } else {     
      exn = exn_type;
      exn_val = arg1;
      prim_err_msg = wrong_type;
      return NULL;
    }
  }
}
    
static SWiT
string_eq_prim (SWiT arg1, SWiT arg2)
{
  int len1 = STRING_LENGTH(arg1);
  int len2 = STRING_LENGTH(arg2);
  return ((len1 == len2)
	  && (memcmp (STRING_ELS(arg1), STRING_ELS(arg2), len1) == 0)
	  ? swit_true : swit_false);
}

static SWiT
string_ci_eq_prim (SWiT arg1, SWiT arg2)
{
  int len1 = STRING_LENGTH(arg1);
  int len2 = STRING_LENGTH(arg2);
  return ((len1 == len2)
	  && (memcicmp (STRING_ELS(arg1), STRING_ELS(arg2), len1) == 0)
	  ? swit_true : swit_false);
}

static int
string_lt (SWiT arg1, SWiT arg2)
{
  int len1 = STRING_LENGTH(arg1);
  int len2 = STRING_LENGTH(arg2);
  if (len1 <= len2) {
    int r = memcmp (STRING_ELS(arg1), STRING_ELS(arg2), len1);
    if (r == 0) {
      return (len1 != len2);
    } else {
      return (r < 0);
    }
  } else { /* len1 > len2 */
    int r = memcmp (STRING_ELS(arg1), STRING_ELS(arg2), len2);
    return (r < 0);
  }
}


static SWiT
string_lt_prim (SWiT arg1, SWiT arg2)
{
  return string_lt(arg1, arg2) ? swit_true : swit_false;
}

static SWiT
string_le_prim (SWiT arg1, SWiT arg2)
{
  return string_lt(arg2, arg1) ? swit_false : swit_true;
}

static SWiT
string_gt_prim (SWiT arg1, SWiT arg2)
{
  return string_lt(arg2, arg1) ? swit_true : swit_false;
}

static SWiT
string_ge_prim (SWiT arg1, SWiT arg2)
{
  return string_lt(arg1, arg2) ? swit_false : swit_true;
}

static int
memcicmp (char *a1, char *a2, size_t size)
{
  register i;
  for (i = 0; (i < size) && (toupper(*a1) == toupper(*a2)); i++, a1++, a2++) {
    continue;
  }
  return ((i == size) ? 0 : (int)(toupper(*a1) - toupper(*a2)));
}

static int
string_ci_lt (SWiT arg1, SWiT arg2)
{
  int len1 = STRING_LENGTH(arg1);
  int len2 = STRING_LENGTH(arg2);
  if (len1 <= len2) {
    int r = memcicmp (STRING_ELS(arg1), STRING_ELS(arg2), len1);
    if (r == 0) {
      return (len1 != len2);
    } else {
      return (r < 0);
    }
  } else { /* len1 > len2 */
    int r = memcicmp (STRING_ELS(arg1), STRING_ELS(arg2), len2);
    return (r < 0);
  }
}

static SWiT
string_ci_lt_prim (SWiT arg1, SWiT arg2)
{
  return string_ci_lt(arg1, arg2) ? swit_true : swit_false;
}

static SWiT
string_ci_le_prim (SWiT arg1, SWiT arg2)
{
  return string_ci_lt(arg2, arg1) ? swit_false : swit_true;
}

static SWiT
string_ci_gt_prim (SWiT arg1, SWiT arg2)
{
  return string_ci_lt(arg2, arg1) ? swit_true : swit_false;
}

static SWiT
string_ci_ge_prim (SWiT arg1, SWiT arg2)
{
  return string_ci_lt(arg1, arg2) ? swit_false : swit_true;
}

static SWiT
substring_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  void *from = STRING_ELS(arg1);
  int len = STRING_LENGTH(arg1);
  int start = SWIT_FIXNUM(arg2);
  int end = SWIT_FIXNUM(arg3);
  int size;
  unsigned char *new;

  /* Check bounds */
  if (start < 0) {
    prim_err_msg = out_of_bound1;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  }
  if ((end < start) || (end > len)) {
    prim_err_msg = out_of_bound2;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  }
  size = end-start;
  if (size == 0) return swit_null_string;
  new = swit_malloc_atomic(size+1);
  memcpy(new, from+start, size);
  new[size] = '\0';
  return (swit_make_string(new, size));
}

static SWiT
string_append_prim (int argc, SWiT *argv)
{
  int len = 0;
  unsigned char *new, *to;
  register i;

  /* First compute the sum of lengths, while type checking */
  for (i = 0; i < argc; i++) {
    SWiT arg = argv[i];
    if (!SWIT_STRINGP(arg)) {
      prim_err_msg = must_be_strings;
      exn = exn_type;
      exn_val = arg;
      return NULL;
    }
    len += STRING_LENGTH(arg);
  }
  /* Then make the new string */
  new = swit_malloc_atomic(len+1);
  to = new;
  for (i = 0; i < argc; i++) {
    SWiT arg = argv[i];
    int len = STRING_LENGTH(arg);
    memcpy(to, STRING_ELS(arg), len);
    to += len;
  }
  *to = '\0';
  return (swit_make_string(new, len));
}

static SWiT
string_to_list_prim (SWiT arg)
{
  unsigned char *str = (unsigned char *)STRING_ELS(arg);
  int len = STRING_LENGTH(arg);

  if (len == 0) {
    return(swit_null_list);
  } else {
    SWiT first, last;
    register i;

    first = last = swit_make_pair(swit_make_char(*str), swit_null_list);
    for (i = 1; i < len; i++) {
      int c = str[i];
      SWiT pair = swit_make_pair(swit_make_char(c), swit_null_list);
      SET_CDR(last, pair);
      last = pair;
    }
    return (first);
  }
}

static SWiT
string2_to_list_prim (SWiT arg)
{
  unsigned char *str = (unsigned char *)STRING_ELS(arg);
  int len = STRING_LENGTH(arg);
  if (len % 2) {
    prim_err_msg = length_must_be_even;
    exn = exn_type;
    exn_val = arg;
    return NULL;
  }
  if (len == 0) {
    return(swit_null_list);
  } else {
    SWiT first, last;
    register i;

    first = last = swit_make_pair(swit_make_char((*str)*256 + str[1]), swit_null_list);
    for (i = 2; i < len; i+=2) {
      int c = 256*str[i]+str[i+1];
      SWiT pair = swit_make_pair(swit_make_char(c), swit_null_list);
      SET_CDR(last, pair);
      last = pair;
    }
    return (first);
  }
}

static SWiT
list_to_string_prim (SWiT arg)
{
  if (arg == swit_null_list) {
    return swit_null_string;
  } else {
    int cnt, len = swit_list_length(arg);
    unsigned char *new, *result;
    if (len == 0) {
      prim_err_msg = finite_list;
      exn = exn_type;
      exn_val = arg;
      return NULL;
    }
    if (len < 0) {
#ifdef STRICT
      prim_err_msg = proper_list;
      exn = exn_type;
      exn_val = arg;
      return NULL;
#endif
      len = -len;
    }
    new = swit_malloc_atomic(2*len+1);
    result = new;
    cnt = 0;
    while SWIT_PAIRP(arg) {
      SWiT ch = SWIT_CAR(arg);
      int n = SWIT_CHAR(ch);
      if (n < 256) {
	*new++ = n;
      } else {
	*new++ = n / 256;
	*new++ = n % 256;
	cnt++;
      }
      arg = SWIT_CDR(arg);
    }
    *new = '\0';
    return swit_make_string(result, len+cnt);
  }
}

static SWiT
string_copy_prim (SWiT arg)
{
  int len = STRING_LENGTH(arg);
  unsigned char *new = (unsigned char *)swit_malloc_atomic(len+1);
  memcpy(new, STRING_ELS (arg), len);
  new[len] = '\0';
  return (swit_make_string(new, len));
}

static SWiT
string_fill_prim (SWiT arg1, SWiT arg2)
{
  if SWIT_CONST_STRINGP(arg1) {
    exn = exn_type;
    exn_val = arg1;
    prim_err_msg = wrong_type;
    return NULL;
  } else {
    unsigned char *els = STRING_ELS(arg1);
    int len = STRING_LENGTH(arg1);
    int ch = SWIT_CHAR(arg2);
    if (ch < 256) {
      memset(els, ch, len);
      return (arg1);
    } else if ((ch < 65536) && (len % 2 == 0)) {
      int i;
      unsigned char c1 = ch / 256; 
      unsigned char c2 = ch % 256; 
      for (i = 0; i < len-1;) {
	els[i++] = c1;
	els[i++] = c2;
      }
      return (arg1);
    } else {
      prim_err_msg = wrong_type;
      exn = exn_type;
      exn_val = arg2;
      return NULL;
    }
  }
}

/******* Extensions *******/

/* GNU memmem is wrong ?? (gcc 2.6.3) */

void *
mymemmem (char *s1, int l1, char *s2, int l2)
{
  if (l1 == 0) {
    return s2;
  } else if (l1 > l2) {
    return NULL;
  } else {
    int l = l2;
    char c = *s1, *s = s2;
    while ((s = memchr(s, c, l))) {
      l = l2 + (s2 - s);
      if (l < l1) return NULL;
      if (memcmp(s, s1, l1)) {
	s++;
      } else {
	return s;
      }
    }
    return NULL;
  }
}

static SWiT minus_one; 

/* the string 'arg1' is search into string 'arg2' 
   starting at position `arg3'*/
static SWiT
string_match_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  void *s1 = STRING_ELS(arg1);
  int len1 = STRING_LENGTH (arg1);
  void *s2 = STRING_ELS(arg2);
  int len2 = STRING_LENGTH (arg2);
  int init = SWIT_FIXNUM(arg3);
  if (init < 0) {
    exn = exn_range;
    prim_err_msg = index_out_of_range;
    exn_val = arg3;
    return NULL;
  } else {
    void *s = mymemmem(s1, len1, s2+init, len2-init);
    if (s == NULL) {
      return minus_one;
    } else {
      return swit_make_fixnum(abs((int)(s-s2)));
    }
  }
}


/* char 'arg2' is search into string 'arg1' starting from position 'arg3': 
   if found, return the position otherwise return -1 */
static SWiT
string_pos_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  unsigned char *s = STRING_ELS(arg1);
  int len = STRING_LENGTH(arg1);
  int c = SWIT_CHAR(arg2);
  int init = SWIT_FIXNUM(arg3);
  if (c < 256) {
    if ((init < len) && (init >= 0)){
      register i;
      s += init;
      for (i = init; (i < len) && (*s != c); i++, s++) continue;
      return (i == len ? minus_one  : swit_make_fixnum(i));
    } else {
      goto err;
    }
  } else { /* extended character */
    if ((init < len-1) && (init >= 0)){
      register i;
      unsigned char c1 = c / 256;
      unsigned char c2 = c % 256;
      s += init;
      for (i = init; (i < len-1) && ((*s != c1) || (*(s+1) != c2)); 
	   i+=2, s+=2) continue;
      return (i >= len-1 ? minus_one : swit_make_fixnum(i));
    } else {
      goto err;
    }
  }
 err:
  prim_err_msg = index_out_of_range;
  exn = exn_range;
  exn_val = arg3;
  return NULL;
}

/* 'str2' is block-transfered into 'str1' from position 'start' */
static SWiT
string_blt_prim (SWiT str1, SWiT start, SWiT str2)
{
  if SWIT_CONST_STRINGP(str1) {
    exn = exn_type;
    exn_val = str1;
    prim_err_msg = wrong_type;
    return NULL;
  } else {
    int len1 = STRING_LENGTH(str1);
    int pos = SWIT_FIXNUM(start);
    int start2 = 0;
    if (pos < 0) {
      start2 = -pos;
      pos = 0;
    }
    if (pos < len1) {
      int len2 = STRING_LENGTH(str2);
      if ((len2 + pos) > len1) len2 = len1 - pos;
      if (len2 == 0) {
	return str1;
      } else {
	unsigned char tmp[len2];
	void *to = STRING_ELS(str1) + pos;
	void *from = STRING_ELS(str2) + start2;
	memcpy(tmp, from, len2);
	memcpy(to, tmp, len2);
	return str1;
      }
    } else {
      return str1;
    }
  }
}
static SWiT
primitive_to_string_prim (SWiT obj)
{
  switch SWIT_TYPE(obj)
    {
    case PRIM0:
    case CURRENT_ENV_PRIM:
      return (PRIM_NAME(SWIT_PRIM0(obj)));
    case PRIM01:
    case PRIM1:
    case CAR_PRIM:
    case CDR_PRIM:
    case CALL_CC_PRIM:
    case PRIM0N:
      return (PRIM_NAME(SWIT_PRIM1(obj)));
    case PRIM12:
    case PRIM1N:
    case PRIM2:
    case CALL_WITH_VALUES_PRIM:
    case CERROR_PRIM:
    case RAISE_PRIM:
    case VALUES_PRIM:
      return (PRIM_NAME(SWIT_PRIM2(obj)));
    case PRIM2N:
    case APPLY_PRIM:
    case PRIM3:
      return (PRIM_NAME(SWIT_PRIM3(obj)));
      
    default:
      exn = exn_type;
      exn_val = obj;
      prim_err_msg = wrong_type;
      return NULL;
    }
}

static SWiT
const_string_p_prim (SWiT arg)
{
  return (SWIT_CONST_STRINGP(arg) ? swit_true : swit_false);
}


/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_string (void)
{
  print_table[STRING >> 1] = print_string;
  print_table[CONST_STRING >> 1] = print_string;

  minus_one = swit_make_fixnum(-1);

  swit_make_prim1 ("string?", string_p_prim,
		   ANY);
  swit_make_prim12 ("make-string", make_string_prim,
		    FIXNUM, CHAR, swit_space);
  swit_make_prim0n ("string", string_prim, 
		    CHAR, CHAR);
  swit_make_prim1 ("string-length", string_length_prim,
		   STRING_M);
  swit_make_prim2 ("string-ref", string_ref_prim,
		   STRING_M, FIXNUM);
  swit_make_prim3 ("string-set!", string_set_prim,
		   STRING_M, FIXNUM, CHAR);
  swit_make_prim2 ("string=?", string_eq_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string-ci=?", string_ci_eq_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string<?", string_lt_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string>?", string_gt_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string<=?", string_le_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string>=?", string_ge_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string-ci<?", string_ci_lt_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string-ci>?", string_ci_gt_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string-ci<=?", string_ci_le_prim,
		   STRING_M, STRING_M);
  swit_make_prim2 ("string-ci>=?", string_ci_ge_prim,
		   STRING_M, STRING_M);
  swit_make_prim3 ("substring", substring_prim,
		   STRING_M, FIXNUM, FIXNUM);
  swit_make_prim0n ("string-append", string_append_prim,
		    STRING_M, STRING_M);
  swit_make_prim1 ("string->list", string_to_list_prim,
		   STRING_M);
  swit_make_prim1 ("list->string", list_to_string_prim,
		   LIST);
  swit_make_prim1 ("string-copy", string_copy_prim,
		   STRING_M);
  swit_make_prim2 ("string-fill!", string_fill_prim,
		   STRING_M, CHAR);
  /* extensions */ 
  swit_make_prim2 ("string-ref2", string_ref2_prim,
		   STRING_M, FIXNUM);
  swit_make_prim1 ("string2->list", string2_to_list_prim,
		   STRING_M);
  swit_make_prim3 ("string-match", string_match_prim,
		   STRING_M, STRING_M, FIXNUM);
  swit_make_prim3 ("string-pos", string_pos_prim,
		   STRING_M, CHAR, FIXNUM);
  swit_make_prim3 ("string-blt!", string_blt_prim,
		   STRING_M, FIXNUM, STRING_M);
  swit_make_prim1 ("primitive->string", primitive_to_string_prim, 
		   ANY);
  swit_make_prim1("constant-string?", const_string_p_prim,
		  ANY);
}
