/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *  switSymbol.c --
 *
 *      This file implement primitives on symbols: standard scheme ones
 *      as defined in chapter 6.4 of R4RS and some extensions that may be
 *      useful mainly (but not only) when dealing with symbols produced
 *      with ``gentemp''.
 *
 *
 *                        6.4 SYMBOLS
 *
 * Standard scheme R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 *   symbol?              'a -> BOOLEAN
 *   symbol->string       'a SYMBOL -> STRING
 *   string->symbol        STRING ->  SYMBOL
 *
 * Extensions :
 * ^^^^^^^^^^
 *   keyword?              'a -> BOOLEAN
 *   gentemp?              'a -> BOOLEAN
 *   gentemp              [STRING]     ->   SYMBOL
 *   symbol-value          SYMBOL X [ENV]   ->  'a
 *   unbound-object?       'a    ->  BOOLEAN
 *   set-symbol!           SYMBOL X 'b ->  'b
 *   hash-table-stats      VOID -> #t
 *   apropos               STRING  ->  SYMBOL LIST
 *
 *==================================================================
 */

#include "swit.h"

/* locals (prototypes) */

static SWiT new_symbol(char *name, int len, SWiT init, unsigned int h);
/* r4rs */
static SWiT symbol_p_prim (SWiT arg);
static SWiT string_to_symbol_prim (SWiT arg);
static SWiT symbol_to_string_prim (SWiT arg);
/* extensions */
static SWiT keyword_p_prim (SWiT arg);
static SWiT gentemp_prim (SWiT arg);
static SWiT gentemp_p_prim (SWiT arg);
static SWiT symbol_value_prim (SWiT sym, SWiT env);
static SWiT unbound_object_p_prim (SWiT arg);
static SWiT set_symbol_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT swit_gentemp_default_prefix;
static SWiT hash_stat_prim (void);
static SWiT apropos_prim (SWiT arg);

/* WARNING : also defined in switRead.c */
#define HASH_TABLE_SIZE 1023

static SWiT *global_env;
static SWiT *primitives_table;

/*==========================
 *      PRINTING SYMBOLS
 *==========================
 */
extern void print_str();
void
print_sym (SWiT_Symbol sym)
{
  SWiT name = SYMBOL_NAME(sym);
  print_str(STRING_ELS(name), STRING_LENGTH(name), 0, 0);
}

static void
print_symbol (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  print_sym(SWIT_SYM(obj));
}

static void
print_unbound (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  str2buff("##unbound");
}

/*==========================
 *   SYMBOL CREATION
 *==========================
 */
static SWiT
new_symbol(char *name, int len, SWiT init, unsigned int h)
{
  SWiT sym = swit_alloc_object();
  SWiT_Symbol symbol = SWIT_ALLOC(Symbol);

  SYMBOL_NAME(symbol) = swit_make_constant_string(name, len);
  SET_SYMBOL_NEXT(symbol, global_env[h]);
  /* if name starts with a colon, makes it read-only
   *  and initialized to 'itself' :
   */
  if (*name == ':') {
    SET_SYMBOL_FLAG(symbol, SYMBOL_READ_ONLY_MASK);
    SYMBOL_VAL(symbol) = sym;
  } else {
    SYMBOL_VAL(symbol) = init;
  }
  SYMBOL_PLIST(symbol) = swit_null_list;
  SWIT_TYPE(sym) = SYMBOL;
  SWIT_SYM(sym) = symbol;

  /* save the newly created symbol in the hash table 'global_env' */
  swit_change_stubborn(global_env);
  global_env[h] = sym;
  swit_end_stubborn_change(global_env);
  return (sym);
}

static SWiT
find_symbol (char *name, int len, unsigned int *entry)
{
  register unsigned int h;
  register int i;
  char *str = name;
  SWiT sym;

  /* find entry in hash-table */
  h = 0;
  for (i = 0; i < len; i++) {
    h += (h << 5) + h + *str++;
  }
  h = h % HASH_TABLE_SIZE;

  /* output result */
  *entry = h;

  /* get sym list */
  sym = global_env[h];

  /* look if symbol already exists and returns it, if found */
  while (sym) {
    SWiT_Symbol x = SWIT_SYM(sym);
    SWiT sym_name = SYMBOL_NAME(x);
    if ((len == STRING_LENGTH(sym_name))
	&& (memcmp (name, STRING_ELS(sym_name), len) == 0)) {
      return (sym);
    }
    sym = SYMBOL_NEXT(x);
  }
  /* not found */
  return NULL;
}

/*---------------------------------------------------------------
 * Same as previous one, but uses primitive table.
 * Not used in this file.
 * Exported to switRead.c to read the syntax #'name-of-primitive.
 *---------------------------------------------------------------
 */
SWiT
find_primitive (char *name, int len)
{
  register unsigned int h;
  register int i;
  char *str = name;
  SWiT list;

  /* find entry  */
  h = 0;
  for (i = 0; i < len; i++) {
    h += (h << 5) + h + *str++;
  }
  h = h % HASH_TABLE_SIZE;

  /* get list */
  list = primitives_table[h];

  /* look if it's a name of a primitive */
  while (list) {
    SWiT_Sym_prim0 x = SWIT_PRIM0(list);
    SWiT sym_name = SYMBOL_NAME(x);
    if ((len == STRING_LENGTH(sym_name))
	&& (memcmp (name, STRING_ELS(sym_name), len) == 0)) {
      return (list);
    } else {
      list = PRIM_NEXT(x);
    }
  }
  /* not found */
  return NULL;
}

/*-------------------------------------------------------
 * Makes a new symbol only if it does not already exist.
 * Argument 'init' is ignored if symbol already exists.
 * The length of the name must be given because strings
 * are not required to be 0-terminated (or could include
 * a 0).
 *-------------------------------------------------------
 */
SWiT
swit_make_symbol (char *name, int len, SWiT init)
{
  unsigned int h;
  SWiT sym = find_symbol(name, len, &h);
  if (sym) {
    return (sym);
  } else {
    char *dup = (char *)swit_malloc_atomic(len+1);
    memcpy(dup, name, len);
    dup[len] = '\0';
    return new_symbol(dup, len, init, h);
  }
}


/*---------------------------------------------------------------
 *
 * The next few functions are only called at initialisation,
 * when primitives are 'interned' (into the main hash table,
 * also named the 'global environment').
 *
 * A secondary table, the 'primitives_table',
 * also stores these (precious) definitions: if a symbol
 * naming a primitive is redefined, the primitive is
 * still recoverable through this table and the #' syntax
 * (see switRead.c and switLex.c).
 *
 *--------------------------------------------------------------
 */
void
swit_make_prim0 (char *name, SWiT_Prim0 prim_proc)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim0 fun = SWIT_ALLOC(Sym_prim0);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  if (strcmp(name,"current-env") == 0) {
    SWIT_TYPE(obj) = CURRENT_ENV_PRIM;
  } else {
    SWIT_TYPE(obj) = PRIM0;
  }
  SWIT_PRIM0(obj) = fun;
}

void
swit_make_prim0n (char *name, SWiT_Prim0n prim_proc,
                 int args, int last)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim0n fun = SWIT_ALLOC(Sym_prim0n);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARGS(fun) = args;
  PRIM_T_LAST(fun) = last;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  SWIT_TYPE(obj) = PRIM0N;
  SWIT_PRIM0n(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim1 (char *name, SWiT_Prim1 prim_proc,
		 int arg1)
{
  if (strcmp(name,"call-with-current-continuation") == 0) {
    /* alias with #'call/cc */
    SWiT sym = swit_make_symbol("call-with-current-continuation", 30, swit_unbound);
    /* call/cc is defined before call-with-... (in switControl.c) */
    SYMBOL_VAL(SWIT_SYM(sym)) = find_primitive("call/cc", 7);
  } else {
    unsigned int h;
    SWiT obj = swit_alloc_object();
    int len = strlen(name);
    SWiT sym = find_symbol(name, len, &h);
    SWiT_Sym_prim1 fun = SWIT_ALLOC(Sym_prim1);
    if (!sym) sym = new_symbol(name, len, obj, h);
    PRIM_FUN(fun) = prim_proc;
    PRIM_T_ARG1(fun) = arg1;
    PRIM_VAL(fun) = NULL;
    PRIM_SIG(fun) = NULL;
    PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
    PRIM_NEXT(fun) = primitives_table[h];
    
    swit_change_stubborn(primitives_table);
    primitives_table[h] = obj;
    swit_end_stubborn_change(primitives_table);
    
    if (strcmp(name,"car") == 0) {
      SWIT_TYPE(obj) = CAR_PRIM;
    } else if (strcmp(name,"cdr") == 0) {
      SWIT_TYPE(obj) = CDR_PRIM;
    } else if (strcmp(name,"call/cc") == 0) {
      SWIT_TYPE(obj) = CALL_CC_PRIM;
    } else {
      SWIT_TYPE(obj) = PRIM1;
    }
    SWIT_PRIM1(obj) = fun;
    
    SYMBOL_VAL(SWIT_SYM(sym)) = obj;
  }
}

void
swit_make_prim01 (char *name, SWiT_Prim1 prim_proc,
                 int arg1, SWiT val)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim1 fun = SWIT_ALLOC(Sym_prim1);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_VAL(fun) = val;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  SWIT_TYPE(obj) = PRIM01;
  SWIT_PRIM1(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim1n (char *name, SWiT_Prim1n prim_proc,
                 int arg1, int args, int last)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim1n fun = SWIT_ALLOC(Sym_prim1n);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_T_ARGS(fun) = args;
  PRIM_T_LAST(fun) = last;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  if (strcmp(name,"values") == 0) {
    SWIT_TYPE(obj) = VALUES_PRIM;
  } else {
    SWIT_TYPE(obj) = PRIM1N;
  }
  SWIT_PRIM1n(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim2 (char *name, SWiT_Prim2 prim_proc,
                 int arg1, int arg2)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim2 fun = SWIT_ALLOC(Sym_prim2);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_T_ARG2(fun) = arg2;
  PRIM_VAL(fun) = NULL;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  if (strcmp(name,"call-with-values") == 0) {
    SWIT_TYPE(obj) = CALL_WITH_VALUES_PRIM;
  } else if (strcmp(name,"cerror") == 0) {
    SWIT_TYPE(obj) = CERROR_PRIM;
  } else {
    SWIT_TYPE(obj) = PRIM2;
  }
  SWIT_PRIM2(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim12 (char *name, SWiT_Prim2 prim_proc,
		  int arg1, int arg2, SWiT val)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim2 fun = SWIT_ALLOC(Sym_prim2);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_T_ARG2(fun) = arg2;
  PRIM_VAL(fun) = val;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

    if (strcmp(name,"raise") == 0) {
    SWIT_TYPE(obj) = RAISE_PRIM;
  } else {
    SWIT_TYPE(obj) = PRIM12;
  }
  SWIT_PRIM2(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim2n (char *name, SWiT_Prim2n prim_proc,
		  int arg1, int arg2, int args,
		  int last)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim2n fun = SWIT_ALLOC(Sym_prim2n);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_T_ARG2(fun) = arg2;
  PRIM_T_ARGS(fun) = args;
  PRIM_T_LAST(fun) = last;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  if (strcmp(name,"apply") == 0) {
    SWIT_TYPE(obj) = APPLY_PRIM;
  } else {
    SWIT_TYPE(obj) = PRIM2N;
  }

  SWIT_PRIM2n(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}

void
swit_make_prim3 (char *name, SWiT_Prim3 prim_proc,
                 int arg1, int arg2, int arg3)
{
  unsigned int h;
  SWiT obj = swit_alloc_object();
  SWiT_Sym_prim3 fun = SWIT_ALLOC(Sym_prim3);
  int len = strlen(name);
  SWiT sym = find_symbol(name, len, &h);
  if (!sym) sym = new_symbol(name, len, obj, h);
  PRIM_FUN(fun) = prim_proc;
  PRIM_T_ARG1(fun) = arg1;
  PRIM_T_ARG2(fun) = arg2;
  PRIM_T_ARG3(fun) = arg3;
  PRIM_SIG(fun) = NULL;
  PRIM_NAME(fun) = SYMBOL_NAME(SWIT_SYM(sym)); /* a swit string */
  PRIM_NEXT(fun) = primitives_table[h];

  swit_change_stubborn(primitives_table);
  primitives_table[h] = obj;
  swit_end_stubborn_change(primitives_table);

  SWIT_TYPE(obj) = PRIM3;
  SWIT_PRIM3(obj) = fun;

  SYMBOL_VAL(SWIT_SYM(sym)) = obj;
}


/*----------------------------------------------------------------
 *  SYNTAX : <primitive> == #'name
 *----------------------------------------------------------------
 */
extern SWiT cur_in_port;
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
static SWiT
read_primitive (int token) /* token ignored */
{
  SWiT prim = find_primitive(tok_text, strlen(tok_text));
  if (prim) {
    return (prim);
  } else {
    prim_err_msg = wrong_prim_name;
    exn = exn_read;
    longjmp(*read_jbuf, 2);
  }
}


/*==================================================================
 *                         PRIMITIVES
 *==================================================================
 */

static SWiT
symbol_p_prim (SWiT arg)
{
  return (SWIT_SYMBOLP(arg) ? swit_true : swit_false);
}

static SWiT
symbol_to_string_prim (SWiT arg)
{
  return (SYMBOL_NAME(SWIT_SYM(arg)));
}

static SWiT
string_to_symbol_prim (SWiT arg)
{
  return (swit_make_symbol(STRING_ELS(arg), STRING_LENGTH(arg), swit_unbound));
}

static SWiT
keyword_p_prim (SWiT val)
{
  return (SYMBOL_KEYWORDP(SWIT_SYM(val)) ? swit_true : swit_false);
}


/*
 * ``gentemp'' produces a new interned symbol.
 * Often used in macros to avoid variable names conflicts.
 */

static SWiT
gentemp_prim (SWiT arg)
{
  static unsigned int counter = -1, h;
  char *prefix = STRING_ELS(arg);
  char counter_string[6];
  int i, len = strlen(prefix);
  char *name = (char *)swit_malloc_atomic(sizeof(char)*(strlen(prefix) + 7));
  SWiT sym;

  /* tolower */
  for (i = 0; i < len; i++) name[i] = tolower(prefix[i]);

 start_over:
  counter++;
  if (counter == 1000000) {
    counter = 0;
  }
  name[len] = '\0';
  sprintf (counter_string,"%d", counter);
  strcat(name+len, counter_string);
  if (find_symbol(name, strlen(name), &h)) {
    goto start_over;
  }
  sym = new_symbol(name, strlen(name), swit_unbound, h);
  SET_SYMBOL_FLAG(SWIT_SYM(sym), SYMBOL_GENTEMP_MASK);
  return (sym);
}

static SWiT
gentemp_p_prim (SWiT val)
{
  return (SYMBOL_GENTEMPP(SWIT_SYM(val)) ? swit_true : swit_false);
}

/* exported to switEval.c */
int
search (SWiT sym, SWiT names)
{
  int idx = 0;
  while (names != swit_null_list) {
    if (sym == SWIT_CAR(SWIT_CAR(names))) {
      return idx;
    }
    idx++;
    names = SWIT_CDR(names);
  }
  return -1;
}

static SWiT
symbol_value_prim (SWiT sym, SWiT env)
{
  if (env != swit_void_env) {
    SWiT_Env e = SWIT_ENV(env);
    while (e) {
      SWiT names = ENV_NAMES(e);
      int idx = search(sym, names);
      if (idx != -1) {
	return ENV_FRAME(e)[idx];
      }
      e = ENV_NEXT(e);
    }
    return swit_unbound;
  } else {
    return SYMBOL_VAL(SWIT_SYM(sym));
  }
}

/* exported to switPrint.c */
SWiT
swit_symbol_value (SWiT sym, SWiT env)
{
  return (symbol_value_prim(sym, env));
}

static SWiT
set_symbol_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv)
{
  if (argc == 0) {
    SWiT_Symbol sym = SWIT_SYM(arg1);

    if SYMBOL_READ_ONLYP(sym) {
      prim_err_msg = cannot_set;
      exn = exn_type;
      exn_val = arg1;
      return NULL;
      }
    
    SYMBOL_VAL(sym) = arg2;
    return (arg2);
  } else if (argc == 1) {
    SWiT_Env e = SWIT_ENV(*argv);
    while (e) {
      SWiT names = ENV_NAMES(e);
      int idx = search(arg1, names);
      if (idx != -1) {
	ENV_FRAME(e)[idx] = arg2;
	return arg2;
      }
      e = ENV_NEXT(e);
    }
    prim_err_msg = isunbound;
    exn = exn_unbound;
    exn_val = arg1;
    return NULL;
  } else {
    prim_err_msg = wrong_number;
    exn = exn_arity;
    exn_val = swit_make_fixnum(argc+2);
    return NULL;
  }
}

static SWiT
unbound_object_p_prim (SWiT arg)
{
  return ((arg == swit_unbound) ? swit_true : swit_false);
}

/* Stats on hash table */
static SWiT
hash_stat_prim (void)
{
  int i,total = 0;
  int non_void = 0;
  int result[8];
  double av;
  for (i = 0; i < 8; i++) result[i] = 0;
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = global_env[i];
    int cnt = 0;
    while (syms) {
      cnt++;
      syms = SYMBOL_NEXT(SWIT_SYM(syms));
    }
    total += cnt;
    if (cnt) non_void++;
    switch (cnt) {
    case 0: result[0]++;break;
    case 1: result[1]++;break;
    case 2: result[2]++;break;
    case 3: result[3]++;break;
    case 4: result[4]++;break;
    case 5: result[5]++;break;
    case 6: result[6]++;break;
    default: result[7]++;break;
    }
  }
  av = 0.5*result[1] + result[2] + 1.5*result[3] + 2.0*result[4]
    + 2.5*result[5] + 3.0*result[6] + 3.5*result[7];
  printf("** Hash table stats **\n Buckets info:\n");
  printf("Empty: %d\n",result[0]);
  printf("One entry: %d\n",result[1]);
  printf("Two entries: %d\n",result[2]);
  printf("Three entries: %d\n",result[3]);
  printf("Four entries: %d\n",result[4]);
  printf("Five entries: %d\n",result[5]);
  printf("Six entries: %d\n",result[6]);
  printf("More entries: %d\n",result[7]);
  printf("Total number of symbols: %d\n", total);
  printf("Load factor: %.1f\n", (total*1.0)/non_void);
  printf("Average search distance: %.1f\n", 1 + av/HASH_TABLE_SIZE);
  return swit_true;
}

extern SWiT get_signature_prim (SWiT obj);

/* Check signatures */

static SWiT
check_sig_prim (void)
{
  int i;
  SWiT missing = swit_null_list;
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = primitives_table[i];
    while (syms) {
      if (!PRIM_SIG(SWIT_PRIM0(syms))) {
	missing = swit_make_pair(syms, missing);
      }
      syms = PRIM_NEXT(SWIT_PRIM0(syms));
    }
  }
  return missing;
}

extern SWiT vector_prim (int argc, SWiT *argv);
SWiT
get_prim_prim (void)
{
  int i;
  SWiT prim0 = swit_null_list;
  SWiT prim01 = swit_null_list;
  SWiT prim0n = swit_null_list;
  SWiT prim1 = swit_null_list;
  SWiT prim1n = swit_null_list;
  SWiT prim2 = swit_null_list;
  SWiT prim12 = swit_null_list;
  SWiT prim2n = swit_null_list;
  SWiT prim3 = swit_null_list;
  SWiT prim[9];
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = primitives_table[i];
    while (syms) {
      int type = SWIT_TYPE(syms);
      switch  (type)
	{
	case PRIM0:
	  prim0 = swit_make_pair(syms, prim0);break;
	case PRIM01:
	  prim01 = swit_make_pair(syms, prim01);break;
	case PRIM0N:
	  prim0n = swit_make_pair(syms, prim0n);break;
	case PRIM1:
	  prim1 = swit_make_pair(syms, prim1);break;
	case PRIM12:
	  prim12 = swit_make_pair(syms, prim12);break;
	case PRIM1N:
	  prim1n = swit_make_pair(syms, prim1n);break;
	case PRIM2:
	  prim2 = swit_make_pair(syms, prim2);break;
	case PRIM2N:
	  prim2n = swit_make_pair(syms, prim2n);break;
	case PRIM3:
	  prim3 = swit_make_pair(syms, prim3);break;
	}
      syms = PRIM_NEXT(SWIT_PRIM0(syms));
    }
  }
  prim[0] = prim0;
  prim[1] = prim01;
  prim[2] = prim0n;
  prim[3] = prim1;
  prim[4] = prim12;
  prim[5] = prim1n;
  prim[6] = prim2;
  prim[7] = prim2n;
  prim[8] = prim3;
  return vector_prim(9, prim);
}


extern void *mymemmem (char *s1, int l1, char *s2, int l2);

static SWiT
apropos_prim (SWiT arg)
{
  char *s = STRING_ELS(arg);
  int l = STRING_LENGTH(arg);
  SWiT result = swit_null_list;
  register i;
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = global_env[i];
    while (syms) {
      SWiT_Symbol x = SWIT_SYM(syms);
      SWiT name = SYMBOL_NAME(x);
      if (mymemmem(s, l, STRING_ELS(name), STRING_LENGTH(name)))
	result = swit_make_pair(syms, result);
      syms = SYMBOL_NEXT(x);
    }
  }
  return result;
}

static SWiT
oblist_prim (void)
{
  SWiT result = swit_null_list;
  register i;
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = global_env[i];
    while (syms) {
      if (SYMBOL_VAL(SWIT_SYM(syms)) != swit_unbound) {
	result = swit_make_pair(syms, result);
      }
      syms = SYMBOL_NEXT(SWIT_SYM(syms));
    }
  }
  return result;
}

/*
 * search if an already port is open with this filedes
 * (to be used in switRead.c)
 */

SWiT
search_port(int descr)
{
  register i;
  for (i = 0; i < HASH_TABLE_SIZE; i++) {
    SWiT syms = global_env[i];
    while (syms) {
      SWiT_Symbol s = SWIT_SYM(syms);
      SWiT val = SYMBOL_VAL(s);
      if (val != swit_unbound) {
	if SWIT_PORTP(val) {
	  int des = PORT_DESCR(SWIT_PORT(val));
	  if (descr == des) {
	    return val;
	  }
	}
      }
      syms = SYMBOL_NEXT(SWIT_SYM(syms));
    }
  }
  return swit_false;
}

static SWiT
putprop_prim (SWiT sym1, SWiT sym2, SWiT obj)
{
   SWiT_Symbol symbol = SWIT_SYM(sym1);
  SWiT plist = SYMBOL_PLIST(symbol);
  SWiT plist1 = plist;
  while ((plist1 != swit_null_list) && (SWIT_CAR(plist1) != sym2)) {
    plist1 = SWIT_CDR(SWIT_CDR(plist1));
  }
  if (plist1 == swit_null_list) {
    SYMBOL_PLIST(symbol) = swit_make_pair(sym2, swit_make_pair(obj, plist));
  } else {
    SET_CAR(SWIT_CDR(plist1), obj);
  }
  return (obj);
}

static SWiT
getprop_prim (SWiT sym1, SWiT sym2)
{
  SWiT_Symbol symbol = SWIT_SYM(sym1);
  SWiT plist = SYMBOL_PLIST(symbol);
  while ((plist != swit_null_list) && (SWIT_CAR(plist) != sym2)) {
    plist = SWIT_CDR(SWIT_CDR(plist));
  }
  if (plist == swit_null_list) {
    return swit_unbound;
  } else {
    return SWIT_CAR(SWIT_CDR(plist));
  }
}

static SWiT
remprop_prim (SWiT sym1, SWiT sym2)
{
  SWiT_Symbol symbol = SWIT_SYM(sym1);
  SWiT plist = SYMBOL_PLIST(symbol),plist1;
  plist1 = plist;
  while ((plist != swit_null_list) && (SWIT_CAR(plist) != sym2)) {
    plist1 = SWIT_CDR(plist);
    plist = SWIT_CDR(plist1);
  }
  if (plist == swit_null_list) {
    return swit_false;
  } else if (plist == plist1) {
    SYMBOL_PLIST(symbol) = SWIT_CDR(SWIT_CDR(plist));
    return swit_true;
  } else {
    SET_CDR(plist1, SWIT_CDR(SWIT_CDR(plist)));
    return swit_true;
  }
}

static SWiT
symbol_plist_prim (SWiT sym)
{
  SWiT_Symbol symbol = SWIT_SYM(sym);
  SWiT plist = SYMBOL_PLIST(symbol);
  SWiT result = swit_null_list;
  while (plist != swit_null_list) {
    SWiT cdr = SWIT_CDR(plist);
    result = swit_make_pair(swit_make_pair(SWIT_CAR(plist), SWIT_CAR(cdr)),
			    result);
    plist = SWIT_CDR(cdr);
  }
  return result;
}

static SWiT
help_prim (SWiT arg)
{
  SWiT doc, sig;
  SWiT_Expr argq;
  if SWIT_SYMBOLP(arg) {
    if (SYMBOL_KEYWORDP(SWIT_SYM(arg))) {
      sig = swit_false;
      argq = swit_make_quote(arg);
      goto l;
    } else {
      doc = swit_make_symbol("doc", 3, swit_unbound);
      argq = swit_make_quote(getprop_prim(arg, doc));
      if (swit_apply_proc(":help", 1, &argq)) {
	return swit_true;
      } else {
	return NULL;
      }
    }
  } else if SWIT_PRIMP(arg) {
    argq = swit_make_quote(arg);
    sig = swit_apply_proc(":show-signature", 1, &argq);
    if (!sig) {
      exn = exn_unknown;
      prim_err_msg = sig_not_defined;
      exn_val = arg;
      return NULL;
    }
  l:
    doc = swit_apply_proc(":get-prim-doc", 1, &argq);
    if (doc != swit_false) {
      if (doc) {
	SWiT_Expr docq = swit_make_quote(doc);
	return swit_apply_proc(":help", 1, &docq);
      } else {
	return NULL;
      }
    }
    return sig;
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    exn_val = arg;
    return NULL;
  }
}

static SWiT
set_help_prim (SWiT arg1, SWiT arg2)
{
  if (SWIT_SYMBOLP(arg1)) {
    SWiT docprop = swit_make_symbol("doc", 3, swit_unbound);
    putprop_prim(arg1, docprop, arg2);
  } else if (SWIT_PRIMP(arg1)) {
    ;
    /*    PRIM_DOC(SWIT_PRIM0(arg1)) = arg2; */
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    exn_val = arg1;
    return NULL;
  }
  return arg2;
}

static SWiT
get_code_prim (SWiT arg)
{
  if SWIT_CLOSUREP(arg) {
    return (LAMBDA_PCODE(CLOSURE_LBD(SWIT_CLOSURE(arg))));
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    exn_val = arg;
    return NULL;
  }
}



/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */
void
swit_init_symbol (void)
{
  int i;

  print_table[SYMBOL >> 1] = print_symbol;
  print_table[UNBOUND >> 1] = print_unbound;

  /* GC_stubborn_init();*/
  global_env = (SWiT *)swit_malloc_stubborn(sizeof(struct SWiT)*HASH_TABLE_SIZE);
  primitives_table = (SWiT *)swit_malloc_stubborn(sizeof(struct SWiT)*HASH_TABLE_SIZE);
  for (i = 0; i < HASH_TABLE_SIZE; i++)
    global_env[i] = NULL;

  read_table[PRIMITIVE] = read_primitive;

  swit_gentemp_default_prefix = swit_make_string ("g:", 2);
  swit_make_prim1 ("symbol?", symbol_p_prim,
		   ANY);
  swit_make_prim1 ("symbol->string", symbol_to_string_prim,
		   SYMBOL);
  swit_make_prim1 ("string->symbol", string_to_symbol_prim,
		   STRING_M);
  swit_make_prim01 ("gentemp", gentemp_prim, 
		    STRING_M,
		    swit_gentemp_default_prefix);
  swit_make_prim1 ("gentemp?", gentemp_p_prim,
		   SYMBOL);
  swit_make_prim12 ("symbol-value", symbol_value_prim,
		    SYMBOL, ENV, swit_void_env);
  swit_make_prim1 ("unbound-object?", unbound_object_p_prim,
		   ANY);
  swit_make_prim2n ("set-symbol!", set_symbol_prim,
		    SYMBOL, ANY, ENV, ENV);
  swit_make_prim0 ("hash-table-stats", hash_stat_prim);
  swit_make_prim0 ("check-signatures", check_sig_prim);
  swit_make_prim3 ("putprop!", putprop_prim, 
		   SYMBOL, SYMBOL, ANY);
  swit_make_prim2 ("remprop!", remprop_prim,
		   SYMBOL, ANY);
  swit_make_prim2 ("getprop", getprop_prim, 
		   SYMBOL, SYMBOL);
  swit_make_prim1 ("symbol-plist", symbol_plist_prim,
		   SYMBOL);
  swit_make_prim1 ("apropos", apropos_prim,
		   STRING_M);
  swit_make_prim1 ("help", help_prim,
		   ANY);
  swit_make_prim2 ("set-help!", set_help_prim,
		   ANY, STRING_M);
  swit_make_prim1 ("get-code", get_code_prim,
		   PROC_M);
  swit_make_prim1 ("keyword?", keyword_p_prim,
		   SYMBOL);
  swit_make_prim0 ("get-prim", get_prim_prim);
  swit_make_prim0 ("oblist", oblist_prim);
}
