/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public 
License for more details.

You should have received a copy of the GNU Library General Public License 
along with SWiT software; see the file COPYING.  
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*----------------------------------------------------------------------
 *              TYPES
 *
 *  type-of                 'a -> SYMBOL or LIST
 *  type?                  SYMBOL -> BOOLEAN
 *
 *----------------------------------------------------------------------
 */

#include "swit.h"
#include <math.h>

static SWiT type_of_prim (SWiT arg);

/* exported to switPrint.c */
SWiT swit_class;

SWiT
make_type(char *type_name) 
{
  SWiT sym = swit_make_symbol(type_name, strlen(type_name), swit_unbound);
  SET_SYMBOL_FLAG(SWIT_SYM(sym), SYMBOL_TYPE_MASK);
  return sym;
}

static SWiT
type_p_prim (SWiT val)
{
  return SYMBOL_TYPEP(SWIT_SYM(val)) ? swit_true : swit_false;
}


/*
 * constant types
 */
/* 'type constants' */
SWiT swit_boolean;
SWiT swit_bignum;
SWiT swit_fixnum;
SWiT swit_int;
SWiT swit_ratio;
SWiT swit_zcplx;
SWiT swit_qcplx;
SWiT swit_rcplx;
SWiT swit_real;
SWiT swit_char;
SWiT swit_string;
SWiT swit_vector;
SWiT swit_symbol;
SWiT swit_pair;
SWiT swit_prim;
SWiT swit_obj;
SWiT swit_promise;
SWiT swit_continuation;
SWiT swit_closure;
SWiT swit_primitive;
SWiT swit_environment;
SWiT swit_port;
SWiT swit_inputport;
SWiT swit_outputport;
extern SWiT swit_turtle;
extern SWiT swit_queue;
extern SWiT swit_quad;

extern SWiT swit_symbol_value (SWiT sym, SWiT env);

static SWiT
type_of_prim (SWiT obj)
{
  unsigned int typ = SWIT_TYPE(obj);
  switch (typ & TYPE_MASK1)
    {
    case BOOLEAN:
      return swit_boolean;
    case NUMBER:
      switch (typ)
	{
	case FIXNUM: return swit_fixnum;
	case BIGNUM: return swit_bignum;
	case RATIO: return swit_ratio;
	case QCPLX: return swit_qcplx;
	case ZCPLX: return swit_zcplx;
	case RCPLX: return swit_rcplx;
	default:
	  {
	    double dbl = SWIT_DBL_VAL(obj);
	    if (dbl == floor(dbl)) {
	      return swit_int;
	    } else {
	      return swit_real;
	    }
	  }
	}
    case CHAR:
      return swit_char;
    case EOF_TYPE:
      return swit_eof;
    case PORT:
      return swit_port;
    case QUEUE:
      return swit_queue;
    case ENV: 
      {
	SWiT class = swit_symbol_value(swit_class, obj);
	if (class != swit_unbound) {
	  return class;
	} else {
	  return swit_environment;
	}
      }
    case PROCEDURE:
      switch (typ & TYPE_MASK2) 
	{
	case PRIM:
	  return swit_primitive;
	case CONT:
	  return swit_continuation;
	case CLOSURE:
	  return swit_closure;
	default:
	  return NULL;
	}
    case SYMBOL:
      return swit_symbol;
    case PROMISE:
      return swit_promise;
    case QUAD:
      return swit_quad;
    case TURTLE:
      return swit_turtle;
    case VECT:
      return swit_vector;
    default:
      switch (typ & TYPE_MASK0) 
	{
	case STRING:
	  return swit_string;
	}
      return SWIT_PAIRP(obj) ? swit_pair : swit_obj;
    }
  return swit_obj;
}


static SWiT
set_signature_prim (SWiT prim, SWiT type)
{
  switch SWIT_TYPE(prim)
    {
    case PRIM0:
    case CURRENT_ENV_PRIM:
      PRIM_SIG(SWIT_PRIM0(prim)) = type;
      return type;
    case PRIM0N:
      PRIM_SIG(SWIT_PRIM0n(prim)) = type;
      return type;
    case PRIM01:
    case CAR_PRIM:
    case CDR_PRIM:
    case CALL_CC_PRIM:
    case PRIM1:
      PRIM_SIG(SWIT_PRIM1(prim)) = type;
      return type;
    case PRIM12:
    case PRIM2:
    case CALL_WITH_VALUES_PRIM:
    case CERROR_PRIM:
    case RAISE_PRIM:
      PRIM_SIG(SWIT_PRIM2(prim)) = type;
      return type;
    case PRIM1N:
    case VALUES_PRIM:
      PRIM_SIG(SWIT_PRIM1n(prim)) = type;
      return type;
    case PRIM2N:
    case APPLY_PRIM:
      PRIM_SIG(SWIT_PRIM2n(prim)) = type;
      return type;
    case PRIM3:
      PRIM_SIG(SWIT_PRIM3(prim)) = type;
      return type;
    default:
      exn = exn_type;
      prim_err_msg = wrong_type;
      return NULL;
    }
}

/* exported to switSymbol.c */
SWiT
get_signature_prim (SWiT prim)
{  
  switch SWIT_TYPE(prim)
    {
    case PRIM0:
    case CURRENT_ENV_PRIM:
      return(PRIM_SIG(SWIT_PRIM0(prim)));
    case PRIM0N:
      return(PRIM_SIG(SWIT_PRIM0n(prim)));
    case PRIM01:
    case CAR_PRIM:
    case CDR_PRIM:
    case CALL_CC_PRIM:
    case PRIM1:
      return(PRIM_SIG(SWIT_PRIM1(prim)));
    case PRIM12:
    case PRIM2:
    case CALL_WITH_VALUES_PRIM:
    case CERROR_PRIM:
    case RAISE_PRIM:
      return(PRIM_SIG(SWIT_PRIM2(prim)));
    case PRIM1N:
    case VALUES_PRIM:
      return(PRIM_SIG(SWIT_PRIM1n(prim)));
    case PRIM2N:
    case APPLY_PRIM:
      return(PRIM_SIG(SWIT_PRIM2n(prim)));
    case PRIM3:
      return(PRIM_SIG(SWIT_PRIM3(prim)));
    default:
      exn = exn_type;
      prim_err_msg = wrong_type;
      return NULL;
    }
}
/* defined in switCode.c */
extern SWiT swit_two;
extern SWiT swit_three;
static SWiT swit_max_nargs; /* SWIT_MAX_ARGS (in swit.h) */

SWiT
prim_arity_prim (SWiT prim)
{  
  switch SWIT_TYPE(prim)
    {
    case PRIM0:
    case CURRENT_ENV_PRIM:
      return swit_make_pair(swit_zero, swit_zero);
    case PRIM0N:
      return swit_make_pair(swit_zero, swit_max_nargs);
    case PRIM01:
      return swit_make_pair(swit_zero, swit_one);
    case PRIM1:
    case CAR_PRIM:
    case CDR_PRIM:
    case CALL_CC_PRIM:
      return swit_make_pair(swit_one, swit_one);
    case RAISE_PRIM:
    case PRIM12:
      return swit_make_pair(swit_one, swit_two);
    case PRIM2:
    case CALL_WITH_VALUES_PRIM:
    case CERROR_PRIM:
      return swit_make_pair(swit_two, swit_two);
    case PRIM1N:
    case VALUES_PRIM:
      return swit_make_pair(swit_one, swit_max_nargs);
    case PRIM2N:
    case APPLY_PRIM:
      return swit_make_pair(swit_two, swit_max_nargs);
    case PRIM3:
      return swit_make_pair(swit_three, swit_three);
    default:
      exn = exn_type;
      prim_err_msg = wrong_type;
      return NULL;
    }
}

void
swit_init_type (void)
{
  swit_max_nargs = swit_make_fixnum(0xffff);
  swit_class = swit_make_symbol("class:", 6, swit_unbound);
  swit_make_prim1 ("type-of", type_of_prim, 
		   ANY);
  swit_make_prim1 ("get-signature", get_signature_prim,
		   ANY);
  swit_make_prim1 ("prim-arity", prim_arity_prim,
		   ANY);
  swit_make_prim2 ("set-signature!", set_signature_prim,
		   ANY, ANY);
  swit_make_prim1 ("type?", type_p_prim,
		   SYMBOL);
  swit_boolean = make_type("<bool>");
  swit_bignum = make_type("<bignum>");
  swit_fixnum = make_type("<fixnum>");
  swit_zcplx = make_type("<zcplx>");  
  swit_ratio = make_type("<ratio>");
  swit_qcplx = make_type("<qcplx>");
  swit_rcplx = make_type("<rcplx>");
  swit_int = make_type("<int>");
  swit_real = make_type("<real>");
  swit_char = make_type("<char>");
  swit_string = make_type("<string>");
  swit_vector = make_type("<vector>");
  swit_symbol = make_type("<symbol>");
  swit_promise = make_type("<promise>");
  swit_closure = make_type("<closure>");
  swit_primitive = make_type("<primitive>");
  swit_environment = make_type("<environment>");
  swit_port = make_type("<port>");
  swit_pair = make_type("<pair>");
  swit_obj = make_type("<obj>");
}

