/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *                         6.8 VECTORS
 *==================================================================
 * Standard scheme R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 *   vector?            OBJ                    ->  BOOLEAN
 *   make-vector        NUMBER * OBJ           -> VECTOR
 *   vector-ref         VECTOR * NUMBER        -> OBJ
 *   vector             OBJ * ...              -> VECTOR
 *   vector-length      VECTOR                 -> NUMBER
 *   vector-set!        VECTOR * NUMBER * OBJ  -> OBJ
 *   vector->list       VECTOR                 -> LIST
 *   list->vector       LIST                   -> VECTOR
 *   vector-fill!       VECTOR * OBJ           -> VECTOR
 *
 * Extensions :
 * ^^^^^^^^^^
 *   vector-append      VECTOR * ...             -> VECTOR
 *   vector-copy        VECTOR                   -> VECTOR
 *   vector-reverse!    VECTOR                   -> VECTOR
 *   vector-sort!       VECTOR x PRED            -> VECTOR
 *   binary-search      OBJ x VECTOR x PRED      -> OBJ
 *   vector-blt!        VECTOR X NUMBER X VECTOR -> VECTOR
 *   subvector          VECTOR X NUMBER X NUMBER -> VECTOR
 *==================================================================
 */

#include "swit.h"

SWiT swit_null_vect;

static SWiT vector_p_prim (SWiT arg);
static SWiT make_vector_prim (SWiT arg1, SWiT arg2);
/* SWiT vector_prim (int argc, SWiT *argv); */
static SWiT vector_length_prim (SWiT arg);
static SWiT vector_ref_prim (SWiT arg1, SWiT arg2);
static SWiT vector_set_prim (SWiT arg1, SWiT arg2, SWiT arg3);
static SWiT vector_to_list_prim (SWiT arg);
static SWiT list_to_vector_prim (SWiT arg);
static SWiT vector_fill_prim (SWiT arg1, SWiT arg2);
static SWiT vector_append_prim (int argc, SWiT *argv);
static SWiT vector_nreverse_prim (SWiT arg);
static SWiT vector_copy_prim (SWiT arg);
static SWiT qsort_prim (SWiT arg1, SWiT arg2);
static SWiT bsearch_prim (SWiT arg1,SWiT arg2, SWiT arg3);

/*==================================================================
 *  SYNTAX : <vector> == #( <datum>* )
 *==================================================================
 */
static SWiT
read_vect (void)
{
  int token = get_tok();
  SWiT list = swit_null_list;
  int size = 0;
  while (token != ')') {
    list = swit_make_pair(rd_datum(token), list);
    token = get_tok();
    size++;
  }
  if (size ==  0) {
    return swit_null_vect;
  } else {
    SWiT *vector = swit_alloc_ptrs(size);
    register int i;
    for (i = size-1; i >= 0; --i) {
      vector[i] = SWIT_CAR(list);
      list = SWIT_CDR(list);
    }
    return swit_make_vector(size, vector);
  }
}

static void
print_vector (SWiT vec, char *init, int esc, int tcl, SWiT fv, int depth)
{
  SWiT *els = VECTOR_ELS(vec);
  int size = VECTOR_LENGTH(vec);

  str2buff(init);
  if (size == 0) {
    if (tcl) {
      str2buff("}");
    } else {
      str2buff(")");
    }
  } else {
    int i;
    depth++;
    print_to_buffer(els[0], esc, tcl, fv, depth);
    for (i = 1; i < size; ++i) {
      space2buff();
      print_to_buffer(els[i], esc, tcl, fv, depth);
    }
    if (tcl) {
      str2buff("}");
    } else {
      str2buff(")");
    }
  }
}

static void
print_vect (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  if (tcl)
    print_vector(obj, "{", esc, tcl, fv, depth);
  else
    print_vector(obj, "#(", esc, tcl, fv, depth);
}

static int
equal_vect (SWiT vect1, SWiT vect2)
{
  int s1 = VECTOR_LENGTH(vect1);
  int s2 = VECTOR_LENGTH(vect2);
  if (s1 == s2) {
    SWiT *elts1 = VECTOR_ELS(vect1);
    SWiT *elts2 = VECTOR_ELS(vect2);
    int i;
    for (i = 0 ; i < s1 ; i++) {
      if (! swit_equal(elts1[i], elts2[i])) {
	return 0;
      }
    }
    return 1;
  }
  return 0;
}


/*==================================================================
 *                         PRIMITIVES
 *==================================================================
 */
static SWiT
vector_p_prim (SWiT arg)
{
  return (SWIT_VECTORP(arg) ? swit_true : swit_false);
}

/* Exported (to switRead.c) */
SWiT
swit_make_vector (int size, SWiT *els)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = VECT + (size << 8);
  VECTOR_ELS(obj) = els;
  return (obj);
}


static SWiT
make_vector_prim (SWiT length, SWiT fill)
{
  int size = SWIT_FIXNUM(length);
  if (size == 0) {
    return swit_null_vect;
  } else if (size > 0) {
    SWiT *els = swit_alloc_ptrs(size);
    register i;
    for (i = 0; i < size; i++) {
      els[i] = fill;
    }
    return (swit_make_vector(size, els));
  } else {
    exn = exn_type;
    prim_err_msg = wrong_type;
    return NULL;
  }
}

/* exported to switSymbol.c */
SWiT
vector_prim (int argc, SWiT *argv)
{
  if (argc == 0) {
    return swit_null_vect;
  } else {
    SWiT *els = swit_alloc_ptrs(argc);
    register i;
    for (i = 0; i < argc; i++) {
      els[i] = argv[i];
    }
    return (swit_make_vector(argc, els));
  }
}

static SWiT
vector_length_prim (SWiT arg)
{
  return (swit_make_fixnum (VECTOR_LENGTH(arg)));
}

static SWiT
vector_ref_prim (SWiT arg1, SWiT arg2)
{
  int len = VECTOR_LENGTH(arg1);
  int i = SWIT_FIXNUM (arg2);
  if ((i < len) && (i >= 0)) {
    return VECTOR_ELS(arg1)[i];
  } else {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    return NULL;
  }
}

static SWiT
vector_set_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  int len = VECTOR_LENGTH(arg1);
  int i = SWIT_FIXNUM (arg2);
  if ((i < len) && (i >= 0)) {
    VECTOR_ELS(arg1)[i] = arg3;
    return (arg3);
  } else {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    return NULL;
  }
}


static SWiT
vector_to_list_prim (SWiT vec)
{
  int len = VECTOR_LENGTH (vec);
  if (len == 0) {
    return (swit_null_list);
  } else {
    SWiT first, last;
    SWiT *els = VECTOR_ELS(vec);
    register i;
    first = last = swit_make_pair(els[0], swit_null_list);
    for (i = 1; i < len; i++) {
      SWiT pair = swit_make_pair(els[i], swit_null_list);
      SET_CDR(last, pair);
      last = pair;
    }
    return (first);
  }
}

SWiT
swit_list_to_vector (SWiT list)
{
  if (list == swit_null_list) {
    return (swit_make_vector(0, NULL));
  } else {
    int size = swit_list_length (list);
    SWiT *els;
    register int i = 0;
    if (size == 0) {
      prim_err_msg = finite_list;
      exn = exn_type;
      return NULL;
    }
    if (size < 0) {
#ifdef STRICT
      prim_err_msg = proper_list;
      exn = exn_type;
      return NULL;
#endif
      size = -size;
    }
    els = swit_alloc_ptrs(size);
    while (SWIT_PAIRP(list)) {
      els[i++] = SWIT_CAR (list);
      list = SWIT_CDR (list);
    }
    return (swit_make_vector(size, els)); 
  }
}

static SWiT
list_to_vector_prim (SWiT arg)
{
  return (swit_list_to_vector (arg));
}

static SWiT
vector_fill_prim (SWiT vect, SWiT fill)
{
  int len = VECTOR_LENGTH(vect);
  SWiT *els = VECTOR_ELS(vect);
  register i;
  for (i = 0; i < len; i++) {
    els[i] = fill;
  }
  return (vect);
}


/*********  Extensions **********/

/* 'vect2' is block-transfered into 'vect1' from position 'start' */
static SWiT
vector_blt_prim (SWiT vect1, SWiT start, SWiT vect2)
{
  int len1 = VECTOR_LENGTH(vect1);
  int pos = SWIT_FIXNUM(start);
  int start1 = 0;
  if (pos < 0) {
    start1 = -pos;
    pos = 0;
  }
  if (pos < len1) {
    int len2 = VECTOR_LENGTH(vect2);
    if ((len2 + pos) > len1) len2 = len1 - pos;
    if (len2 == 0) {
      return vect1;
    } else {
      SWiT *to = VECTOR_ELS(vect1)+pos+len2-1;
      SWiT *from = VECTOR_ELS(vect2)+len2-1+start1;
      register i;
      for (i = len2; i > 0; i--) *to-- = *from--;   
      return vect1;
    }
  } else {
    return vect1;
  }
}

static SWiT
vector_nreverse_prim (SWiT arg)
{
  int len = VECTOR_LENGTH(arg) - 1;
  SWiT *els = VECTOR_ELS(arg);
  SWiT tmp;
  register i;
  for (i = 0; i < len/2; i++) {
    tmp = els[i];
    els[i] = els[len-i];
    els[len-i] = tmp;
  }
  return (arg);
}

static SWiT
subvect_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  SWiT *from = VECTOR_ELS(arg1);
  int len = VECTOR_LENGTH(arg1);
  int start = SWIT_FIXNUM(arg2);
  int end = SWIT_FIXNUM(arg3);
  int size;

  /* Check bounds */
  if (start < 0) {
    prim_err_msg = out_of_bound1;
    exn = exn_range;
    return NULL;
  }
  if ((end < start) || (end > len)) {
    prim_err_msg = out_of_bound2;
    exn = exn_range;
    return NULL;
  }
  size = end-start;
  if (size == 0) {
    return swit_null_vect;
  } else {
    SWiT *els = swit_alloc_ptrs(size);
    int i;
    for (i = 0; i < size; i++) {
      els[i] = from[i+start];
    }
    return swit_make_vector(size, els);
  }
}


static SWiT
vector_copy_prim (SWiT arg)
{
  return(vector_append_prim(1,&arg));
}

static SWiT
vector_append_prim (int argc, SWiT *argv)
{
  SWiT *els, *els1;
  int len = 0;
  register i;
  /*
   * Compute the sum of all lengths
   */
  for (i = 0; i < argc; i++) {
    SWiT arg = argv[i];
    len += VECTOR_LENGTH(arg);
  }
  /*
   * Allocate a new vector even if there is only one argument :
   * the situation is not the same as lists (where the first
   * list is not copied).
   */
  els = els1 = swit_alloc_ptrs(len);
  
  /* fill */
  for (i = 0; i < argc; i++) {
    SWiT vec = argv[i];
    SWiT *arg_els = VECTOR_ELS(vec);
    int len = VECTOR_LENGTH(vec);
    int j;
    for (j = 0; j < len; j++) {
      els[j] = arg_els[j];
    }
    els += len;
  }
  return (swit_make_vector(len, els1));
}


static SWiT fun2 = NULL;
static jmp_buf bjbuf;
static int
comp_fnt (SWiT *x, SWiT *y) {
  SWiT vals[2];
  SWiT test;
  vals[0] = *x;
  vals[1] = *y;
  if ((test = swit_apply(fun2, 2, vals)) == NULL) {
    prim_err_msg = bsearch_error;
    exn = exn_unknown;
    longjmp(bjbuf, 1);
  }  
  if (!SWIT_FIXNUMP(test)) {
    prim_err_msg = bsearch_error2;
    exn = exn_type;
    longjmp(bjbuf, 1);
  }
  return SWIT_FIXNUM(test);
}

static SWiT
bsearch_prim (SWiT arg1, SWiT vect, SWiT arg3)
{
  SWiT *res, *els = VECTOR_ELS(vect);
  size_t count = VECTOR_LENGTH(vect);
  if (setjmp(bjbuf)) {
    return NULL;
  }
  fun2 = arg3;
  res = bsearch(&arg1, els, count, sizeof(struct SWiT *), (void *)comp_fnt);
  if (res) {
    return (*res);
  } else {
    return swit_false;
  }
}

static SWiT
qsort_prim (SWiT vect, SWiT arg2)
{
  SWiT *els = VECTOR_ELS(vect);
  int count = VECTOR_LENGTH(vect);
  if (setjmp(bjbuf)) {
    return NULL;
  }
  fun2 = arg2;
  qsort(els, count, sizeof(struct SWiT *), (void *)comp_fnt);
  return vect;
}


static SWiT
set_vmark_prim (SWiT arg1, SWiT arg2)
{
  SWIT_TYPE(arg1) &=  ~0xf0;
  SWIT_TYPE(arg1) |=  (SWIT_FIXNUM(arg2) & 0xf) << 4;
  return arg1;
}

static SWiT
get_vmark_prim (SWiT arg)
{
  return swit_make_fixnum((SWIT_TYPE(arg) & 0xf0) >> 4);
}


/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */
void
swit_init_vector (void)
{
  read_table[VECTOR] = read_vect;
  print_table[VECT >> 1] = print_vect;
  equal_table[VECT >> 1] = equal_vect;

  swit_make_prim1 ("vector?", vector_p_prim,
		   ANY);
  swit_make_prim12 ("make-vector", make_vector_prim,
		    FIXNUM, ANY, swit_null_list);
  swit_make_prim0n ("vector", vector_prim,
		    ANY, ANY);
  swit_make_prim1 ("vector-length", vector_length_prim,
		   VECT_M);
  swit_make_prim2 ("vector-ref", vector_ref_prim,
		   VECT_M, FIXNUM);
  swit_make_prim3 ("vector-set!", vector_set_prim,
		   VECT_M, FIXNUM, ANY);
  swit_make_prim3 ("vector-blt!", vector_blt_prim,
		   VECT_M, FIXNUM, VECT_M);
  swit_make_prim1 ("vector->list", vector_to_list_prim,
		   VECT_M);
  swit_make_prim1 ("list->vector", list_to_vector_prim,
		   LIST);
  swit_make_prim2 ("vector-fill!", vector_fill_prim,
		   VECT_M, ANY);
  swit_make_prim3 ("subvector", subvect_prim,
		   VECT_M, FIXNUM, FIXNUM);
  swit_make_prim0n ("vector-append", vector_append_prim,
		    VECT_M, VECT_M);
  swit_make_prim1 ("vector-reverse!", vector_nreverse_prim,
		   VECT_M);
  swit_make_prim1 ("vector-copy", vector_copy_prim,
		    VECT_M);
  swit_make_prim2 ("vector-sort!", qsort_prim,
		   VECT_M, PROC_M);
  swit_make_prim3 ("binary-search", bsearch_prim,
		   ANY, VECT_M, PROC_M);
  swit_make_prim1 ("get-vmark", get_vmark_prim,
		   VECT_M);
  swit_make_prim2 ("set-vmark!", set_vmark_prim,
		   VECT_M, FIXNUM);
}
